/*****************************************************************************/
/*                                                                           */
/*                                   bif.C                                   */
/*                                                                           */
/*****************************************************************************/

#include "bif.h"       // BIF class definition
#include <math.h>      // for floor() library function

#define BIFDEBUG  1

/*****************************************************************************/
/*                                                                           */
/* method:  BIF::BIF (int sp, FILE *sf)                                      */
/*                                                                           */
/* inputs:  sp - BIF success completion percentage                           */
/*          sf - pointer to problem scratch file                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: This is a HETEROGENEOUS constructor that will create a dummy     */
/*          cell in a list of heterogeneous tasks that the application will  */
/*          fill using the heterogeneous version of AddTask().  Tasks can be */
/*          removed with the heterogeneous version of RemoveTask().          */
/*                                                                           */
/*****************************************************************************/
BIF::BIF (int sp, FILE *sf)
{
     scr = sf;
#ifdef BIFDEBUG
fprintf (scr, "BIF      : BIF::BIF()       : creating heterogeneous BIF\n");
fflush (scr);
#endif

     //
     // Fill BIF variables with dummy data.
     //
     NumberofTasks = 0;           
     SuccessPercentage = sp;
     TaskSourceFile = new char[4];
     strcpy (TaskSourceFile, "nil"); 
     CPU_Estimate = 0;           
     Memory_Estimate = 0;
     Coroutines = 0;

     //
     // Create new task list (circlist constructor creates a dummy cell).
     //
     TaskList = new circlist (scr);
} 

/*****************************************************************************/
/*                                                                           */
/* method:  BIF::BIF (int nt, int sp, char *tsf, FILE *sf, int cpu, int mem) */
/*                                                                           */
/* inputs:  nt  - initial number of tasks in BIF                             */
/*          sp  - BIF success completion percentage                          */
/*          tsf - task source file name                                      */
/*          sf  - pointer to problem scratch file                            */
/*          cpu - estimated CPU requirements                                 */
/*          mem - estimated memory requirements                              */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: This is a HOMOGENEOUS constructor that will build a list of      */
/*          identical subtasks from the parameters passed to it.  Either the */
/*          homogeneous or heterogeneous versions of AddTask() and           */
/*          RemoveTask() can be used to add or remove tasks to this          */
/*          particular BIF.                                                  */
/*                                                                           */
/*****************************************************************************/
BIF::BIF (int nt, int sp, char *tsf, FILE *sf, int cpu, int mem)
{
     scr = sf;
#ifdef BIFDEBUG
fprintf (scr, "BIF      : BIF::BIF()       : creating homogeneous BIF\n");
fflush (scr);
#endif

     //
     // Fill BIF variables with parameter data.
     //
     NumberofTasks = nt;           
     SuccessPercentage = sp;
     TaskSourceFile = new char[strlen (tsf)];
     strcpy (TaskSourceFile, tsf); 
     CPU_Estimate = cpu;           
     Memory_Estimate = mem;
     Coroutines = 0;

     //
     // Build homogeneous task list.
     //
     BuildTaskList (TaskSourceFile, CPU_Estimate, Memory_Estimate);
} 

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::MakeCoroutines ()                                      */
/*                                                                           */
/* inputs:  none                                                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Designate all subtasks in a BIF to be coroutines.  Since the     */
/*          current version of DSC can only wait for coroutines using the    */
/*          dscpr_wait(-1) call, the success percentage is automatically     */
/*          set to 100%.  At this point, this is merely symbolic, since      */
/*          the SuccessPercentage variable will not be referenced in         */
/*          coroutine processing until dscpr_next() can be used to wait      */
/*          on coroutines, at some point in the future.                      */
/*                                                                           */
/*****************************************************************************/
void BIF::MakeCoroutines (void)
{
     Coroutines = 1;
     SuccessPercentage = 100;
}

/*****************************************************************************/
/*                                                                           */
/* method:  int BIF::RunTasks (int retry, int task_type)                     */
/*                                                                           */
/* inputs:  retry     - number of times to tell Wait() to resubmit tasks in  */
/*                      order to achieve BIF success completion percentage   */
/*                                                                           */
/*          task_type - either COROUTINE or 0, the default                   */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  status                                                           */
/*                                                                           */
/* purpose: Submit all subtasks to DSC and wait for successful completion    */
/*          percentage.  If subtasks are not coroutines, kill all incomplete */
/*          subtasks one percentage has been reached.                        */
/*                                                                           */
/*****************************************************************************/
int BIF::RunTasks (int retry, int task_type)
{
 int rval = 0;                  // return code from DSC routines
 int sum = 0;                   // sum of data files (one_level#.d)
 double threshold = 0;          // BIF completion threshold
 char fileprf[DSC_STRINGZ];     // one_level# file name w/o path
 char DataFile[DSC_STRINGZ];    // subproblem solution data file id
 char pwd[DSC_DIRZ];            // current working directory

   //
   // Submit all subproblems in task list, unless the tasks
   //   are RESUMED coroutines.  (The INITIAL submission of
   //   coroutines will use SubmitTasks.)
   //
   if (task_type != COROUTINE)
      if ((rval = SubmitTasks ()) < 0) {
           fprintf (scr, "BIF      : RunTasks()       : error distributing\n");
           fprintf (scr, "BIF      : RunTasks()       : rval %d\n", rval);
           return rval;
      }

   if (Coroutines) {

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RunTasks()       : waiting on coroutines...\n");
fflush (scr);
#endif
        //
        // Wait until all coroutines terminate or go to sleep
        //   and then mark them all complete.
        //
        rval = dscpr_wait (-1);

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RunTasks()       : dscpr_wait rval %d\n", rval);
fflush (scr);
#endif
        if (! rval)            // Zero means success.  Mark all tasks complete.
             MarkComplete ();
   }
   else {
        //
        // Calculate completion threshold value and wait on that number
        //   of subproblems to complete.
        //
        threshold = floor ((SuccessPercentage * .01) * NumberofTasks);

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RunTasks()       : threshold %.0f\n", threshold);
fflush (scr);
#endif

        //
        // Call Wait(), telling it to recurse a maximum of retry times, if 
        //   necessary, to achieve successful completion of threshold.
        //
        if ((rval = Wait (threshold, retry, 0)) != 0) {
             fprintf (scr, "BIF      : RunTasks()       : error in wait \n");
             fprintf (scr, "BIF      : RunTasks()       : rval %d\n", rval);
             return rval;
        }
        //
        // Kill all incomplete subtasks.
        //
        rval = Kill ();
   }

   return rval;
}

/*****************************************************************************/
/*                                                                           */
/* method:  char *BIF::GetDataFile ()                                        */
/*                                                                           */
/* inputs:  none                                                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  name of a data file                                              */
/*                                                                           */
/* purpose: Return name of a data file from a successful task.  Return NULL  */
/*          if no successful tasks are found.                                */
/*                                                                           */
/*****************************************************************************/
char *BIF::GetDataFile (void)
{
   char datafile[DSC_STRINGZ];
   INFO *task;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : GetDataFile()    : reading data file\n");
fflush (scr);
#endif

   //
   // Step through task list, returning name of data files of successfully
   //   completed subtasks.
   //
   for (cell *x = TaskList -> head () ; 
                           ! TaskList -> atEnd (x) ; 
                                         x = x -> advance ("next")) {
       task = TaskList -> data (x);

       if (task -> success) {     
            //
            // Return name of data file if successful and reset success flag.
            //
            strcpy (datafile, task -> fid);
            strcat (datafile, ".d");
            task -> success = 0;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : GetDataFile()    : returning %s\n", datafile);
fflush (scr);
#endif
            return (datafile);
       }
   }
#ifdef BIFDEBUG
fprintf (scr, "BIF      : GetDataFile()    : returning nil\n");
fflush (scr);
#endif

   return "nil";   // Return "nil" if there are no successful tasks.
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::BuildTaskList (char *tsf, int cpu, int mem)            */
/*                                                                           */
/* inputs:  tsf - task source file name                                      */
/*          cpu - estimated CPU requirements                                 */
/*          mem - estimated memory requirements                              */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Build a list of HOMOGENEOUS subtasks.  Do so by calling the      */
/*          heterogeneous version of AddTask().                              */
/*                                                                           */
/*****************************************************************************/
void BIF::BuildTaskList (char *tsf, int cpu, int mem)
{
     char fid[DSC_STRINGZ];         // subproblem file id
     char build[DSC_STRINGZ];       // subproblem build command
     char exec[DSC_STRINGZ];        // subproblem exec command

#ifdef BIFDEBUG
fprintf (scr, "BIF      : BuildTaskList()  : creating task list\n");
fflush (scr);
#endif

     //
     // Create new task list.
     //
     TaskList = new circlist (scr);

#ifdef BIFDEBUG
fprintf (scr, "BIF      : BuildTaskList()  : adding homogeneous tasks\n");
fflush (scr);
#endif

     //
     // Create identical subproblems and add them to the task list.
     //
     for (int y = 0 ; y < NumberofTasks ; y++) {
          BuildStrings (y, fid, build, exec, tsf);

#ifdef BIFDEBUG
fprintf (scr, "BIF      : BuildTaskList()  : fid %s\n", fid);
fprintf (scr, "BIF      : BuildTaskList()  : build %s\n", build);
fprintf (scr, "BIF      : BuildTaskList()  : exec %s\n", exec);
fflush (scr);
#endif
          //
          // Add subproblem to task list.  "0" overrides the default value
          //   of "1" in the heterogeneous parameter, defined in AddTask's
          //   prototype.
          //
          AddTask (y+1, fid, build, exec, cpu, mem, 0);
  }			
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::AddTask ()                                             */
/*                                                                           */
/* inputs:  none                                                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Add a task to a HOMOGENEOUS BIF, incrementing number of tasks in */
/*          BIF and calling heterogeneous AddTask() to do the dirty work.    */
/*                                                                           */
/*****************************************************************************/
void BIF::AddTask (void)
{
     char fid[DSC_STRINGZ];         // subproblem file id
     char build[DSC_STRINGZ];       // subproblem build command
     char exec[DSC_STRINGZ];        // subproblem exec command

#ifdef BIFDEBUG
fprintf (scr, "BIF      : AddTask()        : homogeneous add\n");
fflush (scr);
#endif

     //
     // Create fid, build and exec strings.
     //
     BuildStrings (NumberofTasks, fid, build, exec, TaskSourceFile);

     //
     // Add subproblem to task list, using BIF's CPU and memory usage values.
     //
     AddTask (NumberofTasks, fid, build, exec, CPU_Estimate, Memory_Estimate);
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::AddTask (int i, char *fid, char *build, char *exec,    */
/*                             int cpu, int mem, int heterogeneous)          */
/*                                                                           */
/* inputs:  i             - task index                                       */
/*          fid           - absolute path to source file                     */
/*          build         - build command                                    */
/*          exec          - execute command                                  */
/*          cpu           - estimated CPU requirements                       */
/*          mem           - estimated memory requirements                    */
/*          heterogeneous - 0 = homogeneous add  (default)                   */
/*                          1 = heterogeneous add                            */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Add a task to a BIF.  If this is a heterogeneous add, increment  */
/*          the total number of tasks.  This is the "heterogeneous version"  */
/*          of AddTask, although it is also used on homogeneous BIF's.       */
/*                                                                           */
/*****************************************************************************/
void BIF::AddTask (int i, char *fid, char *build, char *exec, int cpu, int mem,
                   int heterogeneous)
{
#ifdef BIFDEBUG
fprintf (scr, "BIF      : AddTask()        : heterogeneous %d\n", 
         heterogeneous);
fprintf (scr, "BIF      : AddTask()        : index %d\n", i);
fflush (scr);
#endif

     //
     // Initialize cell information.
     //
     INFO *info = new INFO;
     info -> index = i;
     strcpy (info -> fid, fid);
     strcpy (info -> build, build);
     strcpy (info -> exec, exec);
     info -> cpu = cpu;
     info -> mem = mem;
     info -> complete = 0;
     info -> success = 0;

     //
     // Add cell to task list.
     //
     TaskList -> add (info);

#ifdef BIFDEBUG
fprintf (scr, "BIF      : AddTask()        : fid %s\n", fid);
fprintf (scr, "BIF      : AddTask()        : build %s\n", build);
fprintf (scr, "BIF      : AddTask()        : exec %s\n", exec);
fflush (scr);
#endif

     if (heterogeneous) {   // Increment number of tasks for heterogeneous add.

          NumberofTasks++;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : AddTask()        : total tasks %d\n", NumberofTasks);
fflush (scr);
#endif
     }
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::RemoveTask ()                                          */
/*                                                                           */
/* inputs:  none                                                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Remove a task from a HOMOGENEOUS BIF.  Do so by removing the     */
/*          task at the head of the task list and discarding the data.       */
/*          Decrement the BIF's task count.                                  */
/*                                                                           */
/*****************************************************************************/
void BIF::RemoveTask (void)
{
     INFO *x;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RemoveTask()     : removing a homogeneous task\n");
fflush (scr);
#endif

     //
     // Delete task at head of task list.
     //
     x = TaskList -> pop ();
     delete x;

     //
     // Decrement the number of tasks in the BIF.
     //
     --NumberofTasks ;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RemoveTask()     : total tasks %d\n", NumberofTasks);
fflush (scr);
#endif
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::RemoveTask (int index)                                 */
/*                                                                           */
/* inputs:  index - index of task to remove                                  */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Remove a task from a BIF, refereced by index.  This is the       */
/*          "heterogeneous version" of RemoveTask, although it can also be   */
/*          used to remove tasks from a homogeneous BIF.                     */
/*                                                                           */
/*****************************************************************************/
void BIF::RemoveTask (int index)
{
   INFO *task;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RemoveTask()     : removing a heterogeneous task\n");
fflush (scr);
#endif

   //
   // Step through task list, searching for task with matching index.
   //
   for (cell *x = TaskList -> head () ; 
                           ! TaskList -> atEnd (x) ; 
                                         x = x -> advance ("next")) {
       task = TaskList -> data (x);

       if (index == task -> index) {
            //
            // Delete task from BIF and decrement task count.
            //
            TaskList -> remove (x);
            delete task;
            --NumberofTasks ;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RemoveTask()     : removed task %d\n", index);
fflush (scr);
#endif
       }
   }

#ifdef BIFDEBUG
fprintf (scr, "BIF      : RemoveTask()     : total tasks %d\n", NumberofTasks);
fflush (scr);
#endif
}

/*****************************************************************************/
/*                                                                           */
/* method:  int BIF::SubmitTasks (void)                                      */
/*                                                                           */
/* inputs:  none                                                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  index of last successfully submitted task, or error status       */
/*                                                                           */
/* purpose: Submit uncompleted tasks to DSC.                                 */
/*                                                                           */
/*****************************************************************************/
int BIF::SubmitTasks (void)
{
   INFO *task;
   int  rval;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : SubmitTasks()    : submitting tasks\n");
fflush (scr);
#endif

   //
   // Step through task list, submitting only those tasks which are not 
   //   marked complete.  Assign successfully returned indexes to the
   //   respective list elements.
   //
   for (cell *x = TaskList -> head () ; 
                           ! TaskList -> atEnd (x) ; 
                                         x = x -> advance ("next")) {
       task = TaskList -> data (x);

       if (! task -> complete) {     // Don't submit if complete.

#ifdef BIFDEBUG
fprintf (scr, "BIF      : SubmitTasks()    : fid %s\n", task -> fid);
fprintf (scr, "BIF      : SubmitTasks()    : build %s\n", task -> build);
fprintf (scr, "BIF      : SubmitTasks()    : exec %s\n", task -> exec);
fprintf (scr, "BIF      : SubmitTasks()    : cpu %d\n", task -> cpu);
fprintf (scr, "BIF      : SubmitTasks()    : mem %d\n", task -> mem);
fprintf (scr, "BIF      : SubmitTasks()    : complete %d\n", task -> complete);
fprintf (scr, "BIF      : SubmitTasks()    : success %d\n", task -> success);
fflush (scr);
#endif
          //
          // Submit task.
          //
          if ((rval = dscpr_sub (task -> fid,
  	                         "Ci", 
		    	         task -> build, 
		    	         task -> exec, 
                                 task -> cpu, 
                                 task -> mem)) < 0) {

#ifdef BIFDEBUG
fprintf (scr, "BIF      : SubmitTasks()    : error distributing\n");
fprintf (scr, "BIF      : SubmitTasks()    : rval %d\n", rval);
fflush (scr);
#endif
          }
          else {
            task -> index = (long) rval;    // successfully spawned

#ifdef BIFDEBUG
fprintf (scr, "BIF      : SubmitTasks()    : spawned index %d\n", rval);
fflush (scr);
#endif
          }
       }
   }
   return rval;
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::MarkComplete (long index, int status)                  */
/*                                                                           */
/* inputs:  index  - index of subtask                                        */
/*          status - SUCCESS or FAILURE                                      */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Mark a task complete, referenced by its index, and set its       */
/*          success flag to the appropriate value.                           */
/*                                                                           */
/*****************************************************************************/
void BIF::MarkComplete (long index, int status)
{
   INFO *task;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : MarkComplete()   : searching for task %d\n", index);
fflush (scr);
#endif

   //
   // Step through task list, searching for task with matching index.
   //
   for (cell *x = TaskList -> head () ; 
                           ! TaskList -> atEnd (x) ; 
                                         x = x -> advance ("next")) {
       task = TaskList -> data (x);

       if (index == task -> index) {
#ifdef BIFDEBUG
fprintf (scr, "BIF      : MarkComplete()   : marking task %d\n", index);
fflush (scr);
#endif
            task -> complete = 1;
            task -> success = status;
       }
   }
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::MarkComplete ()                                        */
/*                                                                           */
/* inputs:  none                                                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Mark ALL tasks complete and successful.                          */
/*                                                                           */
/*****************************************************************************/
void BIF::MarkComplete (void)
{
   INFO *task;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : MarkComplete()   : Marking all tasks complete\n");
fflush (scr);
#endif

   //
   // Step through task list, marking all tasks complete and successful.
   //
   for (cell *x = TaskList -> head () ; 
                           ! TaskList -> atEnd (x) ; 
                                         x = x -> advance ("next")) {
        task = TaskList -> data (x);

#ifdef BIFDEBUG
fprintf (scr, "BIF      : MarkComplete()   : marking task %d\n", task -> index);
fflush (scr);
#endif
        task -> complete = 1;
        task -> success = 1;
   }
}

/*****************************************************************************/
/*                                                                           */
/* method:  int Wait (double threshold, int recursions, int count)           */
/*                                                                           */
/* inputs:  threshold  - number of tasks that must complete successfully     */
/*          recursions - number of times to attempt to reach threshold       */
/*          count      - recursion depth level                               */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  status                                                           */
/*                                                                           */
/* purpose: Wait on executing subtasks to complete.  Mark a task complete    */
/*          and successful if it completes successfully, complete and failed */
/*          if it terminates with an unrecoverable error, or leave unmarked  */
/*          if it terminates with a recoverable error.  This latter decision */
/*          will allow the task to be resubmitted if necessary.              */
/*                                                                           */
/*          If no more tasks are running and the threshold has not been met, */
/*          resubmit all incomplete tasks again.  Repeat this a maximum      */
/*          number of times defined by the recursions variable.              */
/*                                                                           */
/*****************************************************************************/
int BIF::Wait (double threshold, int recursions, int count)
{
       int    rval = 0;  
       int    wait_on_subtasks = 1;
       long   sweeps = 0;
       static int successes;

       if (count == 0) successes = 0;   // Initialize only first time.

       while (sweeps++ < threshold) {

            //
            // Wait for the next subproblem to complete.
            //
            if (wait_on_subtasks) {

                 rval = dscpr_next();
#ifdef BIFDEBUG
fprintf (scr, "BIF      : Wait()           : dscpr_next rval %d\n", rval);
fflush (scr);
#endif
            }

            //
            // If subproblem ended correctly mark it complete and successful,
            //   and increment the count of successful problems.
            //
            if ((rval >= SUCCESS_LO) && (rval <= SUCCESS_HI)) {
                 MarkComplete ((long) rval, SUCCESS); // rval == problem index
                 successes++ ;
            } 

            //
            // If subproblem exited with a correctable error, don't mark it
            //   complete so that it can be rerun if necessary.  No action is
            //   necessary.
            //

            //
            // If subproblem exited with an uncorrectable error, mark it 
            //   complete so that it will not be rerun.  Also set its success
            //   flag to FAILURE.
            //
            else if ((rval >= UNCORRECTABLE_LO) && (rval <= UNCORRECTABLE_HI)){
#ifdef BIFDEBUG
fprintf (scr, "BIF      : Wait()           : uncorrectable error\n");
fflush (scr);
#endif
                 MarkComplete ((long) -rval, FAILURE); //-rval == problem index
            }
            //
            // If there are no more subtasks to wait on, set wait_on_subtasks
            //   to 0.
            //
            else if (rval == NO_MORE_SUBTASKS) wait_on_subtasks = 0;
       }

#ifdef BIFDEBUG
fprintf (scr, "BIF      : Wait()           : successes %d\n", successes);
fflush (scr);
#endif

       //
       // If enough subtasks have completed successfully, return a value of 0.
       //   Otherwise, if the maximum tries have not been exceeded, resubmit
       //   uncompleted problems and recursively call Wait().
       //
       if (successes == threshold)
            rval = 0;
       else if (++count < recursions) {

#ifdef BIFDEBUG
fprintf (scr, "BIF      : Wait()           : resubmitting incomplete tasks\n");
fprintf (scr, "BIF      : Wait()           : count %d\n", count);
fflush (scr);
#endif
            rval = SubmitTasks ();
            rval = Wait (threshold, recursions, count);
       }
       return rval;
}

/*****************************************************************************/
/*                                                                           */
/* method:  int BIF::Kill (int kill_all)                                     */
/*                                                                           */
/* inputs:  kill_all - flag indicating whether to kill all tasks (0 default) */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  status                                                           */
/*                                                                           */
/* purpose: Kill incomplete tasks if kill_all is 0 (default), or kill all    */
/*          tasks in BIF if kill_all is 1.                                   */
/*                                                                           */
/*****************************************************************************/
int BIF::Kill (int kill_all)
{
   INFO *task;
   int  rval;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : Kill()           : killing incomplete tasks\n");
fflush (scr);
#endif

   //
   // Step through task list, searching for incomplete tasks.
   //
   for (cell *x = TaskList -> head () ; 
                           ! TaskList -> atEnd (x) ; 
                                         x = x -> advance ("next")) {
       task = TaskList -> data (x);

       if ((! task -> complete) || (kill_all)) {
#ifdef BIFDEBUG
fprintf (scr, "BIF      : Kill()           : killing task %d\n", task->index);
fflush (scr);
#endif
            rval = dscpr_kill (task -> index);
       }
   }
   return rval;
}

/*****************************************************************************/
/*                                                                           */
/* method:  int BIF::ResumeTasks (int resume)                                */
/*                                                                           */
/* inputs:  resume - number of times to tell RunTask to resubmit tasks until */
/*                   BIF completion percentage is reached.                   */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  status                                                           */
/*                                                                           */
/* purpose: Resume sleeping coroutines.  Call RunTasks to watch over them.   */
/*                                                                           */
/*****************************************************************************/
int BIF::ResumeTasks (int resume)
{
   INFO *task;
   int  rval;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : ResumeTasks()    : resuming sleeping coroutines\n");
fflush (scr);
#endif

   //
   // Step through task list, resetting the complete and success flags of
   //   each, and resuming.
   //
   for (cell *x = TaskList -> head () ; 
                           ! TaskList -> atEnd (x) ; 
                                         x = x -> advance ("next")) {
       task = TaskList -> data (x);

#ifdef BIFDEBUG
fprintf (scr, "BIF      : ResumeTasks()    : resuming task %d\n", task->index);
fflush (scr);
#endif

       task -> complete = 0;   // Reset flags.
       task -> success = 0;

       rval = dscpr_coresume ((int) task -> index, "");   // Resume coroutines.

#ifdef BIFDEBUG
fprintf (scr, "BIF      : ResumeTasks()    : dscpr_coresume rval %d\n", rval);
fflush (scr);
#endif
   }

   if (rval > -1) {

#ifdef BIFDEBUG
fprintf (scr, "BIF      : ResumeTasks()    : calling RunTasks\n");
fflush (scr);
#endif
        //
        // Call RunTasks to monitor the coroutines.
        //
        RunTasks (resume, COROUTINE);
   }
   return rval;
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::BuildStrings (int y, char *fid, char *build,           */
/*                                  char *exec, char *tsf)                   */
/*                                                                           */
/* inputs:  y     - task index                                               */
/*          fid   - absolute path to source file                             */
/*          build - build command                                            */
/*          exec  - execute command                                          */
/*          tsf   - task source file name                                    */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Create fid, build and exec strings for a task with index y.      */
/*                                                                           */
/*****************************************************************************/
void BIF::BuildStrings (int y, char *fid, char *build, char *exec, char *tsf)
{
     char SourceFile[DSC_STRINGZ];   // full path to source file w/o extention

     //
     // Get absolute path to subproblem source file.
     //
     strcpy (SourceFile, dscpr_getabs (tsf));

     //
     // Create fid string.
     //
     strcpy (fid, SourceFile);
     strcat (fid, "%d");
     sprintf (fid, fid, y);	

     //
     // Create build string.
     //
     strcpy (build, "CC -o "); 
     strcat (build, tsf);
     strcat (build, "%d");
     sprintf (build, build, y);
     strcat (build, " ");
     strcat (build, tsf);
     strcat (build, "%d");
     sprintf (build, build, y); 
     strcat (build, ".C");

     //
     // Create exec string.
     //
     strcpy (exec, tsf);
     strcat (exec, "%d");
     sprintf (exec, exec, y);
     strcat (exec, " ");
     strcat (exec, tsf);
     strcat (exec, "%d.d");
     sprintf (exec, exec, y);
}

/*****************************************************************************/
/*                                                                           */
/* method:  void BIF::Empty()                                                */
/*                                                                           */
/* inputs:  none                                                             */
/*                                                                           */
/* outputs: none                                                             */
/*                                                                           */
/* return:  none                                                             */
/*                                                                           */
/* purpose: Empty all tasks from BIF.                                        */
/*                                                                           */
/*****************************************************************************/
void BIF::Empty (void)
{
   INFO *x;

#ifdef BIFDEBUG
fprintf (scr, "BIF      : Empty()           : Emptying out BIF\n");
fflush (scr);
#endif

   //
   // Pop all tasks off head of list, leaving only the rear dummy cell.
   //
   while ((x = TaskList -> pop ()) != 0) {

#ifdef BIFDEBUG
fprintf (scr, "BIF      : Empty()          : deleting task %d\n", x->index);
fflush (scr);
#endif
       delete x;
   }
}
