/*****************************************************************************/
/*                                                                           */
/*                                dscpr.C                                    */
/*                                                                           */
/* Change History:                                                           */
/* ~~~~~~~~~~~~~~~                                                           */
/* (AAP) - Ran "protoize" to convert function protocols to ANSI.             */
/* (AAP) - Added (DSCKIN_PTR) casts where necessary.                         */
/* (AAP) - Redeclared "setupname" and "pname" as strings, rather than arrays */
/*         of strings.                                                       */
/* (AAP) - Placed extern "C" {...} around all C header files.                */
/* (AAP) - Added ANSI compatible prototypes for external functions.          */
/* (AAP) - Added prototypes for internal functions.                          */
/* (AAP) - Edited #ifdef statements to eliminate compiler warnings.          */
/* (AAP) - Removed unused variables and values.                              */
/* (AAP) - Could not eliminate the following two errors with -ansi -pedantic:*/
/*          too many arguments to function `struct hostent * gethostbyname()'*/
/*          too many arguments to function `void (* signal())()'             */
/*         (Switching from g++ compiler to CC as a result.)                  */
/* (AAP) - Added CC compiler command in ctrl_dispatch.                       */
/* (AAP) - Added Ctyp processing for C++ file recognition.                   */
/* (AAP) - Added stream I/O.                                                 */
/* (AAP) - Added dscpr_getabs() to append a file name to the PWD path.       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* functions :                                                               */
/*   These functions are imbedded into a problem they allow access to DSC    */
/*  for starting subproblems and waiting for answers.                        */
/*                                                                           */
/*     dscpr_start()      - for problems to start subproblems                */
/*     dsc_dispatch()     - to send start problem msg                        */
/*     dscpr_wait()       - wait for one or more subproblems to complete     */
/*     dscpr_kill()       - kill specified subproblem                        */
/*     dscpr_next()       - wait for next subproblem to complete             */
/*     dscpr_cosetup()    - setup when using coroutines                      */
/*     dscpr_cowait       - put a subproblem to sleep                        */
/*     dscpr_coresume()   - awaken a sleeping subproblem                     */
/*     dscdbg_start()     - DEBUGGING FUNCTION                               */
/*                                                                           */
/*   These functions are used for axulliary support                          */
/*                                                                           */
/*     dscpr_reverse()        - reverse a string in place                    */
/*     dscpr_isnum()          - determine if a string consists of digits     */
/*     dscpr_itoa()           - convert a number to characters               */
/*     dsc_read()             - read designed to handle multipart messages   */
/*     dsc_getpass_sub()      -                                              */
/*     dsc_getpass()          -                                              */
/*     dscpr_getnum_sub()     -                                              */
/*     dscpr_getnum()         - get portnumber from the dsc_setup.csh file   */
/*     dscpr_netinit()        - set up local datagram and stream sockets     */
/*     dscpr_netpnm()                                                        */
/*     dscpr_ctrl_portnmchk() - choose a unused port number                  */
/*     dscpr_ctrl_portnm()    - check if a port number is needed             */
/*     dscpr_getabs()         - append a file name to the PWD path           */
/*                                                                           */
/*****************************************************************************/

#include <stream.h> 

#include <ctype.h>
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <errno.h> 
#include <sys/types.h> 
#include <sys/signal.h> 
#include <sys/socket.h> 
#include <sys/time.h> 
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>  
#include <sys/syscall.h>
#include <sys/param.h>

#include "dsc_defs.h"

//
// External Function Prototypes:
//

extern "C" FILE *fopen (const char *, const char *);
extern "C" char *getenv(const char *);
extern "C" char *getcwd (char *, int);
extern "C" int accept (int, struct sockaddr *, int *);
extern "C" int bind (int, struct sockaddr *, int);
extern "C" int close (int);
//extern "C" int execvp (char *, char **);
extern "C" int fork (void);
extern "C" int gethostname (char *, int);
extern "C" int getpid (void);
extern "C" int getsockname (int, struct sockaddr *, int *);
extern "C" int listen (int, int);
//extern "C" int on_exit (void (*)(void), caddr_t);
extern "C" int read (int, char *, int);
//extern "C" int sendto (int, char *, int, int, struct sockaddr *, int);
//extern "C" int setsockopt (int, int, int, char *, int);
//extern "C" int sleep (int);
extern "C" int socket (int, int, int);
//extern "C" struct hostent * gethostbyname(void);
//extern "C" void bcopy (char *, char *, int);
SIG_PF signal (int, SIG_PF);

//
// Internal Function Prototypes:
//

char *dsc_getpass (void);
char *dsc_getpass_sub (char *);
char *dscpr_getnum (void);
char *dscpr_getnum_sub (char *);
char *dscpr_getabs (char *);
int ctrl_dispatch (char *, char *, char *, char *, char *, int, int, int);
int dsc_read (int, DSCKIN_PTR, int);
int dscdbg_start (void);
int dscpr_coresume (int, char *);
int dscpr_ctrl_portnm (void);
int dscpr_ctrl_portnmchk (void);
int dscpr_end (void);
int dscpr_isnum (char *);
int dscpr_kill (long);
int dscpr_next (void);
int dscpr_sub (char *, char *, char *, char *, int, int);
int dscpr_wait (int);
void dsc_cont (void);
void dscpr_cosetup (void);
void dscpr_cowait (int);
void dscpr_itoa (int, char *);
void dscpr_netinit (void);
void dscpr_netpnm (void);
void dscpr_reverse (char *);

static struct sockaddr_in saddr;   /* server datagram address                */
static struct sockaddr_in paddr;   /* this problem stream address            */
static struct sockaddr_in lisaddr; /* non blocking stream address            */
static int    portnm;              /* integer value of datagram port number  */
static char   *cportnm;            /* char value of datagram port number     */
static char   *password=NULL;      /* DSC_PORTNUM text variable              */
static int    mypid;               /* this process's PID                     */
static char   mynode[DSC_NODEZ];   /* local node name                        */
static long   myhost;              /* local host address                     */
static int    netinit=FALSE;       /* sockets initialized flag               */
static int    passinit=FALSE;      /* password set flag                      */
static int    portinit=FALSE;      /* port number set flag                   */
static int    overide=FALSE;       /* portnumber search overide flag         */
static int    sockdg=(-1);         /* udp datagram socket                    */
static int    sockl=(-1);          /* tcp listen socket                      */
static int    lissock=(-1);        /* non blocking stream socket             */
static int    sock=(-1);           /* tcp comm. socket                       */
static int    subcnt=0;            /*                                        */
static int    dsc_cowait_flag = FALSE; /* flag for the sleep state wait loop */
static char   pwd[100];            /* expanded path information              */



/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_reverse(char *s)                                        */
/*                                                                           */
/* inputs:   s - terminated character string                                 */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   reversed string                                                 */
/*                                                                           */
/* purpose:  reverse string s in place                                       */
/*                                                                           */
/*****************************************************************************/

void dscpr_reverse (char *s)
{
   int c, i, j;

   for (i = 0, j = strlen(s)-1; i < j; i++, j--) {
     c=s[i];
     s[i]=s[j];
     s[j]=c;
   }
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_isnum ( chkword )                                       */
/*                                                                           */
/* inputs:   chkword - terminated character string                           */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   TRUE  = only digits in string                                   */
/*           FLASE = non digits in string                                    */ 
/*                                                                           */
/* purpose:  check string for non digits                                     */
/*                                                                           */
/*****************************************************************************/

int dscpr_isnum (char *chkword)
{

 while ( *chkword != '\0' )
 {
   if (!( ( ((int) *chkword) >=48 ) && ( ((int) *chkword) <=57 ) )) 
    return(FALSE);
   chkword++;
 }

 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_itoa( n, s)                                             */
/*                                                                           */
/* inputs:   n - integer value to be converted                               */
/*           s - terminated character string                                 */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  compute the string equivalent of n                              */
/*                                                                           */
/*****************************************************************************/

void dscpr_itoa (int n, char *s)
{
  int i, sign;

  /*             */
  /* record sign */
  /*             */ 
  if ((sign = n) < 0 ) 
    n = -n;           
  i = 0;
  /*                                  */
  /* generate digits in reverse order */
  /*                                  */
  do { 
      s[i++] = n % 10 + '0'; 
  } while ((n /= 10) > 0 ); 
  if (sign < 0)
     s[i++] = '-';
  s[i] = '\0';

  dscpr_reverse(s);

 } 

/*****************************************************************************/
/*                                                                           */
/* subroutine: int dsc_read( d, buf, nbytes)                                 */
/*                                                                           */
/* inputs:   d      - descriptor                                             */
/*           buf    - message buffer                                         */
/*           nbytes - number of bytes to read                                */
/*                                                                           */
/* outputs:  (receive a file)                                                */
/*                                                                           */
/* return:   length of message to read                                       */
/*                                                                           */
/* purpose: this function will piece together a whole message when it        */
/*          arrives in more than one message. ( this has been observed to    */
/*          happen sometimes with messages over 1024  bytes in length,       */
/*          It may happen other times too....)                               */
/*                                                                           */
/*****************************************************************************/

int dsc_read (int d, DSCKIN_PTR buf, int nbytes)
{
  int cc;               /* number of bytes actually  read on success         */
  int len;              /* size of message                                   */
  int remaining;        /* number of remaining bytes to be read              */ 
  char *xbuf;           /* pointer to message buffer                         */

#ifdef DPRRDDEBUG
cerr << "dscpr    : dsc_read         : int d " << d << "\n";
cerr << "dscpr    : dsc_read         : DSCKIN_PTR buf\n";
cerr << "dscpr    : dsc_read         : int nbytes " << nbytes << "\n";
#endif

  cc = read (d, (char *) buf, nbytes);
  if (cc == 0) return(0);         
  else if (cc > 8) len = (int) ntohl (buf->msg_size); 
  else {
    cerr << "dscpr    : dsc_read         : less than 8 bytes received\n";
    cerr << "dscpr    : dsc_read         : " << cc << " bytes recieved\n";
    cerr << "dscpr    : dsc_read         : errno " << errno << "\n";
    cerr << "dscpr    : dsc_read         : fatal error !!!\n";
    exit(errno);
    } 


  xbuf = (char *) buf;
  remaining = len - cc;
  while (remaining > 0) {

cerr << "dscpr    : dsc_read         : another read required\n";
cerr << "dscpr    : dsc_read         : for " << remaining << " bytes\n";

    /*                                    */
    /* calc pointer for begging next read */
    /*                                    */
     xbuf += cc;

    /*                      */
    /* read rest of message */
    /*                      */
     cc = read( d, xbuf, remaining ); 

    /*                                                        */
    /* determine if we are done yet, done when remaining == 0 */
    /*                                                        */ 
     remaining -= cc;
  }

#ifdef DPRRDDEBUG
cerr << "dscpr    : dsc_read         : returning (" << len << ")\n";
#endif

  return( len );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpass_sub( file )                                       */
/*                                                                           */
/* inputs:   file - pointer to memory containing the "dsc_setup.csh" file    */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PASSWORD = if found                                        */
/*            "NOT FOUND"  = if not found                                    */
/*                                                                           */
/* purpose: loop thru memory in search of the DSC_PASSWORD                   */
/*                                                                           */
/*****************************************************************************/

char* dsc_getpass_sub (char *file)
{
   char *word;       /* word in the "dsc_setup.csh" file                     */
   char *wordbeg;    /* pointer to first character                           */
   int found;        /* flag                                                 */

#ifdef DPRGPSDEBUG
cerr << "dscpr    : dsc_getpass_sub  : char* file\n";
#endif 

   found = FALSE;
   word = (char *)malloc(DSC_BUFFZ);
   wordbeg = word;

   /*                                                           */
   /* loop through file searching for the DSC_PASSWORD variable */
   /*                                                           */

  while ( *file != EOF ) {

   /*                     */
   /* skip leading blanks */
   /*                     */
   while ( *file == ' ' ) file ++;

   /*                                          */
   /* get word , a word is delimited by spaces */
   /*                                          */
   while  ( ( *file != ' ') && ( *file != EOF ) )
    { *word = *file;
      word ++; file++;
    }
   *word = '\0';

   /*                                                         */
   /* the pass word is the word after the string DSC_PASSWORD */
   /*                                                         */
   if ( found == TRUE )
     {
#ifdef DPRGPSDEBUG
cerr << "dscpr    : dsc_getpass_sub  : found password " << wordbeg << "\n";
#endif
      return(wordbeg);
     }

   if (strcmp(wordbeg,DSCTEXTPASS) == 0 )  found = TRUE;

   word = wordbeg;

  }

#ifdef DPRGPSDEBUG
cerr << "dscpr    : dsc_getpass_sub  : could not find password\n";
#endif

 return("NOT_FOUND");
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpass()                                                 */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: passinit - to flag password already found                       */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   DSC_PASSWORD on success                                         */
/*           halt otherwise                                                  */
/*                                                                           */
/* purpose:  search the "dsc_setup.csh" file for a password                  */
/*                                                                           */
/*****************************************************************************/

char *dsc_getpass (void)
{
  char setupfile[10000];      /* memory copy of the dsc_setup.csh file       */
  char *homedir;              /* user's HOME directory path                  */
  char setupname[DSC_FILEZ];  /* full path and name of the set up file       */
  char *passwrd;              /* DSC password text variable                  */
  char let;
  int count;
  FILE *fp;                   /* file pointer to dsc_setup.csh               */
  int errcount;

#ifdef DPRGPDEBUG
cerr << "dscpr    : dsc_getpass      : void \n";
#endif

  /*                        */
  /* Create setup file name */
  /*                        */
  homedir = getenv("HOME");
  sprintf(setupname,"%s/",homedir);
  strcat(setupname,DSC_SETUP_NAME);

  fp = fopen (setupname,"r");

  if (fp == NULL) {
   cerr << "dscpr    : dsc_getpass      : can not read " << setupname << "\n";
   cerr << "dscpr    : dsc_getpass      : forcing problem halt!\n";
   cerr << flush;
   exit(DSC_FAILURE);
  }

  /*                                              */
  /* Read set up file and store it in 'setupfile' */
  /*                                              */
  count = 0;
  errcount = 0;

#ifdef DPRGPDEBUG
cerr << "dscpr    : dsc_getpass      : reading " << setupname << "\n";
#endif

   /*                                                                  */
   /* due to problems relating to reading the setup file to soon after */
   /* another server has accessed it the first getc call might return  */
   /* an EOF. Try some repeated fopens.                                */
   /*                                                                  */
   let = getc(fp);

#ifdef GETPDEBUG
if ( let == EOF )
cerr << "dscpr    : dsc_getpass      : 1st read is an EOF\n";
#endif

   while ( ( let == EOF ) && ( errcount < 1000 ) ) {

    fclose(fp);
    fp = fopen (setupname,"r");
    if (fp == NULL) {
    cerr << "dscpr    : dsc_getpass      : in errcount and EOF loop\n";
    cerr << "dscpr    : dsc_getpass      : can not read " << setupname << "\n";
    cerr << "dscpr    : dsc_getpass      : forcing problem halt!\n";
    cerr << flush;
    exit(DSC_FAILURE);
    }
    let = getc(fp);
    errcount++;
#ifdef GETPDEBUG
cerr << "dscpr    : dsc_getpass      : try #" << errcount << "\n";
#endif
   }

   /*                               */
   /* populate 'setupfile' variable */
   /*                               */
   setupfile[count] = let;
   while (let != EOF )
   {
    count++;
    let = getc(fp);
    setupfile[count] = let;
    }
    setupfile[count+1] = EOF;

    fclose(fp);

   /*                                              */
   /* get the password from the dsc_setup.csh file */
   /*                                              */

   passwrd = dsc_getpass_sub(setupfile);

   if ( strcmp(passwrd,"NOT_FOUND") == 0 )
   {
    cerr << "dscpr    : dsc_getpass      : could not find password\n";
    cerr << "dscpr    : dsc_getpass      : in " << setupname << "\n";
    cerr << "dscpr    : dsc_getpass      : forcing problem halt!\n";
    cerr << flush;
    exit(DSC_FAILURE);
   }

 passinit = TRUE;

#ifdef GETPDEBUG
cerr << "dscpr    : dsc_getpass      : returning " << passwrd 
     << " as password\n";
#endif
 
 return ( passwrd );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getnum_sub( file )                                        */
/*                                                                           */
/* inputs:   file - pointer to memory containing the "dsc_setup.csh" file    */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PORT     = if found                                        */
/*            "NOT FOUND"  = if not found                                    */
/*                                                                           */
/* purpose: loop thru memory in search of the DSC_PORT                       */
/*                                                                           */
/*****************************************************************************/

char *dscpr_getnum_sub (char *file)
{
   char *word;       /* word in the "dsc_setup.csh" file                     */
   char *wordbeg;    /* pointer to first character                           */
   int found;        /* flag                                                 */

#ifdef DPRGNSDEBUG
cerr << "dscpr    : dscpr_getnum_sub : char* file\n";
#endif 

   found = FALSE;
   word = (char *)malloc(DSC_BUFFZ);
   wordbeg = word;

   /*                                                       */
   /* loop through file searching for the DSC_PORT variable */
   /*                                                       */

  while ( *file != EOF ) {

   /*                     */
   /* skip leading blanks */
   /*                     */
   while ( *file == ' ' ) file ++;

   /*                                          */
   /* get word , a word is delimited by spaces */
   /*                                          */
   while  ( ( *file != ' ') && ( *file != EOF ) )
    { *word = *file;
      word ++; file++;
    }
   *word = '\0';

   /*                                                         */
   /* the pass word is the word after the string DSC_PASSWORD */
   /*                                                         */
   if ( found == TRUE )
     {
#ifdef DPRGNSDEBUG
cerr << "dscpr    : dscpr_getnum_sub : found portnumber " << wordbeg << "\n";
#endif
      return(wordbeg);
     }

   if (strcmp(wordbeg,DSCTEXTPORT) == 0 )  found = TRUE;

   word = wordbeg;

  }

#ifdef DPRGNSDEBUG
cerr << "dscpr    : dscpr_getnum_sub : could not find password\n";
#endif
 
 return("NOT_FOUND");
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_getabs (file)                                           */
/*                                                                           */
/* inputs:     relative file name                                            */
/*                                                                           */
/* outputs:    none                                                          */
/*                                                                           */
/* return:     absolute path to relative file name                           */
/*                                                                           */
/* purpose:    return absolute path to relative file name parameter          */
/*                                                                           */
/*****************************************************************************/
char *dscpr_getabs (char *file)
{
     char pwd[100];             // current working directory
     char AbsolutePath[132];    // absolute path to file name

    if ( getcwd(pwd,sizeof(pwd)) == NULL )
     {
      cerr << "dscpr    : dscpr_getabs     : getting cwd\n";
      cerr << "dscpr    : dscpr_getabs     : errno " << errno << "\n";  
      exit(5);
     }

     strcpy (AbsolutePath, pwd);
     strcat (AbsolutePath, "/");
     strcat (AbsolutePath, file);

     return (AbsolutePath);
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_getnum()                                                */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PORT     = if found                                        */
/*            "NOT FOUND"  = if not found                                    */
/*                                                                           */
/* purpose:  search the "dsc_setup.csh" file for a portnumber                */
/*                                                                           */
/*****************************************************************************/

char *dscpr_getnum (void)
{
  char setupfile[DSC_SETUPFZ]; /* memory copy of the dsc_setup.csh file      */
  char *homedir;               /* user's HOME directory path                 */
  char setupname[DSC_FILEZ];   /* full path and name of the set up file      */
  char *portnm;                /* DSC port number text variable              */
  char let;
  int count;
  FILE *fp;                    /* file pointer to dsc_setup.csh              */
  int errcount;

#ifdef DPRGNDEBUG
cerr << "dscpr    : dscpr_getnum     : void \n";
#endif

  /*                        */
  /* Create setup file name */
  /*                        */
  homedir = getenv("HOME");
  sprintf(setupname,"%s/",homedir);
  strcat(setupname,DSC_SETUP_NAME);

  fp = fopen (setupname,"r");

  if (fp == NULL) {
   cerr << "dscpr    : dscpr_getnum     : cannot read " << setupname <<"\n";
   cerr << "dscpr    : dscpr_getnum     : forcing problem  halt!\n";
   cerr << flush;
   exit(DSC_FAILURE);
  }

  /*                                              */
  /* Read set up file and store it in 'setupfile' */
  /*                                              */
  count = 0;
  errcount = 0;

#ifdef DPRGNDEBUG
cerr << "dscpr    : dscpr_getnum     : reading" << setupname << "\n";
#endif

   /*                                                                  */
   /* due to problems relating to reading the setup file to soon after */
   /* another server has accessed it the first getc call might return  */
   /* an EOF. Try some repeated fopens.                                */
   /*                                                                  */
   let = getc(fp);

#ifdef DPRGNDEBUG
cerr << "dscpr    : dscpr_getnum     : 1st read is an EOF\n";
#endif

   while ( ( let == EOF ) && ( errcount < 1000 ) ) {

    fclose(fp);
    fp = fopen (setupname,"r");
    if (fp == NULL) {
    cerr << "dscpr    : dscpr_getnum     : in errcount and EOF loop\n";
    cerr << "dscpr    : dscpr_getnum     : can not read " << setupname << "\n";
    cerr << "dscpr    : dscpr_getnum     : forcing problem halt!\n";
    cerr << flush;
    exit(DSC_FAILURE);
    }
    let = getc(fp);
    errcount++;
#ifdef DPRGNDEBUG
cerr << "dscpr    : dscpr_getnum     : try #" << errcount << "\n";
#endif
   }

   /*                               */
   /* populate 'setupfile' variable */
   /*                               */
   setupfile[count] = let;
   while (let != EOF )
   {
    count++;
    let = getc(fp);
    setupfile[count] = let;
    }
    setupfile[count+1] = EOF;

    fclose(fp);

   /*                                                */
   /* get the portnumber from the dsc_setup.csh file */
   /*                                                */

   portnm = dscpr_getnum_sub(setupfile);

   if ( strcmp(portnm,"NOT_FOUND") == 0 )
   {
    cerr << "dscpr    : dscpr_getnum     : could not find port#\n";
    cerr << "dscpr    : dscpr_getnum     : in " << setupname << "\n";
    cerr << "dscpr    : dscpr_getnum     : forcing problem halt!\n";
    cerr << flush;
    exit(DSC_FAILURE);
   }

 portinit = TRUE;

#ifdef DPRGNDEBUG
cerr << "dscpr    : dscpr_getnum     : returning " << portnm 
     << " as portnumber\n";
#endif

 return ( portnm );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_netinit()                                               */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: netinit - to flag network already initialized                   */
/*           sockdg  - datagram socket                                       */
/*           sockl   - listen socket                                         */
/*                                                                           */ 
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  set up host information, and udp/tcp sockets                    */
/*                                                                           */
/*****************************************************************************/
 
void dscpr_netinit (void)
{
   int dscpr_end(void);      /* close socket and unlink the associated file      */
   struct hostent *hp;   /* host information                                 */
   int on;               

#ifdef DPRNIDEBUG
cerr << "dscpr    : dscpr_netinit    : void\n";
#endif 

   if (netinit!=0) return;

   if (sockdg != (-1)) close(sockdg);
   if (sockl != (-1)) close(sockl);
   if (lissock != (-1)) close(lissock);

#ifdef DPRNIDEBUG
cerr << "dscpr    : dscpr_netinit    : geting process info\n";
#endif

   /*                         */
   /* get process information */
   /*                         */ 

   if ( getcwd(pwd,sizeof(pwd)) == NULL )
    {
     cerr << "dscpr    : dscpr_netinit    : getting cwd\n";
     cerr << "dscpr    : dscpr_netinit    : errno " << errno << "\n";
     exit(5);
    }

   mypid = getpid();
   gethostname(mynode, DSC_NODEZ);
   hp = gethostbyname( mynode );
   myhost = (long)( *(long *)hp->h_addr );

#ifdef DPRNIDEBUG
cerr << "dscpr    : dscpr_netinit    : building name of server\n";
#endif

   /*                      */
   /* build name of server */
   /*                      */
   saddr.sin_family = AF_INET;
   bcopy( hp->h_addr, (char *)  &saddr.sin_addr, hp->h_length);
   saddr.sin_port = htons(portnm);

   sock = (-1);

#ifdef DPRNIDEBUG
cerr << "dscpr    : dscpr_netinit    : setting up udp & tcp sockets\n";
#endif

   /*                   */
   /* set up tcp socket */
   /*                   */
   sockdg = socket(AF_INET, SOCK_DGRAM, 0);
   if (sockdg < 0) {
     cerr << "dscpr    : dscpr_netinit    : can not build udp socket\n";
     cerr << "dscpr    : dscpr_netinit    : forced problem halt!\n";
     perror( "building udp socket");
     exit(errno);
    }
   on = 1;
   setsockopt( sockdg, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));

   sockl = socket(AF_INET, SOCK_STREAM, 0);
   if (sockl < 0) {
     cerr << "dscpr    : dscpr_netinit    : can not build tcp socket\n";
     cerr << "dscpr    : dscpr_netinit    : forced problem halt!\n";
     perror( "building tcp socket");
     exit(errno);
    }
   on = 1;
   setsockopt( sockl, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));
   bcopy( hp->h_addr, (char *) &paddr.sin_addr, hp->h_length);
   paddr.sin_port = htons(0);
   paddr.sin_family = AF_INET;
   if (bind (sockl, (struct sockaddr *) &paddr, sizeof (paddr))){
     cerr << "dscpr    : dscpr_netinit    : can't bind streamsocket\n";
     perror("binding stream socket");
     exit(errno);
   }
   on = sizeof(paddr);
   if (getsockname(sockl, (struct sockaddr *) &paddr, &on)){
     cerr << "dscpr    : dscpr_netinit    : can't get socket name\n";
     perror( "getting socket name ");
     exit(errno);
   }

   listen( sockl, 3);

 /*                                                                        */
 /* The reason that the sockl stream socket created in dscpr_netinit() can */
 /* not be used is that it is a socket that blocks. A blocking socket is   */
 /* one that when there are I/O requests that cannot complete immediately  */
 /* the process that issued the I/O request using the blocking socket will */
 /* suspend awaiting completion. When the viability of the server is being */
 /* checked and if sockl were to be used the accept() connection command   */
 /* will hang if the server is not activated. Once a socket has been       */
 /* created via the socket() call, it may be marked as non-blocking by the */
 /* fcntl() call, thus avoiding the hanging of the dscctrl program if the  */
 /* server died in an irregular manner (ie rebooting of machines).         */
 /*                                                                        */

 /*                                      */
 /* get a descriptor for a strean socket */
 /*                                      */
  lissock = socket(AF_INET, SOCK_STREAM,0);
  if ( lissock < 0 ) {
     cerr << "dscpr    : dscpr_netinit    : can't bld lissock tcp\n";
     cerr << "dscpr    : dscpr_netinit    : forced problem halt!\n";
     perror("building tcp sockect");
     exit(errno);
  }

 /*                     */
 /* set sockect options */
 /*                     */
 on = 1;
 setsockopt( lissock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));

 /*                              */
 /* mark lissock as non-blocking */
 /*                              */
 if (fcntl(lissock, F_SETFL, FNDELAY) < 0 ) {
   cerr << "dscpr    : dscpr_netinit    : can't mark as non-blocking\n";
   cerr << "dscpr    : dscpr_netinit    : forced problem halt!\n";
   perror("fcntl F_SETFL SNDELAY");
   exit(errno);
 }

 /*                    */
 /* give socket a name */
 /*                    */
 bcopy( hp->h_addr, (char *) &lisaddr.sin_addr, hp->h_length );
 lisaddr.sin_port = htons(0);
 lisaddr.sin_family = AF_INET;
  if (bind( lissock, (struct sockaddr *) &lisaddr, sizeof(lisaddr))) {
cerr << "dscpr    : dscpr_netinit    : can't bind lissock stream\n";
perror("binding stream socket");
printf("errno is %d \n",errno);
exit(errno);
  }

  /*                             */
  /* get socket name information */
  /*                             */
  on = sizeof(lisaddr);
  if (getsockname(lissock, (struct sockaddr *) &lisaddr, &on)) {
cerr << "dscpr    : dscpr_netinit    : can't get lissock socket name\n";
perror("getting socket name");
exit(errno);
  }

  /*                                                                        */ 
  /* Set upper bound on number of pending connections that should be queued */
  /*                                                                        */
  listen( lissock, 3);

  netinit = TRUE;

#ifdef sun
// on_exit (dscpr_end);      Commented out for now... (AAP)
#endif

#ifdef DPRNIDEBUG
cerr << "dscpr    : dscpr_netinit    : init complete <<  flag set\n";
cerr << "dscpr    : dscpr_netinit    : returning void\n";
#endif

   return;
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_netpnm()                                                */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: portnm  - datagram socket port number                           */
/*           netinit - socket setup flag                                     */
/*           overide - overide dscpr_netpnm call, use user supplied port #   */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  get the DSC_PASSWORD from the setup file                        */ 
/*           get the DSC_PORT number from the setup file                     */
/*           if one is found test for an active server                       */
/*           if not search for an available portnumber                       */
/*           terminate if no valid portnumber can be found                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_netpnm (void)
{
   int status;           /* port number queuery status                       */
   char *dsc_getpass(void);  /* get password from dsc_setup.csh file             */ 

#ifdef DPRNPNDEBUG
cerr << "dscpr    : dscpr_netpnm     : void\n";
#endif

   if (netinit!=0) return;

   /*                                                        */
   /* set the external password variable to the DSC_PASSWORD */
   /*                                                        */
   password = dsc_getpass();

   if (overide == TRUE) { dscpr_netinit(); return; }

   /*                                               */
   /* get socket number from the dsc_setup.csh file */
   /*                                               */
   cportnm = dscpr_getnum();
   if (strcmp(cportnm,"NOT_FOUND") == 0) { 
   cerr << "dscpr    : dscpr_netpnm     : did not find portnumber\n";
   strcpy(cportnm,"0");
   }

   portnm = atoi(cportnm);
  
#ifdef DPRNPNDEBUG
cerr << "dscpr    : dscpr_netpnm     : using port number" << portnm << "\n";
#endif

 
   /*                                                                     */
   /* If the port number found in the file dsc_setp.csh is zero then this */
   /* indicates that there is no activated server, and thus a unused port */
   /* number must be found and bound to. However if the port number is    */
   /* non zero then local communication ends must be established in order */
   /* to verify the existance of the server program.                      */
   /*                                                                     */

   if (portnm != 0) 
    {
#ifdef DPRNPNDEBUG
cerr << "dscpr    : dscpr_netpnm     : nonzero port number found\n";
cerr << "dscpr    : dscpr_netpnm     : initializing sockets\n";
#endif
     dscpr_netinit();  
    }

#ifdef DPRNPNDEBUG
cerr << "dscpr    : dscpr_netpnm     : checking viability of server\n";
cerr << "dscpr    : dscpr_netpnm     : or searching for port number\n";
#endif

   status = dscpr_ctrl_portnmchk();
   if ( status != DSC_SUCCESS ) 
    {
cerr << "dscpr    : dscpr_netpnm     : cannot connect to server\n";
cerr << "dscpr    : dscpr_netpnm     : or no valid portnumber found\n";
cerr << "dscpr    : dscpr_netpnm     : forced problem halt!\n";
exit(DSC_FAILURE);
    }

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_ctrl_portnmchk()                                        */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: portnm  - datagram socket port number                           */
/*           netinit - socket setup flag                                     */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   DSC_SUCCESS = valid port number verified or found               */
/*           DSC_FAILURE = can not find a portnumber                         */
/*                                                                           */
/* purpose:  if 0 port number search for a valid port number and initialize  */
/*           else if server connection established use port number if not    */
/*           search for a valid port number and initialize                   */
/*                                                                           */
/*****************************************************************************/

int dscpr_ctrl_portnmchk (void)
{
int on;                /* switch for the socket option SO_REUSEADDR          */
int status;
int count;
struct hostent *hp;    /* host information                                   */
DSCQSTA msg;           /* datagram server status message                     */
DSCSSTA msg2;          /* stream socket status message                       */


#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : void\n";
#endif

/*                                                                         */
/* according to the dsc_setup.csh file, no operational server on this node */
/*                                                                         */
if ( portnm == 0 ) {

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : no operational server\n";
cerr << "dscpr    : dscpr_ctrl_pchk  : 0 port number\n";
#endif

 cout << "no operational server on this node \n";

 /*                         */
 /* get process information */
 /*                         */

   if ( getcwd(pwd,sizeof(pwd)) == NULL )
    {
     cerr << "dscpr    : dscpr_ctrl_pchk  : getting cwd\n";
     cerr << "dscpr    : dscpr_ctrl_pchk  : errno " << errno << "\n";
     exit(5);
    }

 mypid = getpid();
 gethostname(mynode,DSC_NODEZ);
 hp = gethostbyname( mynode );
 myhost = (long)( *(long *)hp->h_addr);

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : searching for available port#\n";
#endif

 dscpr_ctrl_portnm();

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : initializing sockets\n";
#endif

 netinit = FALSE;
 dscpr_netinit();

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : returning DSC_SUCCESS\n";
#endif

 return(DSC_SUCCESS);

}
 
 /*                                                                          */
 /* In the case that the DSC_PORT number is not zero the value may or not be */
 /* valid depending on the termination condition of the server program       */
 /* (ie rebooting of machines). Therefore the viability of the server        */
 /* program must be checked before the port number is assigned               */
 /*                                                                          */
 else {

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : checking for server at " << portnm
     << "\n";
#endif

 /*                              */
 /* build status request message */
 /*                              */
 msg.msg_type = msg.sanity = htonl(MSG_QSTAT);
 strcpy(msg.dsc_password , password );
 msg.msg_size = htonl(sizeof( DSCQSTA ));
 msg.flag = htonl(MSG_SSTAT);
 memcpy( &msg.addr, &lisaddr, sizeof(lisaddr));

 /*                                  */
 /* send message to local DSC server */
 /*                                  */
 do {
   errno = 0;
   status = sendto (sockdg, (char *) &msg, sizeof (DSCQSTA), 0,
                    (struct sockaddr *) &saddr, sizeof (saddr));
   if ( errno == ENOBUFS ) sleep(2);
  } while ( ( errno == EINTR ) || ( errno == ENOBUFS ) );

 /*                                            */
 /* try to accept a connection with the server */
 /*                                            */
 count = 0;
 do {
   errno = 0;
   sock = accept( lissock, 0, 0 );
   count = count + 1;
 } while ( (count <= 1000) && (errno > 0 ) );

 /*                                                    */
 /* search for a valid port number, no connection made */
 /*                                                    */
 if  ( errno > 0 )  {

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : no operational server on " << portnm
     << "\n";
#endif

  cout << "no operational server on " << portnm 
       << " (cannot connect via DSC_PORT number)\n" ;

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : searching for available port#\n";
#endif

  dscpr_ctrl_portnm();

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : initializing sockets\n";
#endif

  netinit = FALSE;
  dscpr_netinit();

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : returning DSC_SUCCESS\n";
#endif

  return(DSC_SUCCESS);

 }

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : reading status message \n";
#endif

 /*                           */
 /* get server status message */
 /*                           */
 on=1;
 setsockopt( sock,SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));

 errno = EWOULDBLOCK;
 while ( errno == EWOULDBLOCK )
	 {
  errno = 0;
  status = read ( sock, (char *)&msg2, sizeof(DSCSSTA) );
 }

 if ( status != ntohl(msg2.msg_size) ) {
  cerr << "dscpr    : dscpr_ctrl_pchk  : msg size does not match\n";
  close( lissock );
  close( sock );
  sock = -1;
  return(DSC_FAILURE);
 }

 close( sock );
  sock = -1;

 } 

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_pchk  : returning DSC_SUCCESS\n";
#endif

 return(DSC_SUCCESS);

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_ctrl_portnmchk()                                        */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: portnm  - datagram socket port number                           */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose: execute $DSC_PATH/Netcheck/dsc_nc and use the file dsc_us        */
/*          to determine an unused port number  for datagram communication.  */ 
/*          (Inspected processors are found in the file dscrp_net.add)       */
/*                                                                           */
/*****************************************************************************/

int dscpr_ctrl_portnm (void)
{
  int pid;                 /* the parent process gets the id of the new      */
                           /* process, while the child process gets the      */
                           /* value of 0                                     */
  int rstat;
  int currportnm;
  int n=0;
  FILE *portfp;
  char pname[DSC_FILEZ];
  char *dpath;
  union wait statusp;      /* terminated process status                      */
  static char *argv[10];   /* command line for the execvp                    */


#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: void\n";
#endif

  pid = fork();
  if (pid == 0 ) {

     argv[0] = "rsh";
     argv[1] = mynode;
     argv[2] = "csh";
     argv[3] = "$DSC_PATH/Netcheck/dsc_nc";
     argv[4] = 0;

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: " << argv[0] << " " << argv[1] << " "
     << argv[2] << " " << argv[3] << "\n";
cerr << flush;
#endif

   /*                                                                        */
   /* The child process immediately transfers control to the rsh program     */
   /* The transfer is achieved with the execvp function. This function makes */
   /* a request to the UNIX system to replace the current process (the one   */
   /* calling execvp) with the new process (rsh)                             */
   /*                                                                        */

   pid = execvp( "rsh", argv ); 
   /*                                        */
   /* if control is returned here then error */
   /*                                        */
#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: control is returned <<  error\n";
cerr << "dscpr    : dscpr_ctrl_portnm: errno " << errno << "\n";
cerr << flush;
#endif
   perror( "failure in execvp");
   exit(errno);
  } 
  else {

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: checking for available port#\n";
cerr << flush;
#endif

  cout << "checking for available port numbers ... \n";

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: waiting on " << pid << "\n";
cerr << flush;
#endif
   
  wait(&statusp);

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: child done!\n";
cerr << flush;
#endif

/*
  if ( wait_result != pid )
   {
cerr << "dscpr    : dscpr_ctrl_portnm: can not find a port number!\n";
cerr << "dscpr    : dscpr_ctrl_portnm: wait_result " << wait_result 
     << " != PID " << pid << "\n";
cerr << "dscpr    : dscpr_ctrl_portnm: errno " << errno << "\n";
cerr << "dscpr    : dscpr_ctrl_portnm: forced halt!\n";
cerr << flush;
exit(errno);    
   }
*/
 /*                                                                        */
 /* Read dsc_usedsockets file and determine a usable portnumber to bind to */
 /*                                                                        */
  dpath = getenv("DSC_PATH");
  sprintf(pname,"%s/Netcheck/",dpath);
  strcat(pname,"dsc_us");

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: opening " << pname << "\n";
cerr << flush;
#endif

  portfp = fopen(pname,"r");
  if (portfp == NULL) {

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: can not open " << pname << "\n";
#endif

   cout << "can not open " << pname << "\n";
   rstat = dscpr_end();
   exit(rstat);
  }

 /*                                                        */
 /* Find last used port number from the dsc_usedports file */
 /*                                                        */

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: finding last used portnumber\n";
cerr << flush;
#endif

 currportnm = 1; n=1;
 while ( n == 1) 
  { 
    n =  fscanf(portfp,"%d",&currportnm); 
#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: number " << currportnm << " rval "
     << n << "\n";
cerr << flush;
#endif
 
  }
 fclose(portfp);

 /*                                */
 /* Use next available port number */
 /*                                */
 if (currportnm > 1) portnm = currportnm + 1;

#ifdef DPRCPNDEBUG 
cerr << "dscpr    : dscpr_ctrl_portnm: using port number " << portnm << "\n";
#endif

  } /* end if parent */

#ifdef DPRCPNCDEBUG
cerr << "dscpr    : dscpr_ctrl_portnm: returning DSC_SUCCESS\n";
#endif

 return(DSC_SUCCESS);

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_sub()                                                   */
/*                                                                           */
/* inputs:    fid   - file id of problem file                                */
/*            types - types of input files (c,l,o,i,p)                       */
/*            build - build command                                          */
/*            exec  - execute command                                        */
/*            cpu   - estimated cpu usage (LOW_CPU,MED_CPU,HIGH_CPU)         */
/*            mem   - estimated mem usage in megs                            */
/*                                                                           */
/* ext vars: portnm  - datagram socket port number                           */
/*           cportnm - character string socket port number                   */
/*           netinit - to flag network already initialized                   */
/*           sock    - datagram socket                                       */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   index to new subproblem in DSC Server database                  */
/*                                                                           */
/* purpose:  build message to create new subproblem                          */
/*           send the message to the DSC RP                                  */ 
/*                                                                           */
/*****************************************************************************/

int dscpr_sub (char *fid, char *types, char *build, char *exec, 
               int cpu, int mem)
{
   char pname[DSC_NAMEZ+1];  /* problem name                                 */
   int rstat;                /* status of dscpr_sub                          */

#ifdef DPRSUBDEBUG
cerr << "dscpr    : dscpr_sub        : fid " << fid << "\n";
cerr << "dscpr    : dscpr_sub        : types " << types << "\n";
cerr << "dscpr    : dscpr_sub        : exec " << exec << "\n";
cerr << "dscpr    : dscpr_sub        : cpu " << cpu << "\n";
cerr << "dscpr    : dscpr_sub        : mem " << mem << "\n";
#endif

   if ( portinit == FALSE )
   {
    /*                                                                    */
    /* get the datagram communication port number from the dsc setup file */
    /*                                                                    */
    cportnm = dscpr_getnum();
    if (strcmp(cportnm,"NOT_FOUND") == 0) {
       cerr << "dscpr    : dscpr_sub        : port number not found\n";
       cerr << "dscpr    : dscpr_sub        : forced problem halt!\n";
       rstat = dscpr_end();
       exit(rstat);
    }
   /*                            */
   /* convert it to decimal form */
   /*                            */
   portnm = atoi(cportnm);
   }

   /*                              */
   /* set up communication sockets */
   /*                              */
   dscpr_netinit();

   sprintf( pname, "DSC%4.4x%d", mypid&0xffff, ++subcnt);
   rstat = ctrl_dispatch(pname, fid, types, build, exec, 1, cpu, mem );

   /*                                                          */
   /* small wait so successive builds on-node are out of synch */
   /* if too many children complete at same time, then signals */
   /* get crossed and may not be delivered. This is a kluge    */
   /* because some implimentations of UNIX are flakey          */
   /*                                                          */

   sleep(2);

#ifdef DPRSUBDEBUG
cerr << "dscpr    : dscpr_sub        : returning " << rstat << "\n";
#endif

   return( rstat );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  ctrl_dispatch()                                              */
/*                                                                           */
/* inputs:   probname - unique name of problem                               */
/*           fid      - source file id                                       */
/*           types    - source file types                                    */
/*           build    - command to build source file into executable         */
/*           exec     - command to run executable file                       */
/*           level    - 0 then main problem (must run on local node)         */
/*                      > 0  then subproblem                                 */
/*           cpu      - estimated cpu usage of the problem                   */
/*           mem      - estimated mem usage of the problem                   */
/*                                                                           */
/* ext vars: passinit - password set flag                                    */
/*                                                                           */
/* outputs:  (msg to DSC server)                                             */
/*                                                                           */
/* return:   index of problem within server database                         */
/*                                                                           */
/* purpose:  send start message to DSC Server                                */
/*           wait for response with index of problem                         */
/*                                                                           */
/*****************************************************************************/
int ctrl_dispatch (char *probname, char *fid, char *types, char *build,
                   char *exec, int level, int cpu, int mem)
{
   DSCSTRT msg;          /* start subproblem message                         */
   DSCBEG *msg2;         /* incomming startup message                        */
   int     rlen;         /* lenght of read message                           */
   int     on;
   char   *cptr;
   char   *ctypes;
   char *dsc_getpass(void);  /* get password from dsc_setup.csh file             */
   char *dscpr_getnum(void); /* get port numeber from dsc_setup.csh file         */

#ifdef DPRCDISDEBUG
cerr << "dscpr    : ctrl_dispatch    : probname " << probname << "\n";
cerr << "dscpr    : ctrl_dispatch    : fid " << fid << "\n";
cerr << "dscpr    : ctrl_dispatch    : types " << types << "\n";
cerr << "dscpr    : ctrl_dispatch    : build " << build << "\n";
cerr << "dscpr    : ctrl_dispatch    : exec " << exec << "\n";
cerr << "dscpr    : ctrl_dispatch    : level " << level << "\n";
cerr << "dscpr    : ctrl_dispatch    : cpu " << cpu << "\n";
cerr << "dscpr    : ctrl_dispatch    : mem " << mem << "\n";
#endif

   /*                                    */
   /* save pointer to beginning of types */
   /*                                    */
   ctypes = types;

   if ( passinit == FALSE )
   {

#ifdef DPRCDISDEBUG
cerr << "dscpr    : ctrl_dispatch    : getting password\n";
cerr << flush;
#endif

    password = dsc_getpass();
   }


#ifdef DPRCDISDEBUG
cerr << "dscpr    : ctrl_dispatch    : setup start and begin mess\n";
#endif

   /*                                 */
   /* set up start and begin messages */
   /*                                 */
   msg2 = (DSCBEG *)&msg;
   msg.msg_type = msg.sanity = htonl(MSG_START);
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl(sizeof( DSCSTRT ));
   memcpy( &msg.addr, &paddr, sizeof(paddr));

   msg.parent_PID = htonl(mypid);
   msg.fnode = htonl(myhost);
   msg.subflg = htonl(level);
   msg.cpurating = htonl(cpu);
   msg.memrating = htonl(mem);
   strcpy( msg.prob_name, probname);

   /*                                               */
   /* main problems must deal with path information */
   /*                                               */
   if (level == 0) {
      cptr = strchr( fid, '/');
      if (cptr == 0) {
        sprintf( msg.src_name, "%s/%s", pwd, fid);
        cptr = fid;
        }
      else {
        strcpy( msg.src_name, fid);
        cptr++;
        }
     /* special case for off-node servers */
      if (myhost != saddr.sin_addr.s_addr) {
        strcpy( msg.src_name, fid);
        }
    }/* end if main problem */
   else {
     cptr = fid;
     strcpy( msg.src_name, cptr );
   } /* end if subproblem */

   msg.ctyp = 0;
   msg.Ctyp = 0;          // (AAP)
   msg.ltyp = 0;
   msg.mtyp = 0;
   msg.ptyp = 0;
   msg.ityp = 0;
   msg.otyp = 0;
   msg.dtyp = 0;
 

 /*                                                               */
 /* Set flags and build string if using the Lisp interface        */
 /*                                                               */
 if  (strcmp(ctypes,"lc")==0) 
     {
     msg.ltyp = 1;
     strcpy( msg.src_build, build);
     }
 else if (strcmp(ctypes,"lci")==0) 
     {
     msg.ltyp = 1;
     msg.ityp = 1; 
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
      strcpy( msg.src_build, "*");
     }
 else if (strcmp(ctypes,"lo")==0) 
     {
     msg.ltyp = 1;
     msg.otyp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
      strcpy( msg.src_build, "*");
     }
 else if (strcmp(ctypes,"loi")==0) 
     {
     msg.ltyp = 1;
     msg.otyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
      strcpy( msg.src_build, "*");
     }
 else if (strcmp(ctypes,"li")==0)
     {
     msg.ltyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  else if (strcmp(ctypes,"l")==0)
     {
     msg.ltyp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  /*                                                               */
  /* Set flags and build string if using the Mathematica interface */
  /*                                                               */
  else if (strcmp(ctypes,"m")==0)
     {
     msg.mtyp = 1;
          strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  else if (strcmp(ctypes,"mi")==0)
     {
     msg.mtyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  /*                                                               */
  /* Set flags and build string if using the Maple interface       */
  /*                                                               */
  else if (strcmp(ctypes,"p")==0)
     {
     msg.ptyp = 1;
          strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  else if (strcmp(ctypes,"pi")==0)
     {
     msg.ptyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
    }
 /*                                                               */
 /* Set flags and build string if using the C or C++ interface.   */
 /*                                                               */
 else
 {
   while (*types != '\0') {
      switch (*types) {
      case 'c' :
        msg.ctyp = 1;
        if (*build == '*') msg.src_build[0] = '\0';
        else if (strlen( build ))    strcpy( msg.src_build, build);
        else sprintf( msg.src_build,
		     "cc -o %s %s.c", msg.src_name, msg.src_name);
        break;
      case 'C' :
        msg.Ctyp = 1;
        if (*build == '*') msg.src_build[0] = '\0';
        else if (strlen( build ))    strcpy( msg.src_build, build);
        else sprintf( msg.src_build,
		     "CC -o %s %s.C", msg.src_name, msg.src_name);
        break;
      case 'i':
        msg.ityp = 1;
        if (strlen(msg.src_build) == 0)
	  strcpy( msg.src_build, "*");
        break;
      }              /* end switch */
    types++;         /* advance to next character for examination */
    }                /* end while more types */

 } /* End else */

   msg.cpu_type = htonl(DSC_CPU);

#ifdef DPRCDISDEBUG
cerr << "dscpr    : ctrl_dispatch    : msg.src_build " << msg.src_build << "\n";
cerr << flush;
#endif

   /*                                                               */
   /* Set default execute string if using the Lisp interface        */
   /*                                                               */
   if ((strcmp(ctypes,"li")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.int %s.lsp %s.d %s.i", 
             msg.src_name, msg.src_name, msg.src_name);
   else if ((strcmp(ctypes,"lci")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.cpl %s.lsp %s.d %s.i",
             msg.src_name, msg.src_name, msg.src_name);
   else if ((strcmp(ctypes,"loi")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.obj %s.o %s.d %s.i",
             msg.src_name, msg.src_name, msg.src_name);
   else if  ((strcmp(ctypes,"lc")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.cpl %s.lsp %s.d",
             msg.src_name, msg.src_name);
   else if  ((strcmp(ctypes,"lo")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.obj %s.lsp %s.d",
             msg.src_name, msg.src_name);
   else if ( (*ctypes == 'l') && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.int %s.lsp %s.d",
                   msg.src_name, msg.src_name);
   /*                                                               */
   /* Set default execute string if using the Mathematica interface */
   /*                                                               */
   else if ( (*ctypes == 'm') && (strlen(exec)==0))
           sprintf(msg.src_exec, "Math %s.m %s.d",
                   msg.src_name, msg.src_name);
   else if ((strcmp(ctypes,"mi")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "Math %s.m %s.d %s.i",
                   msg.src_name, msg.src_name, msg.src_name);
   /*                                                               */
   /* Set default execute string if using the Maple interface       */
   /*                                                               */
   else if ( (*ctypes == 'p') && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_maple.int %s.mpl %s.d",
                   msg.src_name, msg.src_name);
   else if ( (strcmp(ctypes,"pi")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_maple.int %s.mpl %s.d %s.i",
             msg.src_name, msg.src_name, msg.src_name);
   /*                                                               */
   /* Set default execute string if using the C interface           */
   /*                                                               */
   else if ( strlen(exec)==0 ) 
           sprintf( msg.src_exec, "%s %s.d", msg.src_name, msg.src_name );
   /*                                                               */
   /* Use user supplied execute string                              */
   /*                                                               */
   else if (strlen( exec ))   strcpy( msg.src_exec, exec);

#ifdef DPRCDISDEBUG
cerr << "dscpr    : ctrl_dispatch    : msg.src_exec " << msg.src_exec << "\n";
cerr << "dscpr    : ctrl_dispatch    : sending message to DSCRP\n";
cerr << flush;
#endif

  /*                              */
  /* send message to server DSCRP */
  /*                              */
   do {
      errno = 0;
      rlen = sendto( sockdg, (char *)&msg, sizeof(DSCSTRT),
		 0, (struct sockaddr *) &saddr, sizeof(saddr));
      if (errno == ENOBUFS) sleep(2);
   } while( (errno==EINTR) || (errno==ENOBUFS));

   if (errno != 0) {
      cerr << "dscpr    : ctrl_dispatch    : error sending start msg\n";
      cerr << "dscpr    : ctrl_dispatch    : forced problem halt!\n";
      perror("sending start msg to DSC server");
      exit(errno);
   }

   if (level == 0) {
      cout << "\n     file: " << msg.src_name << "\n";
      cout << "build cmd: " << msg.src_build << "\n";
      cout << " exec cmd: " << msg.src_exec << "\n";
      if (msg.cpurating == HIGH_CPU)  cout << "  est cpu: HIGH_CPU\n";
      else
      if (msg.cpurating == MED_CPU)  cout << "  est cpu: MED_CPU\n";
      else
      if (msg.cpurating == LOW_CPU)  cout << "  est cpu: LOW_CPU\n";
      else cout << "  est cpu: " << msg.cpurating << " (UNKOWN VALUE)\n";
      cout << "  est mem: " << msg.memrating << " megs\n";
      msg2->index = htonl(0);
    } /* end if main problem */
   else if (level > 0) {

#ifdef DPRCDISDEBUG
cerr << "dscpr    : ctrl_dispatch    : getting index\n";
#endif

      sock = accept( sockl, 0, 0);

if ( sock == -1 )
{
cerr << "dscpr    : ctrl_dispatch    : can not get descriptor " << sock << "\n";
cerr << "dscpr    : ctrl_dispatch    : errno " << errno << "\n";
cerr << "dscpr    : ctrl_dispatch    : fatal error!\n";
exit(errno);
}

      on = 1;
      setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));

     /*           */
     /* get index */
     /*           */
      do {
         errno = 0;
         rlen = dsc_read (sock, (DSCKIN_PTR) msg2, sizeof (DSCBEG));
      } while (errno == EINTR );
      if (rlen != ntohl(msg2->msg_size)) {
        perror("reading problem index");
    cerr  << "dscpr    : ctrl_dispatch    : mismatch read DSCBEG\n";
    cerr  << "dscpr    : ctrl_dispatch    : rlen = " << rlen << " msgsize = " 
	  << ntohl(msg2->msg_size) << "\n";
        cout << flush;
        cerr << flush;
      }
      close( sock );
      sock = -1;
    }/* end if level > 0 (subproblems only ) */

#ifdef DPRCDISDEBUG
cerr << "dscpr    : ctrl_dispatch    : returning index " << ntohl(msg2->index)
     << "\n";
#endif

   return( ntohl(msg2->index) );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  dscpr_wait( index )                                          */
/*                                                                           */
/* inputs: index - index of subproblem to wait on (-1 == wait on all)        */
/*                                                                           */
/* ext vars: passinit - password set flag                                    */
/*           portinit - port set flag                                        */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   status                                                          */
/*                                                                           */
/* purpose:  send a message to DSC RP telling it we want to                  */
/*           know when all subprocesses finish.                              */
/*           block on answer from DSC RP, return success or failure.         */
/*                                                                           */
/*****************************************************************************/
DSCSOLV msgsolv;

int dscpr_wait (int index)
{

   DSCPRBLK msg;         /* block problem message                            */
   DSCSOLV *msg2;        /* incomming unblocking message                     */
   int      rlen;        /* length of read message                           */
   int     rstat;        /* status of wait                                   */
   int     on;
   char *dsc_getpass(void);  /* get password from dsc_setup.csh file             */
   char *dscpr_getnum(void); /* get port numeber from dsc_setup.csh file         */


#ifdef DPRWAITDEBUG
cerr  << "dscpr    : dscpr_wait       : index " << index << "\n";
#endif

   if ( portinit == FALSE )
   {
    /*                                                                    */
    /* get the datagram communication port number from the dsc setup file */
    /*                                                                    */
    cportnm = dscpr_getnum();
    if (strcmp(cportnm,"NOT_FOUND") == 0) {
       cerr << "dscpr    : dscpr_sub        : port number not found\n";
       cerr << "dscpr    : dscpr_sub        : forced problem halt!\n";
       rstat = dscpr_end();
       exit(rstat);
    }
   /*                            */
   /* convert it to decimal form */
   /*                            */
   portnm = atoi(cportnm);
   }

   /*                              */
   /* set up communication sockets */
   /*                              */
   dscpr_netinit();

   if ( passinit == FALSE )
   {
    password = dsc_getpass();
   }

   msg.msg_type = msg.sanity = htonl(MSG_PRBLOCK);
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl(sizeof( DSCPRBLK ));
   memcpy( &msg.addr, &paddr, sizeof(paddr));
   msg.from_PID = htonl(mypid);
   msg.index = htonl(index);

#ifdef DPRWAITDEBUG
cerr  << "dscpr    : dscpr_wait       : send message to DSCRP\n";
#endif

  /*                              */
  /* send message to server DSCRP */
  /*                              */
   do {
      errno = 0;
      rlen = sendto( sockdg, (char *)&msg, sizeof(DSCPRBLK),
		 0, (struct sockaddr *) &saddr, sizeof(saddr));
      if (errno == ENOBUFS) sleep(2);
   } while( (errno==EINTR) || (errno==ENOBUFS));

   if (errno != 0) {
cerr  << "dscpr    : dscpr_wait       : error sending MSG_PRBLOCK\n";
cerr  << "dscpr    : dscpr_wait       : forced problem halt!\n";
      perror("sending start msg to DSC server");
      exit(errno);
   }

   sock = accept( sockl, 0, 0);
   on = 1;
   setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));

  /*                                      */
  /* wait for all subproblems to complete */
  /*                                      */
   msg2 = &msgsolv;
   do {
      errno = 0;
      rlen = dsc_read (sock, (DSCKIN_PTR) msg2, sizeof( DSCSOLV));
   } while (errno == EINTR );

   if (rlen != ntohl(msg2->msg_size)) {
    perror("reading problem index");
    cerr  << "dscpr    : dscpr_wait       : mismatch read DSCBEG\n";
    int avoid_ANSI_warning = (int) ntohl(msg2->msg_size);
    cerr  << "dscpr    : dscpr_wait       : rlen = " << rlen << " msgsize = "
          << avoid_ANSI_warning << "\n";    // (AAP)
     cout << flush;
     cerr << flush;
   }

   close( sock );

   if (ntohl(msg2->msg_type) != MSG_SOLVE) rstat = DSC_FAILURE;
   else rstat = (int) ntohl(msg2->status);
 
#ifdef DPRWAITDEBUG
cerr  << "dscpr    : dscpr_wait       : returning rstat " << rstat << "\n";
#endif

   return( rstat );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  dscpr_next( ans )                                            */
/*                                                                           */
/* inputs: none                                                              */
/*                                                                           */
/* return:   status                                                          */
/*                                                                           */
/* outputs:   ans - name of solution data file                               */
/*                                                                           */
/* purpose:  special form of dscpr_wait()                                    */
/*                                                                           */
/*****************************************************************************/
//
//int dscpr_next (char *ans)
//{
//   int rstat;          /* status of next                                   */
//
//#ifdef DPRNEXTDEBUG
//cerr  << "dscpr    : dscpr_next       : char *ans " << ans << "\n";
//#endif
//
//   rstat = dscpr_wait(-2);
//   if (ans != NULL) sprintf( ans, "%s.d", msgsolv.src_name);
//    else
//     cerr  << "dscpr    : dscpr_next       : no answer file given\n";
//
//#ifdef DPRNEXTDEBUG
//cerr  << "dscpr    : dscpr_next       : returning rstat " << rstat << "\n";
//#endif
//
//   return (rstat);
//}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  dscpr_next (char *ans)                                       */
/*                                                                           */
/* inputs: none                                                              */
/*                                                                           */
/* return:   status                                                          */
/*                                                                           */
/* outputs:   ans - name of solution data file                               */
/*                                                                           */
/* purpose:  special form of dscpr_wait()                                    */
/*                                                                           */
/*****************************************************************************/

//int dscpr_next (char *ans)
int dscpr_next (void)
{
   int rstat;
   int nomore;
   int temp;
   int mask;
   rstat = dscpr_wait(-2); 
   sleep(2);
   nomore = -40000;
   temp = 0;
   mask = (1<<6)|(1<<5)|(1<<7)|(1<<8)|(1<<9)|(1<<13)|(1<<29);
#ifdef DPRNEXTDEBUG
cerr << "dscpr:   dscpr_next():   RSTAT " << rstat << endl;
cerr << "dscpr:   dscpr_next():   MASK " << mask << endl;
cerr << "dscpr:   dscpr_next():   rstat && (!mask) " << (rstat && (!mask)) 
     << endl;
cerr << "dscpr:   dscpr_next():   rstat && mask " << (rstat && mask)  << endl;
#endif
 /*                           */
 /* A problem ended correctly */
 /* (index 1..10000)          */
 /*                           */
   if ( (rstat == 0 ) && (msgsolv.index != -1) )
    {
        cerr << "dscpr:   dscpr_next():   Problem ended correctly" << endl;
        cerr << "dscpr:   dscpr_next():   Msgsolv.src_name is " << msgsolv.src_name << endl;
        cerr << "dscpr:   dscpr_next():   Index " << msgsolv.index << endl; 
        return (msgsolv.index);
    }
 /*                            */
 /* No more active subproblems */
 /* (index -40000)             */
 /*                            */
   if (msgsolv.index == -1 )
    {
        cerr << "dscpr:   dscpr_next():   No more active subproblems." << endl;
        cerr << "dscpr:   dscpr_next():   Index " << msgsolv.index << endl; 
        return(nomore);
    }
 /*                                                 */
 /* A problem ended with a correctable error status */
 /* (index -10001..-20000)                          */
 /*                                                 */

   if (  (rstat && (!mask)) && (rstat && mask) )
    {
        cerr << "dscpr:   dscpr_next():   Subproblem ended with a correctable error status." << endl;
        temp = (int) (-1 * msgsolv.index);
        temp = temp + (- 10000) ;
        cerr << "dscpr:   dscpr_next():   Msgsolv.src_name is " << msgsolv.src_name << endl;
        cerr << "dscpr:   dscpr_next():   Index " << msgsolv.index << endl; 
        return(temp);
    }

 /*                                                    */
 /* A problem ended with an uncorrectable error status */
 /* (index -1..-10000)                                 */
 /*                                                    */

   if ( (rstat !=0) && (msgsolv.index != -1) )
    {
        cerr << "dscpr:   dscpr_next():   Subproblem ended with an uncorrectable error status." << endl;
        temp = (int) (-1 * msgsolv.index);
        cerr << "dscpr:   dscpr_next():   Msgsolv.src_name is " << msgsolv.src_name << endl;
        cerr << "dscpr:   dscpr_next():   Index " << msgsolv.index << endl; 
        return(temp);
    }
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_kill( index )                                           */
/*                                                                           */
/* inputs: index - index of subproblem to kill                               */
/*                                                                           */
/* ext vars: passinit - password set flag                                    */
/*           portinit - port set flag                                        */
/*                                                                           */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   status                                                          */
/*                                                                           */
/* purpose:  send a message to DSC RP requesting a specified subproblem      */
/*           be terminated.                                                  */
/*                                                                           */
/*****************************************************************************/ 

int dscpr_kill (long index)
{

   DSCKILL msg;          /* DSC RP kill message                              */
   char *dsc_getpass(void);  /* get password from dsc_setup.csh file             */
   char *dscpr_getnum(void); /* get port numeber from dsc_setup.csh file         */
   int rstat;            /* status of kill                                   */

#ifdef DPRKILLDEBUG
cerr  << "dscpr    : dscpr_kill       : rindex " << index << "\n";
#endif

   if ( portinit == FALSE )
   {
    /*                                                                    */
    /* get the datagram communication port number from the dsc setup file */
    /*                                                                    */
    cportnm = dscpr_getnum();
    if (strcmp(cportnm,"NOT_FOUND") == 0) {
       cerr << "dscpr    : dscpr_kill       : port number not found\n";
       cerr << "dscpr    : dscpr_kill       : forced problem halt!\n";
       rstat = dscpr_end();
       exit(rstat);
    }
   /*                            */
   /* convert it to decimal form */
   /*                            */
   portnm = atoi(cportnm);
   }

   /*                              */
   /* set up communication sockets */
   /*                              */
   dscpr_netinit();

   if ( passinit == FALSE )
   {
    password = dsc_getpass();
   }

   msg.msg_type = msg.sanity = htonl(MSG_KILL);
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl(sizeof( DSCKILL ));
   msg.from_PID = htonl(mypid);
   msg.index = htonl(index);

  /*                              */
  /* send message to server DSCRP */
  /*                              */
   do {

      errno =0;

      sendto( sockdg, (char *)&msg, sizeof(DSCKILL),
		 0, (struct sockaddr *) &saddr, sizeof(saddr));
      if (errno == ENOBUFS) sleep(2);
   } while( (errno==EINTR) || (errno==ENOBUFS));
   if (errno != 0) {
cerr  << "dscpr    : dscpr_kill       : error sending DSCKILL\n";
cerr  << "dscpr    : dscpr_kill       : forced problem halt!\n";
      perror("sending start msg to DSC server");
      exit(errno);
   }

#ifdef DPRKILLDEBUG
cerr  << "dscpr    : dscpr_kill       : returning errno " << errno << "\n";
#endif

   return( errno );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_end( )                                                  */
/*                                                                           */
/* inputs: none                                                              */
/*                                                                           */
/* ext vars: sockl  - stream socket                                          */
/*           sockdg - datagram socket                                        */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   0                                                               */
/*                                                                           */
/* purpose:  close socket and unlink the associated file                     */
/*                                                                           */
/*****************************************************************************/

int dscpr_end (void)
{

#ifdef DPRENDDEBUG
cerr  << "dscpr    : dscpr_end        : void\n";
#endif

   if (sockl != (-1)) close( sockl );
   if (sockdg != (-1)) close( sockdg );

#ifdef DPRENDDEBUG
cerr  << "dscpr    : dscpr_end        : returning 0\n";
#endif

   return( 0 );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscdbg_start( )                                               */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: passinit - password set flag                                    */
/*           portinit - port set flag                                        */
/*           sock     - datagram socket                                      */
/*                                                                           */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   index to new subproblem in DSC Server database                  */
/*                                                                           */
/* purpose:  build message to create new subproblem                          */
/*           send the message to the DSC RP                                  */
/*                                                                           */
/*****************************************************************************/

int dscdbg_start (void)
{
   DSCSTRT msg;          /* start subproblem message                         */
   DSCBEG *msg2;         /* incomming startup message                        */
   int     rlen;         /* legnth of message                                */
   int     rstat;        /* status of dscdbg call                            */
   int on;
   char *dsc_getpass(void);  /* get password from dsc_setup.csh file             */
   char *dscpr_getnum(void); /* get port numeber from dsc_setup.csh file         */  

#ifdef DPRDBGDEBUG
cerr << "dscpr    : dscdbg_start      : void\n";
#endif

   if ( portinit == FALSE )
   {
    /*                                                                    */
    /* get the datagram communication port number from the dsc setup file */
    /*                                                                    */
#ifdef DPRDBGDEBUG
cerr << "dscpr    : dscdbg_start      : getting port number\n";
#endif
    cportnm = dscpr_getnum();
    if (strcmp(cportnm,"NOT_FOUND") == 0) {
       cerr << "dscpr    : dscdbg_start      : port number not found\n";
       cerr << "dscpr    : dscdbg_start      : forced problem halt!\n";
       rstat = dscpr_end();
       exit(rstat);
    }
   /*                            */
   /* convert it to decimal form */
   /*                            */
   portnm = atoi(cportnm);
   }

   /*                              */
   /* set up communication sockets */
   /*                              */
   dscpr_netinit();

   if ( passinit == FALSE )
   {
#ifdef DPRDBGDEBUG
cerr << "dscpr    : dscdbg_start      : getting password\n";
#endif
    password = dsc_getpass();
   }


   msg.msg_type = htonl(MSG_START);
   strcpy(msg.dsc_password , password );
   msg.sanity = htonl(mypid);
   msg.msg_size = htonl(sizeof( DSCSTRT ));
   memcpy( &msg.addr, &paddr, sizeof(paddr));

   msg.parent_PID = htonl(-1);
   msg.fnode = htonl(myhost);
   msg.subflg = htonl(1);
   sprintf( msg.prob_name, "DSC%4.4x_%d", mypid&0xffff, ++subcnt);
   msg.ctyp = 1;
   msg.cpu_type = htonl(DSC_CPU);
   msg.cpurating = htonl(-1);
   msg.memrating = htonl(-1);

   strcpy( msg.src_name, "dsc_prob_debug" );
   strcpy( msg.src_build, "");
   strcpy( msg.src_exec, "");

  /*                              */
  /* send message to server DSCRP */
  /*                              */
   do {
      errno = 0;
      rlen = sendto( sockdg, (char *)&msg, sizeof(DSCSTRT),
		 0, (struct sockaddr *) &saddr, sizeof(saddr));
      if (errno == ENOBUFS) sleep(2);
   } while( (errno==EINTR) || (errno==ENOBUFS));
   if (errno != 0) {
cerr  << "dscpr    : dscdbg_start     : error sending DSCSTRT\n";
cerr  << "dscpr    : dscdbg_start     : forced problem halt!\n";
      perror("sending start msg to DSC server");
      exit(errno);
   }

   do {
     errno = 0;
     sock = accept( sockl, 0, 0);
   } while (errno == EINTR);
   on = 1;
   setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));

  /*           */
  /* get index */
  /*           */
   msg2 = (DSCBEG *)&msg;
   do {
      errno = 0;
      rlen = dsc_read (sock, (DSCKIN_PTR) msg2, sizeof (DSCBEG));
   } while (errno == EINTR );
   if (rlen != ntohl(msg2->msg_size)) {
     perror("reading problem index");
    cerr  << "dscpr    : dscdbg_start     : mismatch read DSCBEG\n";
    int avoid_ANSI_warning = (int) ntohl(msg2->msg_size);
    cerr  << "dscpr    : dscdbg_star      : rlen = " << rlen << " msgsize = "
          << avoid_ANSI_warning << "\n";  // (AAP)
     cout << flush;
     cerr << flush;
   }

   close( sock );

#ifdef DPRDBGDEBUG
cerr << "dscpr    : dscdbg_start      : returning (" << ntohl (msg2->index) 
     << ")\n";
#endif

   return( ntohl(msg2->index) );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_cont( )                                                   */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: dsc_cowait_flag - flag for sleep wait loop                      */ 
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose: when the user defined signal SIGUSR1 is recieved the subproblem  */
/*          is to resume execution after the dscpr_cowait call. This is      */
/*          accomplished by setting the dsc_cowait_flag to TRUE thus         */
/*          allowing the  wait loop in dscpr_cowait function to terminate.   */
/*          The function dsc_cont sets the dsc_cowait_flag to TRUE so that   */
/*          the subproblem can contine execution.                            */
/*                                                                           */
/*****************************************************************************/

void dsc_cont (void)
{
#ifdef DPRCONTDEBUG
cerr  << "dscpr    : dsc_cont         : void\n";
#endif

dsc_cowait_flag = TRUE;

#ifdef DPRCONTDEBUG
cerr  << "dscpr    : dsc_cont         : returning void\n";
#endif
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_cosetup( )                                              */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: dsc_cowait_flag - flag for sleep wait loop                      */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  the function dscpr_cosetup must be called at the begining       */
/*           of any  subproblem using co-routines. This is necessary so      */
/*           that the signal SIGUSR1 can be interpreted as a command to      */
/*           resume execution of the subproblem.                             */
/*                                                                           */
/*****************************************************************************/

void dscpr_cosetup (void)
{
 void dsc_cont(void);

#ifdef DPRCOSETDEBUG
cerr  << "dscpr    : dscpr_cosetup    : void\n";
#endif

 dsc_cowait_flag = FALSE;

 signal (SIGUSR1, (SIG_PF) dsc_cont);

#ifdef DPRCOSETDEBUG
cerr  << "dscpr    : dscpr_cosetup    : returning void\n";
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_cowait( flag )                                          */
/*                                                                           */
/* inputs:   flag - 1  copy data file                                        */
/*                  0  do not copy data file                                 */
/*                                                                           */
/* ext vars: dsc_cowait_flag - flag for sleep wait loop                      */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  the function dscpr_cowait sends a datagram to the server        */
/*           informing it that it is entering a sleep state. Then it enters  */
/*           a sleep loop waiting to be awakened by the signal SIGUSR1       */
/*           signaled by the server.                                         */
/*                                                                           */
/*****************************************************************************/ 

void dscpr_cowait (int flag)
{
 DSCSLP msg;                  /* DSCSLP message                              */
   int rlen;                  /* read in message length                      */
   int rstat;                 /* status of dscpr_cowait                      */
   struct hostent *hp;        /* host information                            */
   char localnode[DSC_NODEZ]; /* name of local node                          */
   char *dsc_getpass(void);       /* get password from dsc_setup.csh file        */
   char *dscpr_getnum(void);      /* get port numeber from dsc_setup.csh file    */

#ifdef DPRCOWDEBUG
cerr << "dscpr    : dscpr_cowait      : flag " << flag << "\n";
#endif

   if ( portinit == FALSE )
   {
    /*                                                                    */
    /* get the datagram communication port number from the dsc setup file */
    /*                                                                    */
#ifdef DPRCOWDEBUG
cerr << "dscpr    : dscpr_cowait      : getting port number\n";
#endif
    cportnm = dscpr_getnum();
    if (strcmp(cportnm,"NOT_FOUND") == 0) {
       cerr << "dscpr    : dscpr_cowait      : port number not found\n";
       cerr << "dscpr    : dscpr_cowait      : forced problem halt!\n";
       rstat = dscpr_end();
       exit(rstat);
    }
   /*                            */
   /* convert it to decimal form */
   /*                            */
   portnm = atoi(cportnm);
   }

   /*                              */
   /* set up communication sockets */
   /*                              */
   dscpr_netinit();

   if ( passinit == FALSE )
   {
#ifdef DPRCOWDEBUG
cerr << "dscpr    : dscpr_cowait      : getting password\n";
#endif
    password = dsc_getpass();
   }

 msg.msg_type = msg.sanity = htonl(MSG_SLEEPING);
 strcpy(msg.dsc_password, password );
 msg.msg_size = htonl(sizeof( DSCSLP ) );
 msg.PID = htonl(getpid());

 gethostname(localnode, DSC_NODEZ);
 hp = gethostbyname( localnode );
 msg.fnode = (long)(*(long *)hp->h_addr);
 msg.copyflag = flag;
 strcpy(msg.fnname,localnode);
 strcpy(msg.prob_name,""); 

 /*                              */
 /* Send message to server DSCRP */
 /*                              */
 do {
    errno = 0;
    rlen = sendto( sockdg, (char *)&msg, sizeof(DSCSLP),
               0, (struct sockaddr *) &saddr, sizeof(saddr));
    if (errno == ENOBUFS) sleep(2);
    } while ( ( errno == EINTR ) || (errno == ENOBUFS ) ) ;  
 if ( errno != 0 ) {
   perror("sending sleep message to DSC server");
cerr  << "dscpr    : dscpr_cowait     : error sending DSCSLP\n";
cerr  << "dscpr    : dscpr_cowait     : forced problem halt!\n";
   exit(errno);
 }

#ifdef DPRCOWDEBUG
cerr << "dscpr    : dscpr_cowait      : entering idle state\n";
#endif

 while ( !(dsc_cowait_flag) ) ;
 dsc_cowait_flag = FALSE;

#ifdef DPRCOWDEBUG
cerr << "dscpr    : dscpr_cowait      : returning void\n";
#endif
 
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_coresume( int index , char *filname)                    */
/*                                                                           */
/* inputs:    index    - subproblem index                                    */
/*            filename - "" copy the regular input file                      */
/*                       "NULL" do not copy any file                         */
/*                        "filname" vopy the filename                        */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   subproblem index on success, -1 on failure                      */
/*                                                                           */
/* purpose:  the function dscpr_coresume sends a datagram to the server      */
/*           informing it that it is ready to wake up a subproblem           */
/*           specified by the index. It hangs until a a success or failure   */
/*           response that is transmitted via the stream socket.             */
/*                                                                           */
/*****************************************************************************/

int dscpr_coresume (int index , char *filename)
{
   DSCAWK msg;                /* dsc awaken message                          */
   DSCBEG *msg2;              /* incomming response message                  */
   int rlen;                  /* read in message length                      */
   int on;                     
   int rstat;                 /* status of dscpr_coresume                    */
   struct hostent *hp;        /* host information                            */
   char localnode[DSC_NODEZ]; /* name of local node                          */
   char *dsc_getpass(void);       /* get password from dsc_setup.csh file        */
   char *dscpr_getnum(void);      /* get port numeber from dsc_setup.csh file    */



#ifdef DPRCORDEBUG
cerr << "dscpr    : dscpr_coresume    : index " << index << "\n";
cerr << "dscpr    : dscpr_coresume    : filename " << filename << "\n";
#endif

   if ( portinit == FALSE )
   {
    /*                                                                    */
    /* get the datagram communication port number from the dsc setup file */
    /*                                                                    */
#ifdef DPRCORDEBUG
cerr << "dscpr    : dscpr_cowait      : getting port number\n";
#endif
    cportnm = dscpr_getnum();
    if (strcmp(cportnm,"NOT_FOUND") == 0) {
       cerr << "dscpr    : dscpr_coresume    : port number not found\n";
       cerr << "dscpr    : dscpr_coresume    : forced problem halt!\n";
       rstat = dscpr_end();
       exit(rstat);
    }
   /*                            */
   /* convert it to decimal form */
   /*                            */
   portnm = atoi(cportnm);
   }

   /*                              */
   /* set up communication sockets */
   /*                              */
   dscpr_netinit();

   if ( passinit == FALSE )
   {
#ifdef DPRCORDEBUG
cerr << "dscpr    : dscpr_coresume    : getting password\n";
#endif
    password = dsc_getpass();
   }


 msg2 = (DSCBEG *)&msg;
 msg.msg_type = msg.sanity = htonl(MSG_AWAKEN);
 strcpy(msg.dsc_password, password );
 msg.msg_size = htonl(sizeof( DSCAWK ) );
 msg.PID = htonl(getpid());
 msg.index = htonl(index);

 gethostname(localnode, DSC_NODEZ);
 hp = gethostbyname( localnode );
 msg.fnode = (long)(*(long *)hp->h_addr);
 strcpy(msg.fnname,localnode);
 strcpy(msg.prob_name,""); 
 strcpy(msg.filename,filename);
 memcpy( &msg.addr, &paddr, sizeof(paddr));

 /*                              */
 /* Send message to server DSCRP */
 /*                              */
 do {
    errno = 0;
    rlen = sendto( sockdg, (char *)&msg, sizeof(DSCAWK),
               0, (struct sockaddr *) &saddr, sizeof(saddr));
    if (errno == ENOBUFS) sleep(2);
    } while ( ( errno == EINTR ) || (errno == ENOBUFS ) ) ;  
 if ( errno != 0 ) {
   perror("sending awaken message to DSC server");
cerr  << "dscpr    : dscpr_coresume   : error sending DSCAWK\n";
cerr  << "dscpr    : dscpr_coresume   : forced problem halt!\n";
   exit(errno);
 }

 sock = accept(sockl,0,0);
 on = 1;
 setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));

 /*           */
 /* get index */
 /*           */
 rlen = read (sock, (char *) msg2, sizeof (DSCBEG));
 if (rlen != ntohl(msg2->msg_size)) {
  perror("reading resume problem index");
    cerr  << "dscpr    : dscpr_coresume   : mismatch read DSCBEG\n";
    int avoid_ANSI_warning = (int) ntohl(msg2->msg_size);
    cerr  << "dscpr    : dscpr_coresume   : rlen = " << rlen << " msgsize = "
          << avoid_ANSI_warning << "\n";  // (AAP)
     cout << flush;
     cerr << flush;
 }

 close(sock);
 sock = -1;

#ifdef DPRCORDEBUG
cerr << "dscpr    : dscpr_coresume    : returning index " << ntohl (msg2->index)
     << "\n";
#endif

 return (ntohl(msg2->index));
 
} 
