/*****************************************************************************/
/*                                                                           */
/*                                dscpr.c                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* functions :                                                               */
/*   These functions are imbedded into a problem they allow access to DSC    */
/*  for starting subproblems and waiting for answers.                        */
/*                                                                           */
/*     dsc_dispatch()     - to send start problem msg                        */
/*     dscpr_wait()       - wait for one or more subproblems to complete     */
/*                                                                           */
/*   These functions are used for axulliary support                          */
/*                                                                           */
/*     dscpr_reverse()        - reverse a string in place                    */
/*     dscpr_isnum()          - determine if a string consists of digits     */
/*     dscpr_itoa()           - convert a number to characters               */
/*     dsc_read()             - read designed to handle multipart messages   */
/*     dsc_getpass_sub()      -                                              */
/*     dsc_getpass()          -                                              */
/*     dscpr_getnum_sub()     -                                              */
/*     dscpr_getnum()         - get portnumber from the dsc_setup.csh file   */
/*     dscpr_netinit()        - set up local datagram and stream sockets     */
/*     dscpr_netpnm()                                                        */
/*     dscpr_ctrl_portnmchk() - choose a unused port number                  */
/*     dscpr_ctrl_portnm()    - check if a port number is needed             */
/*                                                                           */
/*****************************************************************************/

#include <ctype.h>
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <errno.h> 
#include <sys/types.h> 
#include <sys/signal.h> 
#include <sys/socket.h> 
#include <sys/time.h> 
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>    /* used in fcnlt() function */
#include "dsc_defs.h"

#include <sys/syscall.h>
#include <sys/param.h>


static struct sockaddr_in saddr;   /* server datagram address                */
static struct sockaddr_in paddr;   /* this problem stream address            */
static struct sockaddr_in lisaddr; /* non blocking stream address            */
static int    portnm;              /* integer value of datagram port number  */
static char   *cportnm;            /* char value of datagram port number     */
static char   *password=NULL;      /* DSC_PORTNUM text variable              */
static int    mypid;               /* this process's PID                     */
static char   mynode[DSC_NODEZ];   /* local node name                        */
static long   myhost;              /* local host address                     */
static int    netinit=FALSE;       /* sockets initialized flag               */
static int    passinit=FALSE;      /* password set flag                      */
static int    portinit=FALSE;      /* port number set flag                   */
static int    overide=FALSE;       /* portnumber search overide flag         */
static int    sockdg=(-1);         /* udp datagram socket                    */
static int    sockl=(-1);          /* tcp listen socket                      */
static int    lissock=(-1);        /* non blocking stream socket             */
static int    sock=(-1);           /* tcp comm. socket                       */
static int    subcnt=0;            /*                                        */
static int    dsc_cowait_flag = FALSE; /* flag for the sleep state wait loop */
static char   pwd[100];            /* expanded path information              */



/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_reverse(char *s)                                        */
/*                                                                           */
/* inputs:   s - terminated character string                                 */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   reversed string                                                 */
/*                                                                           */
/* purpose:  reverse string s in place                                       */
/*                                                                           */
/*****************************************************************************/

void dscpr_reverse( s )
char *s;
{
   int c, i, j;

   for (i = 0, j = strlen(s)-1; i < j; i++, j--) {
     c=s[i];
     s[i]=s[j];
     s[j]=c;
   }
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_isnum ( chkword )                                       */
/*                                                                           */
/* inputs:   chkword - terminated character string                           */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   TRUE  = only digits in string                                   */
/*           FLASE = non digits in string                                    */ 
/*                                                                           */
/* purpose:  check string for non digits                                     */
/*                                                                           */
/*****************************************************************************/

int dscpr_isnum ( chkword )
char *chkword;
{

 while ( *chkword != '\0' )
 {
   if (!( ( ((int) *chkword) >=48 ) && ( ((int) *chkword) <=57 ) )) 
    return(FALSE);
   chkword++;
 }

 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_itoa( n, s)                                             */
/*                                                                           */
/* inputs:   n - integer value to be converted                               */
/*           s - terminated character string                                 */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  compute the string equivalent of n                              */
/*                                                                           */
/*****************************************************************************/

void dscpr_itoa( n, s)
int n;
char *s;
{
  int i, sign;

  /*             */
  /* record sign */
  /*             */ 
  if ((sign = n) < 0 ) 
    n = -n;           
  i = 0;
  /*                                  */
  /* generate digits in reverse order */
  /*                                  */
  do { 
      s[i++] = n % 10 + '0'; 
  } while ((n /= 10) > 0 ); 
  if (sign < 0)
     s[i++] = '-';
  s[i] = '\0';

  dscpr_reverse(s);

 } 

/*****************************************************************************/
/*                                                                           */
/* subroutine: int dsc_read( d, buf, nbytes)                                 */
/*                                                                           */
/* inputs:   d      - descriptor                                             */
/*           buf    - message buffer                                         */
/*           nbytes - number of bytes to read                                */
/*                                                                           */
/* outputs:  (receive a file)                                                */
/*                                                                           */
/* return:   length of message to read                                       */
/*                                                                           */
/* purpose: this function will piece together a whole message when it        */
/*          arrives in more than one message. ( this has been observed to    */
/*          happen sometimes with messages over 1024  bytes in length,       */
/*          It may happen other times too....)                               */
/*                                                                           */
/*****************************************************************************/

int dsc_read( d, buf, nbytes)
int d; 
DSCKIN_PTR buf;
int  nbytes;
{
  int cc;               /* number of bytes actually  read on success         */
  int len;              /* size of message                                   */
  int remaining;        /* number of remaining bytes to be read              */ 
  char *xbuf;           /* pointer to message buffer                         */

#ifdef DPRRDDEBUG
fprintf(stderr,"dscpr    : dsc_read         : int d %d\n",d);
fprintf(stderr,"dscpr    : dsc_read         : DSCKIN_PTR buf\n");
fprintf(stderr,"dscpr    : dsc_read         : int nbytes %d\n",nbytes);
#endif

  cc = read( d, buf, nbytes );
  if (cc == 0) return(0);         
  else if (cc > 8) len = ntohl(buf->msg_size); 
  else {
fprintf(stderr,"dscpr    : dsc_read         : less than 8 bytes received\n");
fprintf(stderr,"dscpr    : dsc_read         : %d bytes recieved\n",cc);
fprintf(stderr,"dscpr    : dsc_read         : errno %d\n",errno); 
fprintf(stderr,"dscpr    : dsc_read         : fatal error !!!\n");
    exit(errno);
    } 


  xbuf = (char *) buf;
  remaining = len - cc;
  while (remaining > 0) {

fprintf(stderr,"dscpr    : dsc_read         : another read required\n");
fprintf(stderr,"dscpr    : dsc_read         : for %d bytes\n", remaining);

    /*                                    */
    /* calc pointer for begging next read */
    /*                                    */
     xbuf += cc;

    /*                      */
    /* read rest of message */
    /*                      */
     cc = read( d, xbuf, remaining ); 

    /*                                                        */
    /* determine if we are done yet, done when remaining == 0 */
    /*                                                        */ 
     remaining -= cc;
  }

#ifdef DPRRDDEBUG
fprintf(stderr,"dscpr    : dsc_read         : returning (%d)\n",len);
#endif

  return( len );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpass_sub( file )                                       */
/*                                                                           */
/* inputs:   file - pointer to memory containing the "dsc_setup.csh" file    */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PASSWORD = if found                                        */
/*            "NOT FOUND"  = if not found                                    */
/*                                                                           */
/* purpose: loop thru memory in search of the DSC_PASSWORD                   */
/*                                                                           */
/*****************************************************************************/

char* dsc_getpass_sub( file )
char* file;
{
   char *word;       /* word in the "dsc_setup.csh" file                     */
   char *wordbeg;    /* pointer to first character                           */
   int found;        /* flag                                                 */

#ifdef DPRGPSDEBUG
fprintf(stderr,"dscpr    : dsc_getpass_sub  : char* file\n");
#endif 

   found = FALSE;
   word = (char *)malloc(DSC_BUFFZ);
   wordbeg = word;

   /*                                                           */
   /* loop through file searching for the DSC_PASSWORD variable */
   /*                                                           */

  while ( *file != EOF ) {

   /*                     */
   /* skip leading blanks */
   /*                     */
   while ( *file == ' ' ) file ++;

   /*                                          */
   /* get word , a word is delimited by spaces */
   /*                                          */
   while  ( ( *file != ' ') && ( *file != EOF ) )
    { *word = *file;
      word ++; file++;
    }
   *word = '\0';

   /*                                                         */
   /* the pass word is the word after the string DSC_PASSWORD */
   /*                                                         */
   if ( found == TRUE )
     {
#ifdef DPRGPSDEBUG
fprintf(stderr,"dscpr    : dsc_getpass_sub  : found password %s\n",wordbeg);
#endif
      return(wordbeg);
     }

   if (strcmp(wordbeg,DSCTEXTPASS) == 0 )  found = TRUE;

   word = wordbeg;

  }

#ifdef DPRGPSDEBUG
fprintf(stderr,"dscpr    : dsc_getpass_sub  : could not find password\n");
#endif

 return("NOT_FOUND");
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpass()                                                 */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: passinit - to flag password already found                       */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   DSC_PASSWORD on success                                         */
/*           halt otherwise                                                  */
/*                                                                           */
/* purpose:  search the "dsc_setup.csh" file for a password                  */
/*                                                                           */
/*****************************************************************************/

char *dsc_getpass()
{
  char setupfile[10000];      /* memory copy of the dsc_setup.csh file       */
  char *homedir;              /* user's HOME directory path                  */
  char *setupname[DSC_FILEZ]; /* full path and name of the set up file       */
  char *passwrd;              /* DSC password text variable                  */
  char let;
  int count;
  int rstat;
  FILE *fp;                   /* file pointer to dsc_setup.csh               */
  int errcount;

#ifdef DPRGPDEBUG
fprintf(stderr,"dscpr    : dsc_getpass      : void \n");
#endif

  /*                        */
  /* Create setup file name */
  /*                        */
  homedir = getenv("HOME");
  sprintf(setupname,"%s/",homedir);
  strcat(setupname,DSC_SETUP_NAME);

  fp = fopen (setupname,"r");

  if (fp == NULL) {
   fprintf(stderr,"dscpr    : dsc_getpass      : can not read %s\n",setupname);
   fprintf(stderr,"dscpr    : dsc_getpass      : forcing problem halt!\n");
   fflush(stderr);
   exit(DSC_FAILURE);
  }

  /*                                              */
  /* Read set up file and store it in 'setupfile' */
  /*                                              */
  count = 0;
  errcount = 0;

#ifdef DPRGPDEBUG
fprintf(stderr,"dscpr    : dsc_getpass      : reading %s\n",setupname);
#endif

   /*                                                                  */
   /* due to problems relating to reading the setup file to soon after */
   /* another server has accessed it the first getc call might return  */
   /* an EOF. Try some repeated fopens.                                */
   /*                                                                  */
   let = getc(fp);

#ifdef GETPDEBUG
if ( let == EOF )
fprintf(stderr,"dscpr    : dsc_getpass      : 1st read is an EOF\n");
#endif

   while ( ( let == EOF ) && ( errcount < 1000 ) ) {

    fclose(fp);
    fp = fopen (setupname,"r");
    if (fp == NULL) {
    fprintf(stderr,"dscpr    : dsc_getpass      : in errcount and EOF loop\n");
    fprintf(stderr,"dscpr    : dsc_getpass      : can not read %s\n",setupname);
    fprintf(stderr,"dscpr    : dsc_getpass      : forcing problem halt!\n");
    fflush(stderr);
    exit(DSC_FAILURE);
    }
    let = getc(fp);
    errcount++;
#ifdef GETPDEBUG
fprintf(stderr,"dscpr    : dsc_getpass      : try #%d \n",errcount);
#endif
   }

   /*                               */
   /* populate 'setupfile' variable */
   /*                               */
   setupfile[count] = let;
   while (let != EOF )
   {
    count++;
    let = getc(fp);
    setupfile[count] = let;
    }
    setupfile[count+1] = EOF;

    fclose(fp);

   /*                                              */
   /* get the password from the dsc_setup.csh file */
   /*                                              */

   passwrd = dsc_getpass_sub(setupfile);

   if ( strcmp(passwrd,"NOT_FOUND") == 0 )
   {
    fprintf(stderr,"dscpr    : dsc_getpass      : could not find password\n");
    fprintf(stderr,"dscpr    : dsc_getpass      : in %s\n",setupname);
    fprintf(stderr,"dscpr    : dsc_getpass      : forcing problem halt!\n");
    fflush(stderr);
    exit(DSC_FAILURE);
   }

 passinit = TRUE;

#ifdef GETPDEBUG
fprintf(stderr,"dscpr    : dsc_getpass      : returning %s as password\n"
               ,passwrd );
#endif
 
 return ( passwrd );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getnum_sub( file )                                        */
/*                                                                           */
/* inputs:   file - pointer to memory containing the "dsc_setup.csh" file    */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PORT     = if found                                        */
/*            "NOT FOUND"  = if not found                                    */
/*                                                                           */
/* purpose: loop thru memory in search of the DSC_PORT                       */
/*                                                                           */
/*****************************************************************************/

char* dscpr_getnum_sub( file )
char* file;
{
   char *word;       /* word in the "dsc_setup.csh" file                     */
   char *wordbeg;    /* pointer to first character                           */
   int found;        /* flag                                                 */

#ifdef DPRGNSDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum_sub : char* file\n");
#endif 

   found = FALSE;
   word = (char *)malloc(DSC_BUFFZ);
   wordbeg = word;

   /*                                                       */
   /* loop through file searching for the DSC_PORT variable */
   /*                                                       */

  while ( *file != EOF ) {

   /*                     */
   /* skip leading blanks */
   /*                     */
   while ( *file == ' ' ) file ++;

   /*                                          */
   /* get word , a word is delimited by spaces */
   /*                                          */
   while  ( ( *file != ' ') && ( *file != EOF ) )
    { *word = *file;
      word ++; file++;
    }
   *word = '\0';

   /*                                                         */
   /* the pass word is the word after the string DSC_PASSWORD */
   /*                                                         */
   if ( found == TRUE )
     {
#ifdef DPRGNSDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum_sub : found portnumber %s\n",wordbeg);
#endif
      return(wordbeg);
     }

   if (strcmp(wordbeg,DSCTEXTPORT) == 0 )  found = TRUE;

   word = wordbeg;

  }

#ifdef DPRGNSDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum_sub : could not find password\n");
#endif
 
 return("NOT_FOUND");
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_getnum()                                                */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PORT     = if found                                        */
/*            "NOT FOUND"  = if not found                                    */
/*                                                                           */
/* purpose:  search the "dsc_setup.csh" file for a portnumber                */
/*                                                                           */
/*****************************************************************************/

char *dscpr_getnum()
{
  char setupfile[DSC_SETUPFZ]; /* memory copy of the dsc_setup.csh file      */
  char *homedir;               /* user's HOME directory path                 */
  char *setupname[DSC_FILEZ];  /* full path and name of the set up file      */
  char *portnm;                /* DSC port number text variable              */
  char let;
  int count;
  int rstat;
  FILE *fp;                    /* file pointer to dsc_setup.csh              */
  int errcount;

#ifdef DPRGNDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum     : void \n");
#endif

  /*                        */
  /* Create setup file name */
  /*                        */
  homedir = getenv("HOME");
  sprintf(setupname,"%s/",homedir);
  strcat(setupname,DSC_SETUP_NAME);

  fp = fopen (setupname,"r");

  if (fp == NULL) {
   fprintf(stderr,"dscpr    : dscpr_getnum     : can not read %s\n",setupname);
   fprintf(stderr,"dscpr    : dscpr_getnum     : forcing problem  halt!\n");
   fflush(stderr);
   exit(DSC_FAILURE);
  }

  /*                                              */
  /* Read set up file and store it in 'setupfile' */
  /*                                              */
  count = 0;
  errcount = 0;

#ifdef DPRGNDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum     : reading %s\n",setupname);
#endif

   /*                                                                  */
   /* due to problems relating to reading the setup file to soon after */
   /* another server has accessed it the first getc call might return  */
   /* an EOF. Try some repeated fopens.                                */
   /*                                                                  */
   let = getc(fp);

#ifdef DPRGNDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum     : 1st read is an EOF\n");
#endif

   while ( ( let == EOF ) && ( errcount < 1000 ) ) {

    fclose(fp);
    fp = fopen (setupname,"r");
    if (fp == NULL) {
    fprintf(stderr,"dscpr    : dscpr_getnum     : in errcount and EOF loop\n");
    fprintf(stderr,"dscpr    : dscpr_getnum     : can not read %s\n",setupname);
    fprintf(stderr,"dscpr    : dscpr_getnum     : forcing problem halt!\n");
    fflush(stderr);
    exit(DSC_FAILURE);
    }
    let = getc(fp);
    errcount++;
#ifdef DPRGNDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum     : try #%d \n",errcount);
#endif
   }

   /*                               */
   /* populate 'setupfile' variable */
   /*                               */
   setupfile[count] = let;
   while (let != EOF )
   {
    count++;
    let = getc(fp);
    setupfile[count] = let;
    }
    setupfile[count+1] = EOF;

    fclose(fp);

   /*                                                */
   /* get the portnumber from the dsc_setup.csh file */
   /*                                                */

   portnm = dscpr_getnum_sub(setupfile);

   if ( strcmp(portnm,"NOT_FOUND") == 0 )
   {
    fprintf(stderr,"dscpr    : dscpr_getnum     : could not find port#\n");
    fprintf(stderr,"dscpr    : dscpr_getnum     : in %s\n",setupname);
    fprintf(stderr,"dscpr    : dscpr_getnum     : forcing problem halt!\n");
    fflush(stderr);
    exit(DSC_FAILURE);
   }

 portinit = TRUE;

#ifdef DPRGNDEBUG
fprintf(stderr,"dscpr    : dscpr_getnum     : returning %s as portnumber\n"
               ,portnm);
#endif

 return ( portnm );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_netinit()                                               */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: netinit - to flag network already initialized                   */
/*           sockdg  - datagram socket                                       */
/*           sockl   - listen socket                                         */
/*                                                                           */ 
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  set up host information, and udp/tcp sockets                    */
/*                                                                           */
/*****************************************************************************/
 
void dscpr_netinit()
{
   char* getenv();       /* get value of environment variables               */
   int dscpr_end();      /* close socket and unlink the associated file      */
   struct hostent *hp;   /* host information                                 */
   int on;               
   int status;           

#ifdef DPRNIDEBUG
fprintf(stderr,"dscpr    : dscpr_netinit    : void\n");
#endif 

   if (netinit!=0) return;

   if (sockdg != (-1)) close(sockdg);
   if (sockl != (-1)) close(sockl);
   if (lissock != (-1)) close(lissock);

#ifdef DPRNIDEBUG
fprintf(stderr,"dscpr    : dscpr_netinit    : geting process info\n");
#endif

   /*                         */
   /* get process information */
   /*                         */ 

   if ( getcwd(pwd,sizeof(pwd)) == NULL )
    {
     fprintf(stderr,"dscpr    : dscpr_netinit    : getting cwd\n");
     fprintf(stderr,"dscpr    : dscpr_netinit    : errno %d\n",errno);
     fflush( stderr );
     exit(5);
    }

   mypid = getpid();
   gethostname(mynode, DSC_NODEZ);
   hp = gethostbyname( mynode );
   myhost = (long)( *(long *)hp->h_addr );

#ifdef DPRNIDEBUG
fprintf(stderr,"dscpr    : dscpr_netinit    : building name of server\n");
#endif

   /*                      */
   /* build name of server */
   /*                      */
   saddr.sin_family = AF_INET;
   bcopy( hp->h_addr, &saddr.sin_addr, hp->h_length);
   saddr.sin_port = htons(portnm);

   sock = (-1);

#ifdef DPRNIDEBUG
fprintf(stderr,"dscpr    : dscpr_netinit    : setting up udp & tcp sockets\n");
#endif

   /*                   */
   /* set up tcp socket */
   /*                   */
   sockdg = socket(AF_INET, SOCK_DGRAM, 0);
   if (sockdg < 0) {
     fprintf(stderr,"dscpr    : dscpr_netinit    : can not build udp socket\n");
     fprintf(stderr,"dscpr    : dscpr_netinit    : forced problem halt!\n");
     perror( "building udp socket");
     exit(errno);
    }
   on = 1;
   setsockopt( sockdg, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

   sockl = socket(AF_INET, SOCK_STREAM, 0);
   if (sockl < 0) {
     fprintf(stderr,"dscpr    : dscpr_netinit    : can not build tcp socket\n");
     fprintf(stderr,"dscpr    : dscpr_netinit    : forced problem halt!\n");
     perror( "building tcp socket");
     exit(errno);
    }
   on = 1;
   setsockopt( sockl, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
   bcopy( hp->h_addr, &paddr.sin_addr, hp->h_length);
   paddr.sin_port = htons(0);
   paddr.sin_family = AF_INET;
   if (bind(sockl, &paddr, sizeof(paddr))){
     fprintf(stderr,"dscpr    : dscpr_netinit    : can't bind streamsocket\n");
     perror("binding stream socket");
     exit(errno);
   }
   on = sizeof(paddr);
   if (getsockname(sockl, &paddr, &on)){
     fprintf(stderr,"dscpr    : dscpr_netinit    : can't get socket name\n");
     perror( "getting socket name ");
     exit(errno);
   }

   listen( sockl, 3);

 /*                                                                        */
 /* The reason that the sockl stream socket created in dscpr_netinit() can */
 /* not be used is that it is a socket that blocks. A blocking socket is   */
 /* one that when there are I/O requests that cannot complete immediately  */
 /* the process that issued the I/O request using the blocking socket will */
 /* suspend awaiting completion. When the viability of the server is being */
 /* checked and if sockl were to be used the accept() connection command   */
 /* will hang if the server is not activated. Once a socket has been       */
 /* created via the socket() call, it may be marked as non-blocking by the */
 /* fcntl() call, thus avoiding the hanging of the dscctrl program if the  */
 /* server died in an irregular manner (ie rebooting of machines).         */
 /*                                                                        */

 /*                                      */
 /* get a descriptor for a strean socket */
 /*                                      */
  lissock = socket(AF_INET, SOCK_STREAM,0);
  if ( lissock < 0 ) {
     fprintf(stderr,"dscpr    : dscpr_netinit    : can't bld lissock tcp\n");
     fprintf(stderr,"dscpr    : dscpr_netinit    : forced problem halt!\n");
     perror("building tcp sockect");
     exit(errno);
  }

 /*                     */
 /* set sockect options */
 /*                     */
 on = 1;
 setsockopt( lissock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

 /*                              */
 /* mark lissock as non-blocking */
 /*                              */
 if (fcntl(lissock, F_SETFL, FNDELAY) < 0 ) {
   fprintf(stderr,"dscpr    : dscpr_netinit    : can't mark as non-blocking\n");
   fprintf(stderr,"dscpr    : dscpr_netinit    : forced problem halt!\n");
   perror("fcntl F_SETFL SNDELAY");
   exit(errno);
 }

 /*                    */
 /* give socket a name */
 /*                    */
 bcopy( hp->h_addr, &lisaddr.sin_addr, hp->h_length );
 lisaddr.sin_port = htons(0);
 lisaddr.sin_family = AF_INET;
  if (bind( lissock, &lisaddr, sizeof(lisaddr))) {
fprintf(stderr,"dscpr    : dscpr_netinit    : can't bind lissock stream\n");
perror("binding stream socket");
printf("errno is %d \n",errno);
exit(errno);
  }

  /*                             */
  /* get socket name information */
  /*                             */
  on = sizeof(lisaddr);
  if (getsockname(lissock, &lisaddr, &on)) {
fprintf(stderr,"dscpr    : dscpr_netinit    : can't get lissock socket name\n");
perror("getting socket name");
exit(errno);
  }

  /*                                                                        */ 
  /* Set upper bound on number of pending connections that should be queued */
  /*                                                                        */
  listen( lissock, 3);

  netinit = TRUE;

#ifdef sun
   on_exit( dscpr_end );
#endif

#ifdef DPRNIDEBUG
fprintf(stderr,"dscpr    : dscpr_netinit    : init complete, flag set\n");
fprintf(stderr,"dscpr    : dscpr_netinit    : returning void\n");
#endif

   return;
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_netpnm()                                                */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: portnm  - datagram socket port number                           */
/*           netinit - socket setup flag                                     */
/*           overide - overide dscpr_netpnm call, use user supplied port #   */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  get the DSC_PASSWORD from the setup file                        */ 
/*           get the DSC_PORT number from the setup file                     */
/*           if one is found test for an active server                       */
/*           if not search for an available portnumber                       */
/*           terminate if no valid portnumber can be found                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_netpnm()
{
   int status;           /* port number queuery status                       */
   char *dsc_getpass();  /* get password from dsc_setup.csh file             */ 

#ifdef DPRNPNDEBUG
fprintf(stderr,"dscpr    : dscpr_netpnm     : void\n");
#endif

   if (netinit!=0) return;

   /*                                                        */
   /* set the external password variable to the DSC_PASSWORD */
   /*                                                        */
   password = dsc_getpass();

   if (overide == TRUE) { dscpr_netinit(); return; }

   /*                                               */
   /* get socket number from the dsc_setup.csh file */
   /*                                               */
   cportnm = dscpr_getnum();
   if (strcmp(cportnm,"NOT_FOUND") == 0) { 
   fprintf(stderr,"dscpr    : dscpr_netpnm     : did not find portnumber\n");
   strcpy(cportnm,"0");
   }

   portnm = atoi(cportnm);
  
#ifdef DPRNPNDEBUG
fprintf(stderr,"dscpr    : dscpr_netpnm     : using port number %d\n",portnm);
#endif

 
   /*                                                                     */
   /* If the port number found in the file dsc_setp.csh is zero then this */
   /* indicates that there is no activated server, and thus a unused port */
   /* number must be found and bound to. However if the port number is    */
   /* non zero then local communication ends must be established in order */
   /* to verify the existance of the server program.                      */
   /*                                                                     */

   if (portnm != 0) 
    {
#ifdef DPRNPNDEBUG
fprintf(stderr,"dscpr    : dscpr_netpnm     : nonzero port number found\n");
fprintf(stderr,"dscpr    : dscpr_netpnm     : initializing sockets\n");
#endif
     dscpr_netinit();  
    }

#ifdef DPRNPNDEBUG
fprintf(stderr,"dscpr    : dscpr_netpnm     : checking viability of server\n");
fprintf(stderr,"dscpr    : dscpr_netpnm     : or searching for port number\n");
#endif

   status = dscpr_ctrl_portnmchk();
   if ( status != DSC_SUCCESS ) 
    {
fprintf(stderr,"dscpr    : dscpr_netpnm     : can not connect to server\n");
fprintf(stderr,"dscpr    : dscpr_netpnm     : or no valid portnumber found\n");
fprintf(stderr,"dscpr    : dscpr_netpnm     : forced problem halt!\n");
exit(DSC_FAILURE);
    }

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_ctrl_portnmchk()                                        */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: portnm  - datagram socket port number                           */
/*           netinit - socket setup flag                                     */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   DSC_SUCCESS = valid port number verified or found               */
/*           DSC_FAILURE = can not find a portnumber                         */
/*                                                                           */
/* purpose:  if 0 port number search for a valid port number and initialize  */
/*           else if server connection established use port number if not    */
/*           search for a valid port number and initialize                   */
/*                                                                           */
/*****************************************************************************/

int dscpr_ctrl_portnmchk()
{
char *getenv();        /* get value of environment variables                 */
int on;                /* switch for the socket option SO_REUSEADDR          */
int status;
int count;
int i;
struct hostent *hp;    /* host information                                   */
DSCQSTA msg;           /* datagram server status message                     */
DSCSSTA msg2;          /* stream socket status message                       */

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : void\n");
#endif

/*                                                                         */
/* according to the dsc_setup.csh file, no operational server on this node */
/*                                                                         */
if ( portnm == 0 ) {

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : no operational server\n");
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : 0 port number\n");
#endif

 printf("no operational server on this node \n");

 /*                         */
 /* get process information */
 /*                         */

   if ( getcwd(pwd,sizeof(pwd)) == NULL )
    {
     fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : getting cwd\n");
     fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : errno %d\n",errno);
     fflush( stderr );
     exit(5);
    }

 mypid = getpid();
 gethostname(mynode,DSC_NODEZ);
 hp = gethostbyname( mynode );
 myhost = (long)( *(long *)hp->h_addr);

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : searching for available port#\n");
#endif

 dscpr_ctrl_portnm();

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : initializing sockets\n");
#endif

 netinit = FALSE;
 dscpr_netinit();

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : returning DSC_SUCCESS\n");
#endif

 return(DSC_SUCCESS);

}
 
 /*                                                                          */
 /* In the case that the DSC_PORT number is not zero the value may or not be */
 /* valid depending on the termination condition of the server program       */
 /* (ie rebooting of machines). Therefore the viability of the server        */
 /* program must be checked before the port number is assigned               */
 /*                                                                          */
 else {

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : checking for server at %d\n",
               portnm);
#endif

 /*                              */
 /* build status request message */
 /*                              */
 msg.msg_type = msg.sanity = htonl(MSG_QSTAT);
 strcpy(msg.dsc_password , password );
 msg.msg_size = htonl(sizeof( DSCQSTA ));
 msg.flag = htonl(MSG_SSTAT);
 memcpy( &msg.addr, &lisaddr, sizeof(lisaddr));

 /*                                  */
 /* send message to local DSC server */
 /*                                  */
 do {
   errno = 0;
   status = sendto( sockdg, &msg, sizeof(DSCQSTA),0,&saddr,sizeof(saddr));
   if ( errno == ENOBUFS ) sleep(2);
  } while ( ( errno == EINTR ) || ( errno == ENOBUFS ) );

 /*                                            */
 /* try to accept a connection with the server */
 /*                                            */
 count = 0;
 do {
   errno = 0;
   sock = accept( lissock, 0, 0 );
   count = count + 1;
 } while ( (count <= 1000) && (errno > 0 ) );

 /*                                                    */
 /* search for a valid port number, no connection made */
 /*                                                    */
 if  ( errno > 0 )  {

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : no operational server on %d\n",
         portnm);
#endif

  printf("no operational server on %d (cannot connect via DSC_PORT number) \n",
         portnm);

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : searching for available port#\n");
#endif

  dscpr_ctrl_portnm();

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : initializing sockets\n");
#endif

  netinit = FALSE;
  dscpr_netinit();

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : returning DSC_SUCCESS\n");
#endif

  return(DSC_SUCCESS);

 }

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : reading status message \n");
#endif

 /*                           */
 /* get server status message */
 /*                           */
 on=1;
 setsockopt( sock,SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

 errno = EWOULDBLOCK;
 while ( errno == EWOULDBLOCK )
	 {
  errno = 0;
  status = read ( sock, (char *)&msg2, sizeof(DSCSSTA) );
 }

 if ( status != ntohl(msg2.msg_size) ) {
  fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : msg size does not match\n");
  close( lissock );
  close( sock );
  sock = -1;
  return(DSC_FAILURE);
 }

 close( sock );
  sock = -1;

 } 

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_pchk  : returning DSC_SUCCESS\n");
#endif

 return(DSC_SUCCESS);

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_ctrl_portnmchk()                                        */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* ext vars: portnm  - datagram socket port number                           */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose: execute $DSC_PATH/Netcheck/dsc_nc and use the file dsc_us        */
/*          to determine an unused port number  for datagram communication.  */ 
/*          (Inspected processors are found in the file dscrp_net.add)       */
/*                                                                           */
/*****************************************************************************/

int dscpr_ctrl_portnm()
{
  int pid;                 /* the parent process gets the id of the new      */
                           /* process, while the child process gets the      */
                           /* value of 0                                     */
  int wait_result;         /* return value of wait                           */
  int rstat;
  int currportnm;
  int n=0;
  FILE *portfp;
  char *pname[DSC_FILEZ];
  char *dpath;
  union wait statusp;      /* terminated process status                      */
  static char *argv[10];   /* command line for the execvp                    */


#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: void\n");
#endif

  pid = fork();
  if (pid == 0 ) {

     argv[0] = "rsh";
     argv[1] = mynode;
     argv[2] = "csh";
     argv[3] = "$DSC_PATH/Netcheck/dsc_nc";
     argv[4] = 0;

#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: %s %s %s %s \n",
              argv[0],argv[1],argv[2],argv[3]);
fflush(stderr);
#endif

   /*                                                                        */
   /* The child process immediately transfers control to the rsh program     */
   /* The transfer is achieved with the execvp function. This function makes */
   /* a request to the UNIX system to replace the current process (the one   */
   /* calling execvp) with the new process (rsh)                             */
   /*                                                                        */

   pid = execvp( "rsh", argv ); 
   /*                                        */
   /* if control is returned here then error */
   /*                                        */
#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: control is returned, error\n");
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: errno %d\n",errno);
fflush(stderr);
#endif
   perror( "failure in execvp");
   exit(errno);
  } 
  else {

#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: checking for available port#\n");
fflush(stderr);
#endif

  printf("checking for available port numbers ... \n");

#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: waiting on %d\n",pid);
fflush(stderr);
#endif
   
  wait(&statusp);

#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: child done!\n");
fflush(stderr);
#endif

/*
  if ( wait_result != pid )
   {
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: can not find a port number!\n");
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: wait_result %d != PID %d\n",
               wait_result,pid);
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: errno %d\n",errno);
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: forced halt!\n");
fflush(stderr);
exit(errno);    
   }
*/
 /*                                                                        */
 /* Read dsc_usedsockets file and determine a usable portnumber to bind to */
 /*                                                                        */
  dpath = getenv("DSC_PATH");
  sprintf(pname,"%s/Netcheck/",dpath);
  strcat(pname,"dsc_us");

#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: opening %s\n",pname);
fflush(stderr);
#endif

  portfp = fopen(pname,"r");
  if (portfp == NULL) {

#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: can not open %d\n",pname);
#endif

   printf("can not open %d\n",pname);
   rstat = dscpr_end();
   exit(rstat);
  }

 /*                                                        */
 /* Find last used port number from the dsc_usedports file */
 /*                                                        */

#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: finding last used portnumber\n");
fflush(stderr);
#endif

 currportnm = 1; n=1;
 while ( n == 1) 
  { 
    n =  fscanf(portfp,"%d",&currportnm); 
#ifdef DPRCPNDEBUG 1
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: number %d rval %d\n",currportnm,
               n);
fflush(stderr);
#endif
 
  }
 fclose(portfp);

 /*                                */
 /* Use next available port number */
 /*                                */
 if (currportnm > 1) portnm = currportnm + 1;

#ifdef DPRCPNDEBUG 
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: using port number %d\n",portnm); 
#endif

  } /* end if parent */

#ifdef DPRCPNCDEBUG
fprintf(stderr,"dscpr    : dscpr_ctrl_portnm: returning DSC_SUCCESS\n");
#endif

 return(DSC_SUCCESS);

}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  ctrl_dispatch()                                              */
/*                                                                           */
/* inputs:   probname - unique name of problem                               */
/*           fid      - source file id                                       */
/*           types    - source file types                                    */
/*           build    - command to build source file into executable         */
/*           exec     - command to run executable file                       */
/*           level    - 0 then main problem (must run on local node)         */
/*                      > 0  then subproblem                                 */
/*           cpu      - estimated cpu usage of the problem                   */
/*           mem      - estimated mem usage of the problem                   */
/*                                                                           */
/* ext vars: passinit - password set flag                                    */
/*                                                                           */
/* outputs:  (msg to DSC server)                                             */
/*                                                                           */
/* return:   index of problem within server database                         */
/*                                                                           */
/* purpose:  send start message to DSC Server                                */
/*           wait for responce with index of problem                         */
/*                                                                           */
/*****************************************************************************/

int ctrl_dispatch( probname, fid, types, build, exec, level , cpu , mem )
char *probname;
char *fid;
char *types;
char *build;
char *exec;
int level;
int cpu;
int mem;
{
   DSCSTRT msg;          /* start subproblem message                         */
   DSCBEG *msg2;         /* incomming startup message                        */
   int     rlen;         /* lenght of read message                           */
   int     on;
   char   *cptr;
   char   *ctypes;
   char *dsc_getpass();  /* get password from dsc_setup.csh file             */
   char *dscpr_getnum(); /* get port numeber from dsc_setup.csh file         */

#ifdef DPRCDISDEBUG
fprintf(stderr,"dscpr    : ctrl_dispatch    : probname %s\n",probname);
fprintf(stderr,"dscpr    : ctrl_dispatch    : fid %s\n",fid); 
fprintf(stderr,"dscpr    : ctrl_dispatch    : types %s\n",types);
fprintf(stderr,"dscpr    : ctrl_dispatch    : build %s\n",build);
fprintf(stderr,"dscpr    : ctrl_dispatch    : exec %s\n",exec);
fprintf(stderr,"dscpr    : ctrl_dispatch    : level %d\n",level);
fprintf(stderr,"dscpr    : ctrl_dispatch    : cpu %d\n",cpu);
fprintf(stderr,"dscpr    : ctrl_dispatch    : mem %d\n",mem);
#endif

   /*                                    */
   /* save pointer to beginning of types */
   /*                                    */
   ctypes = types;

   if ( passinit == FALSE )
   {

#ifdef DPRCDISDEBUG
fprintf(stderr,"dscpr    : ctrl_dispatch    : getting password\n");
fflush(stderr);
#endif

    password = dsc_getpass();
   }


#ifdef DPRCDISDEBUG
fprintf(stderr,"dscpr    : ctrl_dispatch    : setup start and begin mess\n");
#endif

   /*                                 */
   /* set up start and begin messages */
   /*                                 */
   msg2 = (DSCBEG *)&msg;
   msg.msg_type = msg.sanity = htonl(MSG_START);
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl(sizeof( DSCSTRT ));
   memcpy( &msg.addr, &paddr, sizeof(paddr));

   msg.parent_PID = htonl(mypid);
   msg.fnode = htonl(myhost);
   msg.subflg = htonl(level);
   msg.cpurating = htonl(cpu);
   msg.memrating = htonl(mem);
   strcpy( msg.prob_name, probname);

   /*                                               */
   /* main problems must deal with path information */
   /*                                               */
   if (level == 0) {
      cptr = strchr( fid, '/');
      if (cptr == 0) {
        sprintf( msg.src_name, "%s/%s", pwd, fid);
        cptr = fid;
        }
      else {
        strcpy( msg.src_name, fid);
        cptr++;
        }
     /* special case for off-node servers */
      if (myhost != saddr.sin_addr.s_addr) {
        strcpy( msg.src_name, fid);
        }
    }/* end if main problem */
   else {
     cptr = fid;
     strcpy( msg.src_name, cptr );
   } /* end if subproblem */

   msg.ctyp = 0;
   msg.ltyp = 0;
   msg.mtyp = 0;
   msg.ptyp = 0;
   msg.ityp = 0;
   msg.otyp = 0;
   msg.dtyp = 0;
 

 /*                                                               */
 /* Set flags and build string if using the Lisp interface        */
 /*                                                               */
 if  (strcmp(ctypes,"lc")==0) 
     {
     msg.ltyp = 1;
     strcpy( msg.src_build, build);
     }
 else if (strcmp(ctypes,"lci")==0) 
     {
     msg.ltyp = 1;
     msg.ityp = 1; 
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
      strcpy( msg.src_build, "*");
     }
 else if (strcmp(ctypes,"lo")==0) 
     {
     msg.ltyp = 1;
     msg.otyp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
      strcpy( msg.src_build, "*");
     }
 else if (strcmp(ctypes,"loi")==0) 
     {
     msg.ltyp = 1;
     msg.otyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
      strcpy( msg.src_build, "*");
     }
 else if (strcmp(ctypes,"li")==0)
     {
     msg.ltyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  else if (strcmp(ctypes,"l")==0)
     {
     msg.ltyp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  /*                                                               */
  /* Set flags and build string if using the Mathematica interface */
  /*                                                               */
  else if (strcmp(ctypes,"m")==0)
     {
     msg.mtyp = 1;
          strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  else if (strcmp(ctypes,"mi")==0)
     {
     msg.mtyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  /*                                                               */
  /* Set flags and build string if using the Maple interface       */
  /*                                                               */
  else if (strcmp(ctypes,"p")==0)
     {
     msg.ptyp = 1;
          strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
     }
  else if (strcmp(ctypes,"pi")==0)
     {
     msg.ptyp = 1;
     msg.ityp = 1;
     strcpy( msg.src_build, build);
     if (strlen(msg.src_build) == 0)
       strcpy( msg.src_build, "*");
    }
 /*                                                               */
 /* Set flags and build string if using the C interface           */
 /*                                                               */
 else
 {
   while (*types != '\0') {
      switch (*types) {
      case 'c' :
        msg.ctyp = 1;
        if (*build == '*') msg.src_build[0] = '\0';
        else if (strlen( build ))    strcpy( msg.src_build, build);
        else sprintf( msg.src_build,
		     "cc -o %s %s.c", msg.src_name, msg.src_name);
        break;
      case 'i':
        msg.ityp = 1;
        if (strlen(msg.src_build) == 0)
	  strcpy( msg.src_build, "*");
        break;
      }              /* end switch */
    types++;         /* advance to next character for examination */
    }                /* end while more types */

 } /* End else */

   msg.cpu_type = htonl(DSC_CPU);

#ifdef DPRCDISDEBUG
fprintf(stderr,"dscpr    : ctrl_dispatch    : msg.src_build %s\n",
                msg.src_build);
fflush(stderr);
#endif

   /*                                                               */
   /* Set default execute string if using the Lisp interface        */
   /*                                                               */
   if ((strcmp(ctypes,"li")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.int %s.lsp %s.d %s.i", 
             msg.src_name, msg.src_name, msg.src_name);
   else if ((strcmp(ctypes,"lci")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.cpl %s.lsp %s.d %s.i",
             msg.src_name, msg.src_name, msg.src_name);
   else if ((strcmp(ctypes,"loi")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.obj %s.o %s.d %s.i",
             msg.src_name, msg.src_name, msg.src_name);
   else if  ((strcmp(ctypes,"lc")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.cpl %s.lsp %s.d",
             msg.src_name, msg.src_name);
   else if  ((strcmp(ctypes,"lo")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.obj %s.lsp %s.d",
             msg.src_name, msg.src_name);
   else if ( (*ctypes == 'l') && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_lisp.int %s.lsp %s.d",
                   msg.src_name, msg.src_name);
   /*                                                               */
   /* Set default execute string if using the Mathematica interface */
   /*                                                               */
   else if ( (*ctypes == 'm') && (strlen(exec)==0))
           sprintf(msg.src_exec, "Math %s.m %s.d",
                   msg.src_name, msg.src_name);
   else if ((strcmp(ctypes,"mi")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "Math %s.m %s.d %s.i",
                   msg.src_name, msg.src_name, msg.src_name);
   /*                                                               */
   /* Set default execute string if using the Maple interface       */
   /*                                                               */
   else if ( (*ctypes == 'p') && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_maple.int %s.mpl %s.d",
                   msg.src_name, msg.src_name);
   else if ( (strcmp(ctypes,"pi")==0) && (strlen(exec)==0))
           sprintf(msg.src_exec, "dsc_maple.int %s.mpl %s.d %s.i",
             msg.src_name, msg.src_name, msg.src_name);
   /*                                                               */
   /* Set default execute string if using the C interface           */
   /*                                                               */
   else if ( strlen(exec)==0 ) 
           sprintf( msg.src_exec, "%s %s.d", msg.src_name, msg.src_name );
   /*                                                               */
   /* Use user supplied execute string                              */
   /*                                                               */
   else if (strlen( exec ))   strcpy( msg.src_exec, exec);

#ifdef DPRCDISDEBUG
fprintf(stderr,"dscpr    : ctrl_dispatch    : msg.src_exec %s\n",
                msg.src_exec);
fprintf(stderr,"dscpr    : ctrl_dispatch    : sending message to DSCRP\n");
fflush(stderr);
#endif

  /*                              */
  /* send message to server DSCRP */
  /*                              */
   do {
      errno = 0;
      rlen = sendto( sockdg, (char *)&msg, sizeof(DSCSTRT),
		 0, &saddr, sizeof(saddr));
      if (errno == ENOBUFS) sleep(2);
   } while( (errno==EINTR) || (errno==ENOBUFS));

   if (errno != 0) {
      fprintf(stderr,"dscpr    : ctrl_dispatch    : error sending start msg\n");
      fprintf(stderr,"dscpr    : ctrl_dispatch    : forced problem halt!\n");
      perror("sending start msg to DSC server");
      exit(errno);
   }

   if (level == 0) {
      printf( "\n     file: %s\n", msg.src_name);
      printf(   "build cmd: %s\n", msg.src_build);
      printf(   " exec cmd: %s\n", msg.src_exec);
      if ( msg.cpurating == HIGH_CPU ) printf("  est cpu: HIGH_CPU\n");
      else
      if ( msg.cpurating == MED_CPU )  printf("  est cpu: MED_CPU\n");
      else
      if ( msg.cpurating == LOW_CPU )  printf("  est cpu: LOW_CPU\n");
      else printf("  est cpu: %d (UNKOWN VALUE)\n",msg.cpurating);
      printf("  est mem: %d megs\n", msg.memrating);
      msg2->index = htonl(0);
    } /* end if main problem */
   else if (level > 0) {

#ifdef DPRCDISDEBUG
fprintf(stderr,"dscpr    : ctrl_dispatch    : getting index\n");
#endif

      sock = accept( sockl, 0, 0);

if ( sock == -1 )
{
fprintf(stderr,"dscpr    : ctrl_dispatch    : can not get descriptor %d\n"
              ,sock);
fprintf(stderr,"dscpr    : ctrl_dispatch    : errno %d\n",errno);
fprintf(stderr,"dscpr    : ctrl_dispatch    : fatal error!\n");
exit(errno);
}

      on = 1;
      setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

     /*           */
     /* get index */
     /*           */
      do {
         errno = 0;
         rlen = dsc_read( sock, msg2, sizeof(DSCBEG));
      } while (errno == EINTR );
      if (rlen != ntohl(msg2->msg_size)) {
        perror("reading problem index");
    fprintf( stderr ,"dscpr    : ctrl_dispatch    : mismatch read DSCBEG\n");
    fprintf( stderr ,"dscpr    : ctrl_dispatch    : rlen = %d, msgsize = %d\n",
	     rlen, ntohl(msg2->msg_size));
        fflush( stdout );
        fflush( stderr );
      }
      close( sock );
      sock = -1;
    }/* end if level > 0 (subproblems only ) */

#ifdef DPRCDISDEBUG
fprintf(stderr,"dscpr    : ctrl_dispatch    : returning index %d\n",
        ntohl(msg2->index));
#endif

   return( ntohl(msg2->index) );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  dscpr_wait( index )                                          */
/*                                                                           */
/* inputs: index - index of subproblem to wait on (-1 == wait on all)        */
/*                                                                           */
/* ext vars: passinit - password set flag                                    */
/*           portinit - port set flag                                        */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   status                                                          */
/*                                                                           */
/* purpose:  send a message to DSC RP telling it we want to                  */
/*           know when all subprocesses finish.                              */
/*           block on answer from DSC RP, return success or failure.         */
/*                                                                           */
/*****************************************************************************/
DSCSOLV msgsolv;

int dscpr_wait( index )
int index;
{

   DSCPRBLK msg;         /* block problem message                            */
   DSCSOLV *msg2;        /* incomming unblocking message                     */
   int      rlen;        /* length of read message                           */
   int     rstat;        /* status of wait                                   */
   int on;
   char *dsc_getpass();  /* get password from dsc_setup.csh file             */
   char *dscpr_getnum(); /* get port numeber from dsc_setup.csh file         */


#ifdef DPRWAITDEBUG
fprintf( stderr ,"dscpr    : dscpr_wait       : index %d\n",index); 
#endif

   if ( portinit == FALSE )
   {
    /*                                                                    */
    /* get the datagram communication port number from the dsc setup file */
    /*                                                                    */
    cportnm = dscpr_getnum();
    if (strcmp(cportnm,"NOT_FOUND") == 0) {
       fprintf(stderr,"dscpr    : dscpr_sub        : port number not found\n");
       fprintf(stderr,"dscpr    : dscpr_sub        : forced problem halt!\n");
       rstat = dscpr_end();
       exit(rstat);
    }
   /*                            */
   /* convert it to decimal form */
   /*                            */
   portnm = atoi(cportnm);
   }

   /*                              */
   /* set up communication sockets */
   /*                              */
   dscpr_netinit();

   if ( passinit == FALSE )
   {
    password = dsc_getpass();
   }

   msg.msg_type = msg.sanity = htonl(MSG_PRBLOCK);
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl(sizeof( DSCPRBLK ));
   memcpy( &msg.addr, &paddr, sizeof(paddr));
   msg.from_PID = htonl(mypid);
   msg.index = htonl(index);

#ifdef DPRWAITDEBUG
fprintf( stderr ,"dscpr    : dscpr_wait       : send message to DSCRP\n");
#endif

  /*                              */
  /* send message to server DSCRP */
  /*                              */
   do {
      errno = 0;
      rlen = sendto( sockdg, (char *)&msg, sizeof(DSCPRBLK),
		 0, &saddr, sizeof(saddr));
      if (errno == ENOBUFS) sleep(2);
   } while( (errno==EINTR) || (errno==ENOBUFS));

   if (errno != 0) {
fprintf( stderr ,"dscpr    : dscpr_wait       : error sending MSG_PRBLOCK\n");
fprintf( stderr ,"dscpr    : dscpr_wait       : forced problem halt!\n");
      perror("sending start msg to DSC server");
      exit(errno);
   }

   sock = accept( sockl, 0, 0);
   on = 1;
   setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

  /*                                      */
  /* wait for all subproblems to complete */
  /*                                      */
   msg2 = &msgsolv;
   do {
      errno = 0;
      rlen = dsc_read( sock, (char *)msg2, sizeof( DSCSOLV));
   } while (errno == EINTR );

   if (rlen != ntohl(msg2->msg_size)) {
    perror("reading problem index");
    fprintf( stderr ,"dscpr    : dscpr_wait       : mismatch read DSCBEG\n");
    fprintf( stderr ,"dscpr    : dscpr_wait       : rlen = %d, msgsize = %d\n",
             rlen, ntohl(msg2->msg_size));
     fflush( stdout );
     fflush( stderr );
   }

   close( sock );

   if (ntohl(msg2->msg_type) != MSG_SOLVE) rstat = DSC_FAILURE;
   else rstat = ntohl(msg2->status);
 
#ifdef DPRWAITDEBUG
fprintf( stderr ,"dscpr    : dscpr_wait       : returning rstat %d\n",
        rstat);
#endif

   return( rstat );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_end( )                                                  */
/*                                                                           */
/* inputs: none                                                              */
/*                                                                           */
/* ext vars: sockl  - stream socket                                          */
/*           sockdg - datagram socket                                        */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   0                                                               */
/*                                                                           */
/* purpose:  close socket and unlink the associated file                     */
/*                                                                           */
/*****************************************************************************/

int dscpr_end()
{

#ifdef DPRENDDEBUG
fprintf( stderr ,"dscpr    : dscpr_end        : void\n");
#endif 

   if (sockl != (-1)) close( sockl );
   if (sockdg != (-1)) close( sockdg );

#ifdef DPRENDDEBUG
fprintf( stderr ,"dscpr    : dscpr_end        : returning 0\n");
#endif 

   return( 0 );
}

