/*****************************************************************************/
/*                                                                           */
/*                                 dsc_maple.c                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* functions:                                                                */
/*      main                   - wait for dist cmds from Maple session       */
/*      dscpr_maple_pwd()      - populate the answerfile with the cwd        */
/*      dscpr_maple_sub()      - send a dscpr_sub request to the DSC server  */
/*      dscpr_maple_wait()     - send a dscpr_wait request to the DSC server */
/*      dscpr_maple_next()     - send a dscpr_next request to the DSC server */
/*      dscpr_maple_kill()     - send a dscpr_kill request to the DSC server */
/*      dscpr_maple_cosetup()  - send a cosetup request to the DSC server    */
/*      dscpr_maple_cowait()   - send a cowait request to the DSC server     */
/*      dscpr_maple_coresume() - send a coresume request to the DSC server   */
/*      dscdbg_maple_start()   - send a dscdbg request to the DSC server     */
/*      maple_getinst()        - get particular instance from command file   */
/*      maple_getpid()         - get and save local pid                      */
/*      dsc_maple_setup()      - setup coroutine like behavior for deamon    */
/*      dsc_maple_handle()     - handle SIGUSER2                             */
/*      fscanstring()          - extract string between ""                   */
/*                                                                           */
/*****************************************************************************/

#include "dscpr.c"        /* include DSC C user interface                    */

/*****************************************************************************/
/*                                                                           */
/*                                 definitions                               */
/*                                                                           */
/*****************************************************************************/

#define MSG_MPLKILL      -1 /* Exit dsc_maple deamon                         */ 
#define MSG_DSCPR_PWD     0 /* Get PWD                                       */
#define MSG_DSCPRSTART    1 /* For problems to start subproblems             */
#define MSG_DSCPRWAIT     2 /* wait for one or more subproblems to complete  */
#define MSG_DSCPRKILL     3 /* kill specified subproblem                     */
#define MSG_DSCPRNEXT     4 /* Wait for next subproblem to complete          */
#define MSG_DSCPRCOSETUP  5 /* Setup when using coroutines                   */
#define MSG_DSCPRCOWAIT   6 /* Put a subproblem to sleep                     */
#define MSG_DSCPRCORESUME 7 /* Awaken a sleeping subproblem                  */
#define MSG_DSCDBGSTART   8 /* DEBUGGING FUNCTION                            */
#define DEBUG 1             /* For debugging purposes                        */

/*****************************************************************************/
/*                                                                           */
/*                           externals & prototypes                          */
/*                                                                           */
/*****************************************************************************/

void dsc_maple_setup();
int maple_getpid();
int maple_getinst();
void dscpr_maple_pwd();
void dscpr_maple_sub();
void dscpr_maple_wait();
void dscpr_maple_next();
void dscpr_maple_kill();
void dscpr_maple_cosetup();
void dscpr_maple_cowait();
void dscpr_maple_coresume();
void dscdbg_maple_start();

/*****************************************************************************/
/*                                                                           */
/*                           module level variables                          */
/*                                                                           */
/*****************************************************************************/

int dsc_maple_flag = FALSE;
int mypid;

/*****************************************************************************/
/*                                                                           */
/* subroutine: main (argc, argv)                                             */
/*                                                                           */
/* inputs:    argc    - number of command line arguments                     */
/*            argv[1] - (command file)                                       */ 
/*                      dsc C user functions that are to be invoked          */
/*            argv[2] - (pid file)                                           */
/*                      file that contains the PID of the waiting process    */
/*            argv[3] - (dsc_maple_pid_file)                                 */
/*                      file that contains dsc_maple PID                     */
/*                                                                           */
/* outputs:  file updates, (sub)problems, ...                                */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  daemon program to await for distributing commands from a Maple  */
/*           session                                                         */
/*                                                                           */
/*****************************************************************************/

int main(argc,argv)
int argc;
char **argv;

{
  int pid;       /* pid of process to awaken                                 */
  int instance;  /* instance of DSC C user function                          */
  int i;         /* loop counter                                             */
  FILE *cfile;   /* command file pointer                                     */
  FILE *pfile;   /* pid file pointer                                         */
 
#ifdef DEBUG
fprintf(stderr,"dsc_maple: main()           : starting\n");
fprintf(stderr,"dsc_maple: main()           : argc %d\n",argc);
for(i=0;i<argc;i++)
fprintf(stderr,"dsc_maple: main()           : argv[%d] %s\n",i,argv[i]);
i=0;
fflush(stderr);
#endif

if ( argc !=4 )
{
fprintf(stderr,"dsc_maple: main()           : invalid number of arguments %d\n",
              argc);
fprintf(stderr,"dsc_maple: main()           : dsc_maple\n");
fprintf(stderr,"dsc_maple: main()           : <command_file> \n");
fprintf(stderr,"dsc_maple: main()           : <Awaken_PID_file>\n");
fprintf(stderr,"dsc_maple: main()           : <dsc_maple_PID_file>\n");
fprintf(stderr,"dsc_maple: main()           : forced halt!\n");
fflush(stderr);
exit(1);
}

#ifdef DEBUG
fprintf(stderr,"dsc_maple: main()           : setting up signal handler\n");
fflush(stderr);
#endif

 /*                       */
 /* set up signal handler */
 /*                       */
 
 dsc_maple_setup(argv[3]);             

 /*                                        */
 /* in deamon mode we will run a long time */
 /*                                        */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: main()           : entering deamon mode\n");
fflush(stderr);
#endif

 while ( TRUE )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: main()           : sleeping\n");
fflush(stderr);
#endif
  
     /*                     */
     /* wait in loop again  */
     /*                     */
     dsc_maple_flag = FALSE;           
     while ( !(dsc_maple_flag) )
     sleep(1);                

#ifdef DEBUG
fprintf(stderr,"dsc_maple: main()           : awake\n");
fflush(stderr);
#endif

      /*                                    */
      /* get particualr DSC C user function */ 
      /*                                    */
      instance=maple_getinst(argv[1]); 

      /*                                  */
      /* get pid of the program to awaken */
      /*                                  */
      if ( instance != MSG_MPLKILL ) pid=maple_getpid(argv[2]); 

#ifdef DEBUG
fprintf(stderr,"dsc_maple: main()           : Awaken program PID %d \n", pid);
fprintf(stderr,"dsc_maple: main()           : DSC C user function number : %d\n"
        ,instance);
fflush(stderr);
#endif

      switch(instance) {
      case MSG_MPLKILL:
       unlink(argv[1]);
       unlink(argv[2]);
       unlink(argv[3]);
       fprintf(stderr,"dsc_maple: main()           : exiting\n");
       fflush(stderr);
       exit(0);
       break;      
      case MSG_DSCPR_PWD:
       dscpr_maple_pwd(argv[1],pid);
       break;
      case  MSG_DSCPRSTART:
       dscpr_maple_sub(argv[1],pid); 
       break;
      case MSG_DSCPRWAIT:
       dscpr_maple_wait(argv[1],pid);
       break;
      case MSG_DSCPRKILL:
       dscpr_maple_kill(argv[1],pid);
       break;
      case MSG_DSCPRNEXT:
       dscpr_maple_next(argv[1],pid);
       break;
      case MSG_DSCPRCOSETUP:
       dscpr_maple_cosetup(argv[1],pid); 
       break;
      case MSG_DSCPRCOWAIT:
       dscpr_maple_cowait(argv[1],pid);
       break;
      case MSG_DSCPRCORESUME:
       dscpr_maple_coresume(argv[1],pid);
       break;
      case MSG_DSCDBGSTART:
       dscdbg_maple_start(argv[1],pid);
       break;
      default:
fprintf(stderr,"dsc_maple: main()           : invalid command in command file\n"
       );
fprintf(stderr,"dsc_maple: main()           : forced halt!\n");
       fflush(stderr);
       exit(1);
       break;
      }
     
   } 

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_maple_pwd(name1,apid)                                   */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  populate the answerfile with the cwd                            */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_pwd(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
char ansfile[DSC_FILEZ];   /* file to contain the result of this function    */
char pwd[100];             /* current working directory                      */
FILE *cfile;               /* command file pointer                           */
FILE *ofile;               /* output file pointer                            */


#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_pwd        : name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_pwd        : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_pwd        : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_pwd        : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                          */
/* command file parameters for dscpr_maple_pwd instance are */
/*  instance   - dsc_maple command instance                 */
/*  answerfile - where to deposit the response              */
/*                                                          */

  if ( fscanf(cfile,"%d %s"
      ,&instance,ansfile) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_pwd        : *** START dscpr_maple_pwd ***\n");
fprintf(stderr,"dsc_maple: maple_pwd        : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_pwd        : answer file %s\n",ansfile);
fflush(stderr);
#endif

   if ( getcwd(pwd,sizeof(pwd)) == NULL )
    {
     fprintf(stderr,"dsc_maple: maple_pwd        : getting cwd\n");
     fprintf(stderr,"dsc_maple: maple_pwd        : errno %d\n",errno);
     fflush( stderr );
     exit(5);
    }

     unlink(ansfile);
     ofile = fopen(ansfile,"w");
     if ( ofile == NULL )
      {
fprintf(stderr,"dsc_maple: maple_pwd        : could not open %s\n",ansfile);
fprintf(stderr,"dsc_maple: maple_pwd        : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_pwd        : forced halt!\n");
fflush(stderr);
exit(errno);
      }
     else
      {
       fprintf(ofile,"DSC_M_PWD := `%s`;\n",pwd);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_pwd        : pwd %s\n",pwd);
fflush(stderr);
#endif

       fclose(ofile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_pwd        : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_pwd        : *** END   dscpr_maple_pwd ***\n");
fflush(stderr);
#endif

      if ( kill(apid,SIGUSR2) == -1 )
       {
fprintf(stderr,"dsc_maple: maple_pwd        : could not awaken %d\n",apid);
fflush(stderr);
       }
      }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_pwd        : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_pwd        : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_pwd        : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }
fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_pwd        : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_maple_sub(name1,apid)                                   */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscpr_sub request to the local DSC server                */
/*           populate the answerfile with the returned index                 */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_sub(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
int index;                 /* (sub)problem index                             */
char file[DSC_FILEZ];      /* file id                                        */ 
char types[DSC_FILEZ];     /* file types used                                */
char build[DSC_FILEZ];     /* (sub)problem build                             */
char exec[DSC_FILEZ];      /* (sub)problem exec                              */
int cpu;                   /* estimated cpu ussage                           */
int mem;                   /* estimated mem usage                            */
int f1,f2;
char ansfile[DSC_FILEZ];   /* file to contain the result of this function    */
FILE *cfile;               /* command file pointer                           */
FILE *ofile;               /* output file pointer                            */
char *fscanstring();

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_sub        : name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_sub        : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_sub        : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_sub        : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                          */
/* command file parameters for dscpr_maple_sub instance are */
/*  instance   - dsc_maple command instance                 */
/*  file       - file id                                    */
/*  types      - file types used                            */
/*  build      - (sub)problem build                         */
/*  exec       - (sub)problem exec                          */
/*  cpu        - estimated cpu usage                        */
/*  mem        - estimated mem usage                        */
/*  answerfile - where to deposit the response              */
/*                                                          */

      f1 = fscanf(cfile,"%d",&instance);
      fscanstring(cfile,file);
      fscanstring(cfile,types);
      fscanstring(cfile,build);
      fscanstring(cfile,exec);
      f2 = fscanf(cfile,"%d %d %s",&cpu,&mem,ansfile);


  if ( ( f1 > 0) && (f2 >0) )
   {
#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_sub        : *** START dscpr_sub ***\n"); 
fprintf(stderr,"dsc_maple: maple_sub        : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_sub        : file %s\n",file);
fprintf(stderr,"dsc_maple: maple_sub        : types %s\n",types);
fprintf(stderr,"dsc_maple: maple_sub        : build %s\n",build);
fprintf(stderr,"dsc_maple: maple_sub        : exec %s\n",exec);
fprintf(stderr,"dsc_maple: maple_sub        : cpu %d\n",cpu);
fprintf(stderr,"dsc_maple: maple_sub        : mem %d\n",mem);
fprintf(stderr,"dsc_maple: maple_sub        : answer file %s\n",ansfile);
fflush(stderr);
#endif

     index = dscpr_sub(file,types,build,exec,cpu,mem);
     unlink(ansfile);
     ofile = fopen(ansfile,"w");
     if ( ofile == NULL )
      {
fprintf(stderr,"dsc_maple: maple_sub        : could not open %s\n",ansfile);
fprintf(stderr,"dsc_maple: maple_sub        : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_sub        : forced halt!\n");
fflush(stderr);
exit(errno);
      }
     else
      {
       fprintf(ofile,"DSC_M_INDEX := %d;\n",index);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_sub        : %d\n",index);
fflush(stderr);
#endif 

       fclose(ofile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_sub        : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_sub        : *** END   dscpr_sub ***\n");
fflush(stderr);
#endif

      if ( kill(apid,SIGUSR2) == -1 ) 
       {
fprintf(stderr,"dsc_maple: maple_sub        : could not awaken %d\n",apid); 
fflush(stderr);
       }
      }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_sub        : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_sub        : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_sub        : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }
fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_sub        :  returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_maple_wait(name1,apid)                                  */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscpr_wait request to the local DSC server               */
/*           populate the answerfile with the returned status                */
/*           awaken the sleeping requestor                                   */ 
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_wait(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
int index;                 /* index to wait on (-1 is all)                   */
int status;                /* dscpr_wait return value                        */
char ansfile[DSC_FILEZ];   /* file to contain the result of this function    */
FILE *cfile;               /* command file pointer                           */
FILE *ofile;               /* output file pointer                            */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_wait       :  name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_wait       : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_wait       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_wait       : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                           */
/* command file parameters for dscpr_maple_wait instance are */
/*  instance   - dsc_maple command instance                  */
/*  index      - index to wait on                            */
/*  answerfile - where to deposit the response               */
/*                                                           */

  if ( fscanf(cfile,"%d %d %s",&instance,&index,ansfile) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_wait       : *** START dscpr_wait *** \n");
fprintf(stderr,"dsc_maple: maple_wait       : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_wait       : index %d\n",index);
fprintf(stderr,"dsc_maple: maple_wait       : answer file %s\n",ansfile);
fflush(stderr);
#endif

     status = dscpr_wait(index);
     unlink(ansfile);
     ofile = fopen(ansfile,"w");
     if ( ofile == NULL )
      {
fprintf(stderr,"dsc_maple: maple_wait       : could not open %s\n",ansfile);
fprintf(stderr,"dsc_maple: maple_wait       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_wait       : forced halt!\n");
fflush(stderr);
exit(errno);
      }
     else
      {
       fprintf(ofile,"DSC_M_STATUS := %d;\n",status);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_wait       : status %d\n",status);
fflush(stderr);
#endif

       fclose(ofile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_wait       : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_wait       : *** END   dscpr_wait *** \n");
fflush(stderr);
#endif
      if ( kill(apid,SIGUSR2) == -1 ) 
       {
fprintf(stderr,"dsc_maple: maple_wait       : could not awaken %d\n",apid);
fflush(stderr); 
       }
      }
   }
   else
   {
fprintf(stderr,"dsc_maple: maple_wait       : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_wait       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_wait       : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_wait       : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_maple_next(name1,apid)                                  */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscpr_next request to the local DSC server               */
/*           populate the answerfile with the returned status                */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_next(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
int status;                /* dscpr_next return value                        */
char ansfile[DSC_FILEZ];   /* file to contain the result of this function    */
char afile[DSC_FILEZ];     /* parameter to dscpr_next (answer file)          */
char tfile[DSC_FILEZ];     /* parameter to dscpr_next (answer file)          */
FILE *cfile;               /* command file pointer                           */
FILE *ofile;               /* output file pointer                            */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_next       :  name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_next       : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_next       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_next       : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                           */
/* command file parameters for dscpr_maple_next instance are */
/*  instance   - dsc_maple command instance                  */
/*  afile      - file to deposit next completed subproblem   */
/*  answerfile - where to deposit the response               */
/*                                                           */

  if ( fscanf(cfile,"%d %s %s"
      ,&instance,afile,ansfile) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_next       : *** START dscpr_maple_next ***\n"
       );
fprintf(stderr,"dsc_maple: maple_next       : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_next       : afile %s\n",afile);
fprintf(stderr,"dsc_maple: maple_next       : answer file %s\n",ansfile);
fflush(stderr);
#endif

     status = dscpr_next(tfile);
     unlink(ansfile);
     ofile = fopen(ansfile,"w");
     if ( ofile == NULL )
      {
fprintf(stderr,"dsc_maple: maple_next       : could not open %s\n",ansfile);
fprintf(stderr,"dsc_maple: maple_next       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_next       : forced halt!\n");
fflush(stderr);
exit(errno);
      }
     else
      {
       fprintf(ofile,"DSC_M_STATUS := %d;\n",status);
       fprintf(ofile,"DSC_M_ANSWER := `%s`;\n",tfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_next       : status %d\n",status);
fprintf(stderr,"dsc_maple: maple_next       : next completed subproblem %s\n"
              ,tfile);
fflush(stderr);
#endif

       fclose(ofile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_next       : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_next       : *** END   dscpr_maple_next ***\n"
);
fflush(stderr);
#endif

      if ( kill(apid,SIGUSR2) == -1 )
       {
fprintf(stderr,"dsc_maple: maple_next       : could not awaken %d\n",apid);
fflush(stderr);
       }
      }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_next       : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_next       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_next       : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_next       : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  dscpr_maple_kill(name1,apid)                                 */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscpr_kill request to the local DSC server               */
/*           populate the answerfile with the returned status                */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_kill(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
int index;                 /* index of subproblem to kill                    */
int status;                /* dscpr_kill return value                        */
char ansfile[DSC_FILEZ];   /* file to contain the result of this function    */
FILE *cfile;               /* command file pointer                           */
FILE *ofile;               /* output file pointer                            */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_kill       : name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_kill       : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_kill       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_kill       : orced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                           */
/* command file parameters for dscpr_maple_kill instance are */
/*  instance   - dsc_maple command instance                  */
/*  index      - index of subproblem to kill                 */
/*  answerfile - where to deposit the response               */
/*                                                           */

  if ( fscanf(cfile,"%d %d %s"
      ,&instance,&index,ansfile) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_kill       : *** START dscpr_maple_kill  ***\n"
       );
fprintf(stderr,"dsc_maple: maple_kill       : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_kill       : answer file %s\n",ansfile);
fflush(stderr);
#endif

     status = dscpr_kill(index);
     unlink(ansfile);
     ofile = fopen(ansfile,"w");
     if ( ofile == NULL )
      {
fprintf(stderr,"dsc_maple: maple_kill       : could not open %s\n",ansfile);
fprintf(stderr,"dsc_maple: maple_kill       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_kill       : forced halt!\n");
fflush(stderr);
exit(errno);
      }
     else
      {
       fprintf(ofile,"DSC_M_ERROR := %d;\n",status);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_kill       : status %d\n",status);
#endif

       fclose(ofile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_kill       : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_kill       : *** END   dscpr_maple_kill ***\n"
);
fflush(stderr);
#endif
      if ( kill(apid,SIGUSR2) == -1 )
       {
fprintf(stderr,"dsc_maple: maple_kill       : could not awaken %d\n",apid);
fflush(stderr);
       }
      }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_kill       : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_kill       : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_kill       : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_kill       : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_maple_cosetup(name1,apid)                               */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscpr_cosetup request to the local DSC server            */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_cosetup(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
FILE *cfile;               /* command file pointer                           */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cosetup    : name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_cosetup    : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_cosetup    : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_cosetup    : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                              */
/* command file parameters for dscpr_maple_cosetup instance are */
/*  instance   - dsc_maple command instance                     */
/*                                                              */

  if ( fscanf(cfile,"%d",&instance) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cosetup    : *** START dscpr_maple_cosetup ***\n");
fprintf(stderr,"dsc_maple: maple_cosetup    : instance %d\n",instance);
fflush(stderr);
#endif

      dscpr_cosetup();   

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cosetup    : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_cosetup    : *** END   dscpr_maple_cosetup ***\n");
fflush(stderr);
#endif

      if ( kill(apid,SIGUSR2) == -1 )
       {
fprintf(stderr,"dsc_maple: maple_cosetup    : could not awaken %d\n",apid);
fflush(stderr);
       }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_cosetup    : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_cosetup    : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_cosetup    : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cosetup    : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_maple_cowait(name1,apid)                                */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscpr_cowait request to the local DSC server             */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_cowait(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
int flag;                  /* 1 = copy data file, 0 = do not copy data file  */
FILE *cfile;               /* command file pointer                           */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cowait     : name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_cowait     : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_cowait     : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_cowait     : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                             */
/* command file parameters for dscpr_maple_cowait instance are */
/*  instance   - dsc_maple command instance                    */
/*  flag       - data file copy flag                           */
/*                                                             */

  if ( fscanf(cfile,"%d %d",&instance,&flag) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cowait     : *** START dscpr_maple_cowait ***\n");
fprintf(stderr,"dsc_maple: maple_cowait     : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_cowait     : flag %d\n",flag);
fflush(stderr);
#endif

   dscpr_cowait(flag);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cowait     : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_cowait     : *** END   dscpr_maple_cowait ***\n");
fflush(stderr);
#endif

      if ( kill(apid,SIGUSR2) == -1 )
       {
fprintf(stderr,"dsc_maple: maple_cowait     : could not awaken %d\n",apid);
fflush(stderr);
       }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_cowait     : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_cowait     : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_cowait     : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_cowait     : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscpr_maple_coresume(name1,apid)                              */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscpr_coresume request to the local DSC server           */
/*           populate the answerfile with the returned status                */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscpr_maple_coresume(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
int index;                 /* index to awaken                                */
int status;                /* dscpr_coresume return value                    */
char ansfile[DSC_FILEZ];   /* file to contain the result of this function    */
char com[DSC_FILEZ];       /* input file to [copy] back                      */
FILE *cfile;               /* command file pointer                           */
FILE *ofile;               /* output file pointer                            */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_coresume   : name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_coresume   : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_coresume   : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_coresume   : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                               */
/* command file parameters for dscpr_maple_coresume instance are */
/*  instance   - dsc_maple command instance                      */
/*  index      - index to awaken                                 */
/*  com        - input file to optionaly copy back               */
/*  answerfile - where to deposit the response                   */

  if ( fscanf(cfile,"%d %d %s %s"
      ,&instance,&index,com,ansfile) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_coresume   : *** START dscpr_maple_coresume  *** \n");
fprintf(stderr,"dsc_maple: maple_coresume   : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_coresume   : index sent %d\n",index);
fprintf(stderr,"dsc_maple: maple_coresume   : com %s\n",com);
fprintf(stderr,"dsc_maple: maple_coresume   : answer file %s\n",ansfile);
fflush(stderr);
#endif

     status = dscpr_coresume(index,com);
     unlink(ansfile);
     ofile = fopen(ansfile,"w");
     if ( ofile == NULL )
      {
fprintf(stderr,"dsc_maple: maple_coresume   : could not open %s\n",ansfile);
fprintf(stderr,"dsc_maple: maple_coresume   : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_coresume   : forced halt!\n");
fflush(stderr);
exit(errno);
      }
     else
      {
       fprintf(ofile,"DSC_M_INDEX := %d;\n",status);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_coresume   : index recieved %d\n",status);
#endif

       fclose(ofile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_coresume   : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_coresume   : *** END   dscpr_maple_coresume  *** \n");
fflush(stderr);
#endif

      if ( kill(apid,SIGUSR2) == -1 )
       {
fprintf(stderr,"dsc_maple: maple_coresume   : could not awaken %d\n",apid);
fflush(stderr);
       }
      }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_coresume   : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_coresume   : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_coresume   : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_coresume   : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscdbg_maple_start(name1,apid)                                */
/*                                                                           */
/* inputs:   name1 - command file name                                       */
/*           apid  - pid of waiting process                                  */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  send a dscdbg_start request to the local DSC server             */
/*           populate the answerfile with the returned index                 */
/*           awaken the sleeping requestor                                   */
/*                                                                           */
/*****************************************************************************/

void dscdbg_maple_start(name1,apid)
char *name1;
int apid;
{
int instance;              /* DSC maple interface command instance           */
int index;                 /* (sub)problem index                             */
char ansfile[DSC_FILEZ];   /* file to contain the result of this function    */
FILE *cfile;               /* command file pointer                           */
FILE *ofile;               /* output file pointer                            */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_dscdbg     :  name1 %s apid %d\n",name1,apid);
fflush(stderr);
#endif

cfile = fopen(name1,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_dscdbg     : could not open %s\n",name1);
fprintf(stderr,"dsc_maple: maple_dscdbg     : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_dscdbg     : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {

/*                                                             */
/* command file parameters for dscdbg_maple_start instance are */
/*  instance   - dsc_maple command instance                    */
/*  answerfile - where to deposit the response                 */
/*                                                             */

  if ( fscanf(cfile,"%d %s"
      ,&instance,ansfile) > 0 )
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_dscdbg     : *** START dscdbg_maple_start  ***\n");
fprintf(stderr,"dsc_maple: maple_dscdbg     : instance %d\n",instance);
fprintf(stderr,"dsc_maple: maple_dscdbg     : answer file %s\n",ansfile);
fflush(stderr);
#endif

     index = dscdbg_start();
     unlink(ansfile);
     ofile = fopen(ansfile,"w");
     if ( ofile == NULL )
      {
fprintf(stderr,"dsc_maple: maple_dscdbg     : could not open %s\n",ansfile);
fprintf(stderr,"dsc_maple: maple_dscdbg     : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_dscdbg     : forced halt!\n");
fflush(stderr);
exit(errno);
      }
     else
      {
       fprintf(ofile,"DSC_M_INDEX := %d;\n",index);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_dscdbg     : index %d\n",index);
fflush(stderr);
#endif

       fclose(ofile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_dscdbg     : awakening message sender %d\n"
              ,apid);
fprintf(stderr,"dsc_maple: maple_dscdbg     : *** END  dscdbg_maple_start  ***\n");
fflush(stderr);
#endif

      if ( kill(apid,SIGUSR2) == -1 )
       {
fprintf(stderr,"dsc_maple: maple_dscdbg     : could not awaken %d\n",apid);
fflush(stderr);
       }
      }
     }
   else
   {
fprintf(stderr,"dsc_maple: maple_dscdbg     : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_dscdbg     : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_dscdbg     : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_dscdbg     : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: maple_getinst(name)                                           */
/*                                                                           */
/* inputs:   name - name of file to read for DSC C function instance         */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   message instance number                                         */
/*                                                                           */
/* purpose:  read message instance found in the file "name"                  */
/*                                                                           */
/*****************************************************************************/

int maple_getinst(name)
char *name;
{
int instance;              /* DSC maple interface command instance           */
FILE *cfile;               /* command file pointer                           */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_getinst    : name %s\n",name);
fflush(stderr);
#endif

cfile = fopen(name,"r");
if ( cfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_getinst    : could not open %s\n",name);
fprintf(stderr,"dsc_maple: maple_getinst    : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_getinst    : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {
  if ( fscanf(cfile,"%d",&instance) > 0 ) 
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_getinst    : returning %d\n",instance);
fflush(stderr);
#endif

    fclose(cfile);
    return(instance);
   }
  else
   {
fprintf(stderr,"dsc_maple: maple_getinst    : can not read command file\n");
fprintf(stderr,"dsc_maple: maple_getinst    : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_getinst    : forced halt!\n");
fflush(stderr);
exit(errno);
   }
  }

 fclose(cfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_getinst    : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: maple_getpid(name)                                            */
/*                                                                           */
/* inputs:   name - name of file to read for PID                             */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   integer pid found in the file "name"                            */
/*                                                                           */
/* purpose:  read pid found in the file "name"                               */
/*                                                                           */
/*****************************************************************************/

int maple_getpid(name)
char *name;
{
int pid;                   /*  DSC maple interface command instance          */
FILE *pfile;               /* command file pointer                           */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_getpid     : name %s\n",name);
fflush(stderr);
#endif

pfile = fopen(name,"r");
if ( pfile == NULL )
 {
fprintf(stderr,"dsc_maple: maple_getpid     : could not open %s\n",name);
fprintf(stderr,"dsc_maple: maple_getpid     : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_getpid     : forced halt!\n");
fflush(stderr);
exit(errno);
 }
else
 {
  if ( fscanf(pfile,"%d",&pid) > 0 ) 
   {

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_getpid     : returning %d\n",pid);
fflush(stderr);
#endif

    fclose(pfile);
    return(pid);
   }
  else
   {

fprintf(stderr,"dsc_maple: maple_getpid     : problem reading PID \n");
fprintf(stderr,"dsc_maple: maple_getpid     : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_getpid     : forced halt!\n");
fflush(stderr);
exit(errno);
   }
 }

 fclose(pfile);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_getpid     : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_maple_setup(name)                                         */
/*                                                                           */
/* inputs:   name - file to contain mypid                                    */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  set up for catch for SIGUSR2 and set the wait loop flag to false*/
/*                                                                           */
/*****************************************************************************/

void dsc_maple_setup(name)
char *name;
{
 FILE *mfile;               /* mypid file pointer                            */
 void dsc_maple_handle();   /* signal handler                                */

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_setup      : name %s\n",name);
fflush(stderr);
#endif

 dsc_maple_flag = FALSE;     
 mypid = getpid();            

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_setup      : dsc_maple_flag set to FALSE\n");
fprintf(stderr,"dsc_maple: maple_setup      : mypid %d\n",mypid);
fflush(stderr);
#endif

 mfile = fopen(name,"w");
 if ( mfile == NULL )
  {
fprintf(stderr,"dsc_maple: maple_setup      : could not open %s\n",name);
fprintf(stderr,"dsc_maple: maple_setup      : errno %d\n",errno);
fprintf(stderr,"dsc_maple: maple_setup      : forced halt!\n");
fflush(stderr);
exit(errno);
  }
 else
  {
   fprintf(mfile,"%d",mypid);
  }
 fclose(mfile);
 signal( SIGUSR2, dsc_maple_handle);

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_setup      : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* signal handler: dsc_maple_handle()                                        */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  reset the wait loop flag to true                                */
/*                                                                           */
/*****************************************************************************/

void dsc_maple_handle()
{

#ifdef DEBUG
fprintf(stderr,"dsc_maple: maple_handle     : recieved SIGUSR2\n");
fflush(stderr);
#endif

dsc_maple_flag = TRUE;

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: fscanstring(string,save)                                      */
/*                                                                           */
/* inputs:   string - file pointer to string                                 */
/*           save   - characters within the ""                               */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  put the character between "" into the memory pointed to by save */
/*                                                                           */
/*****************************************************************************/

char *fscanstring(string,save)
FILE *string;
char *save;
{
char c;

/*                                      */
/* take off leading characters before " */
/*                                      */
c = fgetc(string);
while ( c  != '"') c = fgetc(string);

c = fgetc(string);
while ( c != '"' )
 {
  if ( c == EOF )
   {
    fprintf(stderr,"dsc_maple: fscanstring      : EOF found before EOS\n");
    fflush(stderr);
    exit(1);
   }
  *save++ = c;
  c = fgetc(string);
 }
 *save = '\0';
 return(save);
}

