##############################################################################
#                                                                            #
#                            dsc_maple_lib.mpl                               #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
# functions:                                                                 #
#      copyfiles        - make copies of a source file                       #
#      writefiles       - populate files with quoted expression              #
#      is_file_exist    - check existance of a file in the CWD               #
#      dataconvert      - convert integer file to Maple assignment stmnt     #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
#                                   globals                                  #
#                                                                            #
##############################################################################

# git     - global integer variable 

##############################################################################
#                                                                            #
# subroutine: copyfiles                                                      #
#                                                                            #
# inputs:  numfiles     - number of copies of SOURCE_FILE.ext needed         #
#          SOURCE_FILE  - the path and name of the source file w/o           #
#                         the extension                                      #
#          extension    - the language extension of the source file          #
#                         for example c,lsp,mpl,m,o                          #
#                                                                            #
# outputs: copied files                                                      #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: make numfiles copies of SOURCE_FILE to SOURCE_FILE#.extension     #
#                                                                            #
##############################################################################

copyfiles := proc(numfiles,SOURCE_FILE,extension)
 local readfile,         # source of copy string
       writefile,        # destination of copy string
       y,                # copy loop counter
       intfs_echo,       # value of echo
       intfs_pprint,     # value of prettyprint
       copystr;          # unix copy command

 #                                               #
 # Get the current value of echo and prettyprint #
 #                                               #
 intfs_echo:=interface(echo);
 intfs_pprint:=interface(prettyprint);

 #                                                                        # 
 # Do not echo under any circumstance and use the lprint method to render #
 # maple expressions in the user interface                                #
 #                                                                        #
 interface(echo=0,prettyprint=false);

 #                                                   #
 # Copy from "SOURCE_FILE.ext" to "SOURCE_FILE#.ext" #
 #                                                   #
 readfile:= cat(SOURCE_FILE,`.`,extension); 
 for y from 0 to (numfiles - 1) do
   writefile:= cat(SOURCE_FILE,y,`.`,extension);
   copystr:= cat(`cp `,readfile,` `,writefile);
   system(copystr);
 od;

 #                                        #
 # Restore values of echo and prettyprint #
 #                                        #
 interface(echo=intfs_echo,prettyprint=intfs_pprint);

end;

##############################################################################
#                                                                            #
# subroutine: writefiles                                                     #
#                                                                            #
# inputs:  numfiles     - number of copies of SOURCE_FILE.ext needed         #
#          SOURCE_FILE  - the path and name of the source file w/o           #
#                         the extension                                      #
#          extension    - the language extension of the source file          #
#                         for example c,lsp,mpl,m,o                          #
#          com          - command that is evaluated, and its result saved    #
#                         in the file                                        # 
#                                                                            #
# globals: git - global integer variable, used locally as a loop counter     #
#                com can be and expression that depends on git and thus      #
#                change at each iteration                                    #
#                                                                            #
# example: writefiles(2,`file`,`i`,                                          #
#          'print(cat(`I1:=`,git,`;I2:=`,git,`;`))');                        #
#                                                                            #
# outputs: created files                                                     #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: create numfiles of SOURCE_FILE#.extension                         #
#          with each file containing what com evaluates to                   #
#                                                                            #
##############################################################################

writefiles := proc(numfiles,SOURCE_FILE,extension,command)
 local 
       fid,           # file id to write to 
       intfs_echo,    # value of echo
       intfs_pprint,  # value of prettyprint
       copystr;       # unix copy command

 #                                               #
 # Get the current value of echo and prettyprint #
 #                                               #
 intfs_echo:=interface(echo);
 intfs_pprint:=interface(prettyprint);

 #                                                                        #
 # Do not echo under any circumstance and use the lprint method to render #
 # maple expressions in the user interface                                #
 #                                                                        #
 interface(echo=0,prettyprint=false);

 #                             #
 # Create "SOURCE_FILE#.ext"   #
 # populate each file with com # 
 #                             #
 for git from 0 to (numfiles - 1) do
   fid:=cat(SOURCE_FILE,git,`.`,extension);

   #                                                                  #
   # write to the file what is found in the command quoted expression #
   #                                                                  #
   writeto(fid);
   eval(command);

   #                          #
   # stop writing to the file #
   #                          #
   writeto(terminal);
  od;

 #                                        #
 # Restore values of echo and prettyprint #
 #                                        #
 interface(echo=intfs_echo,prettyprint=intfs_pprint);

end;

##############################################################################
#                                                                            #
# subroutine: is_file_exist                                                  #
#                                                                            #
# inputs:  filename - name of file to check                                  #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  1 = if file exists                                                #
#          0 = if file not exists                                            #
#                                                                            #
# purpose: check for the existance of a file in the current working directory#
#                                                                            #
##############################################################################

is_file_exist := proc(filename)
 local result,     # boolean result returned
       test_str,   # string to test file name
       sys_status; # status of system command

 test_str:= cat(`test -f `.filename);

 sys_status:= system(test_str);

 if sys_status = 0 then result:=1;
 else result:=0;
 fi;

 RETURN(result);

end;

##############################################################################
#                                                                            #
# subroutine: dataconvert                                                    #
#                                                                            #
# inputs:   filename - name of data file containing an integer               #
#           var      - name of variable that will hold the data              #
#                                                                            #
# example: fid :=`file`; dataconvert(fid,DATA);                              #  
#                                                                            #
# outputs: .dd file                                                          #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: convert the integer found in filename to a Maple assignment       #
#          statement (var:=integer in file) and save the assignment          #
#          into filenam.dd file                                              #
#                                                                            #
##############################################################################

dataconvert:= proc(filename,var)

 local newfile;

 newfile:= cat(filename,`.dd`);

 #                                             #
 # write the .dd file in maple readable format #
 #                                             #
 system(`echo -n "`.var.`:=" > `.newfile);
 system(`cat `.filename.` >> `.newfile);
 system(`sed s/\\\$/\\\;/ `.newfile.` > `.newfile.`.tmp`);
 system(`mv `.newfile.`.tmp `.newfile);

end;
