/*****************************************************************************/
/*                                                                           */
/*                              dsc_copy.c                                   */
/*                                                                           */
/*****************************************************************************/

#include "dscrp.h"        /* include server external variables               */
#include <sys/time.h>     /* time functions                                  */

/*****************************************************************************/
/*                                                                           */
/* functions :                                                               */
/*     dsc_cpyctrl( msg, s )       service a copy request                    */
/*     dsc_cpyreq( flg, name, np)  server copy function                      */
/*     dscc_send( sock, name )     send local file to remote node            */
/*     dscc_recv( sock, name)      recieve file from remote node             */
/*                                                                           */
/*****************************************************************************/

static DSCCPY               msgcpy;
static int                  cpysok;

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscd_cpyctrl( msg, s )                                        */
/*                                                                           */
/* inputs:   msg - message requesting a file be sent                         */
/*           s   - socket for the transmission                               */
/*                                                                           */
/* outputs:  (either send a file or receive a file)                          */
/*                                                                           */
/* return:   0 = DSC_SUCCESS                                                 */
/*           n = error mask                                                  */
/*                                                                           */
/* purpose: service a copy request                                           */
/*          if request is to send a file then call dscc_send()               */
/*          else call dscc_recv()                                            */
/*          close stream socket                                              */
/*                                                                           */
/*****************************************************************************/

int dscd_cpyctrl( msg, s )
DSCCPYCTRL *msg;
int s;
{
   int rstat;              /* subroutine status                              */

#ifdef DCPYCTRLDEBUG
fprintf(stderr,"dsc_serv : dscd_cpyctrl     : DSCCPYCTRL *msg, int s\n");
fprintf(stderr,"dsc_serv : dscd_cpyctrl     : msg_type %ld msg_size %ld \n",
        ntohl(msg->msg_type),ntohl(msg->msg_size));
if ( ntohl(msg->flag) == -1 )
fprintf(stderr,"dsc_serv : dscd_cpyctrl     : EOF reached\n"); 
#endif


  /*                */
  /*  send the file */
  /*                */
  if (ntohl(msg->flag) == 0) {
    rstat = dscc_send( s, msg->fname );

#ifdef DCPYCTRLDEBUG
fprintf(stderr,"dsc_serv : dscd_cpyctrl     : send %s\n",msg->fname);
#endif 

    }
  /*                */
  /* recv the file  */
  /*                */
  else if (ntohl(msg->flag) == 1) {

#ifdef DCPYCTRLDEBUG
fprintf(stderr,"dsc_serv : dscd_cpyctrl     : recieve %s\n",msg->fname); 
#endif

    rstat = dscc_recv( s, msg->fname );
    }

   /*                        */
   /* close socket when done */
   /*                        */
   close( s );

#ifdef DCPYCTRLDEBUG
fprintf(stderr,"dsc_serv : dscd_cpyctrl     : returning rstat %d\n",rstat);
#endif

   return( rstat );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_cpyreq(flg, name, np  )                                   */
/*                                                                           */
/* inputs:   flg  - 1 = send, 0 = receive                                    */
/*           name - name of file to send/recv                                */
/*           np   - node to get/put file                                     */
/*                                                                           */
/* outputs: (file contents to disk/remote-computer)                          */
/*                                                                           */
/* return:   0 = DSC_SUCCESS                                                 */
/*           n = error mask                                                  */
/*                                                                           */
/* purpose:  create local socket                                             */
/*           send msg to remote computer to connect with this socket         */
/*           wait for remote computer  to connect with socket                */
/*              retry 3 times on timeout, eventually give up.                */
/*           copy the file                                                   */
/*           close the socket(s)                                             */
/*                                                                           */
/*****************************************************************************/ 

int dsc_cpyreq( flg, name, np  )
int flg;
char *name;
NODE_REC *np;
{
  int rstat = DSC_SUCCESS;    /* status of dsc_cpyreq                        */
  DSCCPYCTRL msgreq;          /* request for copy message                    */
  int sock, on;
  struct sockaddr_in server;
  char *cptr;
  struct linger lo;
  int    i;                   /* number of ready descriptors                 */
  int retry_cnt;              /* node retry for copy request                 */
  fd_set ready;
  struct timeval to;

#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : flg %d\n",flg);
fprintf(stderr,"dsc_serv : dsc_cpyreq       : name %s\n",name);
fprintf(stderr,"dsc_serv : dsc_cpyreq       : NODE_REC *np\n");
#endif
 
  if ( np->fsystem == fsystem ) 
  {
#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : no copy!\n");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : same file system\n");
fflush(stderr);
#endif
   return(DSC_SUCCESS);
  } 
#ifdef DCPYREQDEBUG
  else
   {
fprintf(stderr,"dsc_serv : dsc_cpyreq       : different file system\n");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : must copy!\n");
fflush(stderr);
   }
#endif
  
  /*                */
  /* function start */
  /*                */
  cpysok = -1;
  sock = -1;
  msgreq.msg_type = msgreq.sanity = htonl(MSG_COPYCTRL);
  strcpy(msgreq.dsc_password , password );
  msgreq.msg_size = htonl(sizeof(DSCCPYCTRL));
  msgreq.flag = htonl(flg);

  /*                                   */
  /* disect the file name, remote side */
  /* does not want directory path      */
  /*                                   */
  cptr = strrchr( name, '/');
  if (cptr == 0) cptr = name;
  else cptr++;
  strcpy( msgreq.fname, cptr );

  /*                                          */
  /* construct inet socket with remote server */
  /*                                          */
  sock = socket( AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
     perror( "error opening socket for copy");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : error opening socket for copy\n");
     rstat = errno;
     goto XDONE;
     }

  on = 1;
  setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
  lo.l_onoff = TRUE;
  lo.l_linger = 120;           /* wait for remote node to accept msg */
  setsockopt( sock, SOL_SOCKET, SO_LINGER, &lo, sizeof(lo));

  server.sin_family = AF_INET;
  server.sin_port = 0;
  server.sin_addr.s_addr = INADDR_ANY;
  if (bind( sock, &server, sizeof( server ))) {
    perror( "binding copy socket");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : binding copy socket\n");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : fatal error!\n");
    exit(errno);
    }

  on = sizeof(server);
  if (getsockname( sock, &(msgreq.addr), &on)){
    perror( "getting copy socket name");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : getting copy socket name\n");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : fatal error!\n");
    exit(errno);
    }

  memcpy( &msgreq.addr.sin_addr, &my_addr, sizeof( my_addr ));
  listen( sock, 3);

  /*                           */
  /* send the request for copy */
  /*                           */
  for (retry_cnt = 0; retry_cnt < MAX_RETRY; retry_cnt++) {
    rstat = dsc_dg_send( &msgreq, sizeof( DSCCPYCTRL ), sockdg, 
       np->portg, np->naddr);
    if (rstat != DSC_SUCCESS ) goto XDONE;

#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : retry number %d\n",retry_cnt);
#endif

    /*                               */
    /* wait up to 20 seconds per msg */
    /*                               */
    to.tv_sec = 60;
    to.tv_usec = 0; 
    FD_ZERO( &ready );
    FD_SET( sock, &ready);
    if ((i=select( sock+1, &ready, 0, 0, &to)) < 0) {
      perror( "select");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : select\n");
fprintf(stderr,"dsc_serv : dsc_cpyreq       : fatal error!\n");
      exit(errno);
      }

#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : select returns %d\n",i);
#endif

    if (i>0) break;
  }

  if (retry_cnt >= MAX_RETRY) {
fprintf(stderr,"dsc_serv : dsc_cpyreq       : copy with %s timed out.\n"
               , np->nname);
fprintf(stderr,"dsc_serv : dsc_cpyreq       : incrementing node err cnt\n");
     np->err_cnt++;
     rstat = DSCERR_P_COMM;
     goto XDONE;
     }

  on = sizeof( server );
  cpysok = accept( sock, &server, &on);

  /*                */
  /*  SEND THE FILE */
  /*                */
  if (flg == 1) {

#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : about to copy file %s to %s\n",
               name, np->nname);
fprintf( stderr, logstr);
#endif

    rstat = dscc_send( cpysok, name );
    }        /* end if sending file from here to there */

  /*                */
  /*  RECV THE FILE */
  /*                */
  else if (flg == 0) {

#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : about to recive file %s from %s\n"
               ,name, np->nname);
   fprintf( stderr, logstr);
#endif

    rstat = dscc_recv( cpysok, name );
    }

#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : copy complete!\n");
fprintf( stderr, logstr);
#endif

XDONE:

   if (cpysok != (-1)) close( cpysok );
   if (sock != (-1))close( sock );

#ifdef DCPYREQDEBUG
fprintf(stderr,"dsc_serv : dsc_cpyreq       : returning rstat %d\n",rstat);
#endif

  return( rstat );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscc_send( ssok, name )                                       */
/*                                                                           */
/* inputs:   ssok - socket to send file on                                   */
/*           name - name of file to send                                     */
/*                                                                           */
/* outputs: (file contents to disk/remote-computer)                          */
/*                                                                           */
/* return:   0 = DSC_SUCCESS                                                 */
/*           n = error mask                                                  */
/*                                                                           */
/* purpose:  open local file for reading.                                    */
/*           read each record and send it to remote computer.                */
/*           send end of file message                                        */
/*           close source file                                               */
/*           read acknowledge message                                        */
/*                                                                           */
/*****************************************************************************/

int dscc_send( ssok, name )
int ssok;
char *name;
{
  FILE *fp;
  long rstat;        /* status of dscc_send                                  */
  DSCCPYCTRL *msg2;  /* request for copy message                             */
  long totz=0;

#ifdef DCSENDDEBUG
fprintf(stderr,"dsc_serv : dscc_send        : ssok %d\n",ssok);
fprintf(stderr,"dsc_serv : dscc_send        : name %s\n",name);
#endif
 
  msgcpy.msg_type = msgcpy.sanity = htonl(MSG_COPY);
  strcpy(msgcpy.dsc_password , password );
  msgcpy.msg_size = htonl(sizeof( DSCCPY) );
  msg2 = (DSCCPYCTRL *) &msgcpy;

  /*                             */
  /* open local file for reading */
  /*                             */
  fp = fopen( name, "r");
  if (fp == NULL) {
    perror( "opening file for copy");
    fprintf( stderr, "\t%s\n", name );
fprintf(stderr,"dsc_serv : dscc_send        : error opening file for copy\n");
fprintf(stderr,"dsc_serv : dscc_send        : filename %s\n",name);
    rstat = errno;
    goto XDONE;
    }

  /*                                  */
  /*        Loop while not EOF        */
  /*                                  */
  /*                                  */
  while (!feof(fp)) {
    rstat = fread( msgcpy.buffer, 1, DSC_BUFFZ, fp);
    if (rstat == 0) {
      perror ("reading file during copy");
fprintf(stderr,"dsc_serv : dscc_send        : err reading file during copy\n");
fprintf(stderr,"dsc_serv : dscc_send        : filename %s\n",name); 
fprintf(stderr,"dsc_serv : dscc_send        : errno %d\n",errno);
      break;
      }
        
#ifdef DCSENDDEBUG
fprintf(stderr,"dsc_serv : dscc_send        : sending %d bytes\n", rstat);
#endif

    msgcpy.cnt = htonl(rstat);
    totz += rstat;
    rstat = write( ssok, &msgcpy, sizeof( DSCCPY ));

    if (rstat != ntohl(sizeof(DSCCPY))) {
      perror( "sending part of file to remote node");
fprintf(stderr,"dsc_serv : dscc_send        : err sending part of file to remote node\n");
fprintf(stderr,"dsc_serv : dscc_send        : filename %s\n",name);
fprintf(stderr,"dsc_serv : dscc_send        : errno %d\n",errno);
      rstat = errno;
      
      break;
      }    /* if error sending */
    else rstat = DSC_SUCCESS;
    }      /* end while not end of file */

  /*                  */
  /* send end of file */
  /*                  */
  msg2->msg_type = msg2->sanity = htonl(MSG_COPYCTRL);
  strcpy(msg2->dsc_password , password);
  msg2->msg_size = htonl( sizeof(DSCCPYCTRL ));
  msg2->flag = (-1);
  msg2->totz = htonl( totz );

#ifdef DCSENDDEBUG
fprintf(stderr,"dsc_serv : dscc_send        : sending EOF : totz = %d bytes\n"
              ,totz);
#endif

  if (write( ssok, msg2, sizeof( DSCCPYCTRL )) <= 0)
     {
     perror( "\tsending EOF");
fprintf(stderr,"dsc_serv : dscc_send        : err sending EOF\n");     
fprintf(stderr,"dsc_serv : dscc_send        : errno %d\n",errno);
     }

  /*                   */
  /* close source file */
  /*                   */
  fclose( fp );

  /*                               */
  /* read acknowledge message      */
  /* with signal interupt handling */
  /*                               */
  do {
     errno = 0;
     msgcpy.msg_type = (-1);
     strcpy(msgcpy.dsc_password , password );	
     rstat = dsc_read( ssok, &msgcpy, sizeof( DSCCPYCTRL ));

#ifdef DCSENDDEBUG
fprintf(stderr,"dsc_serv : dscc_send        : read size = %d\n",rstat);
fprintf(stderr,"dsc_serv : dscc_send        : asked for %d\n"
              ,sizeof(DSCCPYCTRL));
fprintf(stderr,"dsc_serv : dscc_send        : expected %d\n"
              ,htonl(msgcpy.msg_size) );
fflush( stderr );
#endif


     /*                                                        */
     /* if we get bad record, simulate interupt to force retry */
     /*                                                        */
     if ((rstat != 0) && (htonl(msgcpy.msg_type) != MSG_COPYCTRL)) {
fprintf(stderr,"dsc_serv : dscc_send        : send, reading ack, retry, msg_type: %d\n",htonl( msgcpy.msg_type));
        printf( "send, reading ack, retry, msg_type: %d\n",
            htonl( msgcpy.msg_type));
        errno = EINTR;
     }

  } while (errno == EINTR );

  if (rstat <= 0)
    {
     perror( "\treceiving EOF");
fprintf(stderr,"dsc_serv : dscc_send        : error receiving EOF\n");
    }
  else {

#ifdef DCSENDDEBUG
totz = ntohl( msg2->totz);
fprintf(stderr,"dsc_serv : dscc_send        : rcv EOF : totz = %d bytes\n"
              , totz);
#endif

     rstat = DSC_SUCCESS;
     }

XDONE:

#ifdef DCSENDDEBUG
fprintf(stderr,"dsc_serv : dscc_send        : returning rstat %d\n",rstat);
#endif

  return( rstat );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscc_recv(rsok, name  )                                       */
/*                                                                           */
/* inputs:   rsok - socket to read file on                                   */
/*           name - name of file to create                                   */
/*                                                                           */
/* outputs: (file contents to disk)                                          */
/* return:   0 = DSC_SUCCESS                                                 */
/*           n = error mask                                                  */
/*                                                                           */
/* purpose:  open file locally to put contents of remote file.               */
/*                                                                           */
/*****************************************************************************/

int dscc_recv( rsok, name )
int rsok;
char *name;
{
  DSCCPYCTRL *msg2;     /* recive file message                               */
  FILE *fp;
  long rstat, size;
  long totz=0, totx=0;

#ifdef DCRECVDEBUG
fprintf(stderr,"dsc_serv : dscc_recv        : rsok %d\n",rsok);
fprintf(stderr,"dsc_serv : dscc_recv        : name %s\n",name);
#endif 

  msg2 = (DSCCPYCTRL *) &msgcpy;
  msg2->msg_type = 0;
  strcpy(msg2->dsc_password , password );
  msg2->msg_size = 0;

  /*                              */
  /* open source file for writing */
  /*                              */
  fp = fopen( name, "w");
  if (fp == NULL) {
    perror( "opening/create file for copy");
    fprintf( stderr, "\t%s\n", name );
    rstat = errno;
fprintf(stderr,"dsc_serv : dscc_recv        : opening/create %s for copy\n",
                name);
fprintf(stderr,"dsc_serv : dscc_recv        : errno %d\n",rstat);
    goto XDONE;
    }

  /*                     */
  /* loop for whole file */
  /*                     */
  do {
     /*                                        */
     /* read file record                       */
     /* repost read when interupted by signals */
     /*                                        */
     do {
        errno = 0;
        rstat = dsc_read( rsok, &msgcpy, sizeof( DSCCPY ));

#ifdef DCRECVDEBUG
fprintf(stderr,"dsc_serv : dscc_recv        : read size = %d\n",rstat);
fprintf(stderr,"dsc_serv : dscc_recv        : asked for %d\n",sizeof(DSCCPY));
fprintf(stderr,"dsc_serv : dscc_recv        : expected %d\n"
              ,htonl(msgcpy.msg_size) );
fflush( stderr );
#endif
        if (errno == EINTR) 
          fprintf(stderr,"dsc_serv : dscc_recv        : EINTR during read\n");
     } while (errno == EINTR );

    /*                                      */
    /* let the world know when errors occur */
    /*                                      */
    if (rstat == 0) {
      perror( "\tcontents of msg trashed during copy");
fprintf(stderr,"dsc_serv : dscc_recv        : msg trashed during copy \n");
fprintf(stderr,"dsc_serv : dscc_recv        : ABNORMAL END OF COPY, zero bytes read\n");
      break;     /* socket terminated ? */
      }

    /*                                             */
    /* next COPYCTRL message indicates end of file */
    /*                                             */
    if (ntohl(msgcpy.msg_type) == MSG_COPYCTRL) {

#ifdef DCRECVDEBUG
totz = ntohl( msg2->totz);
fprintf(stderr,"dsc_serv : dscc_recv        : rcv EOF : totz = %ld bytes\n"
              , totz);
#endif
      rstat=DSC_SUCCESS;
      break;
      }  /* end if end of file */

    /*                               */
    /* COPY msg contains next record */
    /* anything else here is error   */
    /*                               */
    else if (ntohl(msgcpy.msg_type) != MSG_COPY) {
      perror( "\tcontents of msg trashed during copy");
fprintf(stderr,"dsc_serv : dscc_recv        : contents of msg trashed during copy\n");
fprintf(stderr,"dsc_serv : dscc_recv        : ABNORMAL END OF COPY\n");
fprintf(stderr,"dsc_serv : dscc_recv        : bad msg type %d, size %d\n",
        ntohl(msgcpy.msg_type), rstat);
      break;     /* socket terminated ? */
      }
    size = ntohl( msgcpy.cnt );
    totx += size;

#ifdef DCRECVDEBUG
fprintf(stderr,"dsc_serv : dscc_recv        : %ld bytes copied, msg size = %ld\n", size, rstat);
#endif

    rstat = fwrite( msgcpy.buffer, 1, size, fp);
    if (rstat != size) {
      perror( "writing file on this node");
fprintf(stderr,"dsc_serv : dscc_recv        : writing file on this node %s\n",
               my_node);
fprintf(stderr,"dsc_serv : dscc_recv        : errno %d\n",errno);
      rstat = errno;
      break;
      }    /* if error sending */
    else rstat = DSC_SUCCESS;
  } while( TRUE );      /* end while not end of file */

  
  /*                                                 */
  /* when whole file is copied, close the local file */
  /*                                                 */
  fclose( fp );

  /*                                                              */
  /* acknowledge completion and allow stream socket to be dropped */
  /* if socket is dropped to soon, we may lose part of file       */
  /*                                                              */
  msg2->msg_type = msg2->sanity = htonl(MSG_COPYCTRL);
  strcpy(msg2->dsc_password , password );
  msg2->msg_size = htonl( sizeof(DSCCPYCTRL ));
  msg2->flag = (-1);
  msg2->totz = htonl( totx );

#ifdef DCRECVDEBUG
fprintf(stderr,"dsc_serv : dscc_recv        : returning EOF : totz = %d bytes\n", totx);
#endif

  if (write( rsok, msg2, sizeof( DSCCPYCTRL )) <= 0)
    {
     perror( "\tsending EOF");
fprintf(stderr,"dsc_serv : dscc_recv        : sending EOF\n");
fprintf(stderr,"dsc_serv : dscc_recv        : errno %d\n",errno);
    }

XDONE:

#ifdef DCRECVDEBUG
fprintf(stderr,"dsc_serv : dscc_recv        : returning rstat %d\n",rstat);
#endif

  return( rstat );
}
