/*****************************************************************************/
/*                                                                           */
/*                                 dsc_nodes.c                               */
/*                                                                           */
/* Change History:                                                           */
/* ~~~~~~~~~~~~~~~                                                           */
/* (AAP) - Modified dscn_chkin() to recognize new dscrp_net.lis format. (Cff)*/
/* (AAP) - Modified dscn_find() to check for C++ type files (dubious).  (Cff)*/
/* (AAP) - Added Ctyp where needed for recognition of C++ files.             */
/*                                                                           */
/*****************************************************************************/

#include "dscrp.h"        /* include server external variables               */

/*****************************************************************************/
/*                                                                           */
/* functions:                                                                */
/*     dscn_init()     initialize local node data                            */
/*     dscn_chkin()    build node database & checkin                         */
/*     dscn_find()     find a node in the database                           */
/*     insert()        insert a node at end of LPTR s node list              */
/*     pr_list()       print current node choice list                        */
/*     dscn_sock()     build stream socket with client process               */
/*     dscn_allchkin() checkin with each node with parent problems           */ 
/*     chkin_loop()    loop thru problems, sending checkin messages          */
/*                                                                           */
/*****************************************************************************/

static NODE_REC node_db[DSC_NODEX];  /* node database                        */
static int ncnt;                     /* number of nodes in the node database */
long nice1=1;                        /* default server nice value            */ 
long nice2=15;                       /* default (sub)problem nice value      */

/*****************************************************************************/
/*                                                                           */
/* dscn_init( sokg )                                                         */
/*                                                                           */
/* Input:   sokg - datagram socket                                           */
/*                                                                           */
/* Output: (node database is initialized)                                    */
/*                                                                           */
/* Return: (int) 0 = DSC_SUCCESS                                             */
/*                                                                           */
/*                                                                           */
/* Descriptions:                                                             */
/*         determine local network name, address, ports                      */
/*                                                                           */
/*****************************************************************************/

int dscn_init()
{
   struct hostent *hp, *gethostbyname(); /* network  host  data              */
   struct servent *sp, *getservbyname(); /* network  host  data              */

   loop_pos = 0;                     /* loop position for checkin            */
   loop_flg = FALSE;                 /* outstanding problem check flag       */

#ifdef NINITDEBUG
fprintf(stderr,"dsc_serv : dscn_init()      : void \n");
fprintf(stderr,"dsc_serv : dscn_init()      : loop position %d \n",loop_pos);
fprintf(stderr,"dsc_serv : dscn_init()      : loop flag %d \n",loop_flg);
fflush( stderr );
#endif

   /*                                    */
   /*    read local network configuation */
   /*      - host address                */
   /*      - server nice value           */
   /*      - (sub) problem nice value    */
   /*                                    */
   gethostname( my_node, DSC_NAMEZ );
   hp = gethostbyname( my_node );
   my_addr = (unsigned long) (*(long *)hp->h_addr);

#ifdef NINITDEBUG
fprintf(stderr,"dsc_serv : dscn_init()      : local host name %s \n",
        my_node);
fflush( stderr );
#endif

   /* Check to see if it is a service */
   /* This is not used due to the run time socket number enhancement */

   /*
   sp = getservbyname( DSC_ING_NAME, "udp");           
   if (sp != 0) my_portg = (unsigned short) sp->s_port;
   */

   my_portg = (unsigned short) htons(portnm);

#ifdef NINITDEBUG
fprintf(stderr,"dsc_serv : dscn_init()      : my_portg = %X(hex), %u(dec)\n",
        (unsigned int)my_portg, (unsigned int)my_portg);
fprintf(stderr,"dsc_serv : dscn_init()      : returning (void) \n");
fflush( stderr );
#endif
}

/*****************************************************************************/
/*                                                                           */
/* dscn_chkin( sokg )                                                        */
/*                                                                           */
/* Input:   sokg - datagram socket                                           */
/*                                                                           */
/* Output:                                                                   */
/*                                                                           */
/* Return: (int) 0 = DSC_SUCCESS                                             */
/*               n = failure                                                 */
/*                                                                           */
/* Descriptions:                                                             */
/*         build checkin message                                             */
/*         read thru network config file                                     */
/*         build node database                                               */
/*         if local node config record, save nice values                     */
/*         if remote node config record, send checkin msg                    */
/*                                                                           */
/*****************************************************************************/

int dscn_chkin( sokg )
int sokg;
{
   int rstat = DSC_SUCCESS; /* assume subproblem(s) not done                 */
   int cflag;               /* C language suport flag                        */ 
   int Cflag;               /* C++ language suport flag (AAP)                */ 
   int lflag;               /* lisp language suport flag                     */
   int mflag;               /* mathematica language suport flag              */
   int pflag;               /* maple language suport flags                   */
   int cpurating;           /* cpu rating of current node in network file    */
   int memrating;           /* memory rating of current node in network file */
   int numprocessors;       /* # of processors of current node               */
   int fsys;                /* file system number                            */
   int tmp;                 /* flag used in node field parsing               */
   long x;                  /* server nice value                             */
   long y;                  /* subproblem nice value                         */
   char *cptr;              /* used in node field parsing                    */
   char temp[180];          /* array used in node field parsing              */
   FILE *fp=(NULL);         /* file pointer to the DSCRP network file        */
   struct hostent *hp;      /* network  host  data                           */
   DSCKIN  msg;             /* node checkin message                          */
   struct hostent *gethostbyname(); /* network  host  data                   */

  loop_poss = 0;

#ifdef NCHKINDEBUG
fprintf(stderr,"dsc_serv : dscn_chkin()     : int sokg \n");
fprintf(stderr,"dsc_serv : dscn_chkin()     : datagram socket %d \n",sokg);
fprintf(stderr,"dsc_serv : dscn_chkin()     : loop position %d \n",loop_pos);
fprintf(stderr,"dsc_serv : dscn_chkin()     : loop flag %d \n",loop_flg);
fflush( stderr );
#endif

   /*                       */
   /* build checkin message */
   /*                       */
   msg.msg_type = msg.sanity = htonl( MSG_CHECKIN );
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl( sizeof( DSCKIN ));
   msg.fnode = my_addr;
   strcpy( msg.fnname, my_node);
   msg.act_prob = ntohl(act_prob);
   msg.blk_prob = ntohl(blk_prob);
   msg.flag = 0;
   msg.reply = htonl(TRUE);

   /*                       */ 
   /* open the network file */
   /*                       */
   fp = fopen( DSCRP_NETFILE, "r");
   if (fp == NULL) {
       rstat = errno;
       perror( "error opening DSC net configuration file");
       fprintf( stderr, "dsc_serv : dscn_chkin()     : can not open%s\n",
                DSCRP_NETFILE);
       fflush( stderr );
       goto XDONE;
       }

#ifdef NCHKINDEBUG
fprintf(stderr,"dsc_serv : dscn_chkin()     : building node database \n");
fflush( stderr );
#endif

  /*                     */
  /* build node database */
  /*                     */
  for (ncnt = 0; ncnt<DSC_NODEX; ) {
     cptr = node_db[ncnt].nname;
     fgets( cptr, DSC_NODEZ, fp);
     if (feof(fp)) break;
     if (*cptr == '*') continue;            /* comment line */

     /*                                                       */
     /* first field is the node name                          */
     /* second field is server nice value                     */
     /* third field is subproblem nice value                  */
     /* fourth field is the C language support field          */
     /* fifth field is the lisp language support field        */
     /* sixth field is the mathematica language support field */
     /* seventh field is the maple language support field     */ 
     /* eigth field is the cpu mips rating                    */ 
     /* ninth field is the megs of memory available           */
     /* tenth field is the number of processors on board      */
     /* eleventh field is the file system identifier          */
     /* fields are delimited by vertical bar character (|)    */
     /*                                                       */
     x = 1; y = 15;   /* assumed nice values                  */

     /*                               */
     /* get the DSC server nice value */
     /*                               */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     tmp = 0;
     while ( *cptr != '|' ) { temp[tmp] = *cptr; cptr++; tmp++;}
     temp[tmp] = '\0';
     if (tmp != 0) x = atoi( temp );

     /*                               */
     /* get the subproblem nice value */
     /*                               */ 
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     tmp = 0;
     while ( *cptr != '|' ) { temp[tmp] = *cptr; cptr++; tmp++;}
     temp[tmp] = '\0';
     if (tmp != 0) y = atoi( temp );

     /*                                    */
     /* get C language support information */
     /*                                    */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     if ( *cptr == 'c' ) cflag = 1; else cflag = 0;

     /*                                            */
     /* get C++ language support information (AAP) */
     /*                                            */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     if ( *cptr == 'C' ) Cflag = 1; else Cflag = 0;

     /*                                       */
     /* get lisp language support information */
     /*                                       */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     if ( *cptr == 'l' ) lflag = 1; else lflag = 0;

     /*                                              */
     /* get mathematica language support information */
     /*                                              */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     if ( *cptr == 'm' ) mflag = 1; else mflag = 0;

     /*                                        */
     /* get maple language support information */
     /*                                        */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     if ( *cptr == 'p' ) pflag = 1; else pflag = 0;

     /*                                    */
     /* get the processor cpu rating       */
     /*                                    */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     tmp = 0;
     while ( *cptr != '|' ) { temp[tmp] = *cptr; cptr++; tmp++;}
     temp[tmp] = '\0';
     if (tmp != 0) cpurating = atoi( temp );

     /*                                    */
     /* get the processor memory avaliable */
     /*                                    */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     tmp = 0;
     while ( *cptr != '|' ) { temp[tmp] = *cptr; cptr++; tmp++;}
     temp[tmp] = '\0';
     if (tmp != 0) memrating = atoi( temp );

     /*                              */
     /* get the number of processors */
     /*                              */
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     tmp = 0;
     while ( isdigit(*cptr) != 0 ) { temp[tmp] = *cptr; cptr++; tmp++;}
     temp[tmp] = '\0';
     if (tmp != 0) numprocessors = atoi( temp );

     /*                            */
     /* get the file system number */
     /*                            */   
     cptr = strchr( cptr, '|');
     *cptr++ = 0;
     tmp = 0;
     while ( isdigit(*cptr) != 0 ) { temp[tmp] = *cptr; cptr++; tmp++;}
     temp[tmp] = '\0';
     if (tmp != 0) fsys = atoi( temp );

     cptr = strchr( node_db[ncnt].nname, '\n');
     if (cptr != 0) *cptr = '\0';
     if (node_db[ncnt].nname[0] == '\0') continue;   /* empty record */

#ifdef NCHKINDEBUG
fprintf(stderr,"dsc_serv : dscn_chkin()     : NODE NAME : %s \n",
        node_db[ncnt].nname);
fprintf(stderr,"dsc_serv : dscn_chkin()     : server nice value %d \n",x);
fprintf(stderr,"dsc_serv : dscn_chkin()     : subproblem nice value %d \n",y);
fprintf(stderr,"dsc_serv : dscn_chkin()     : c language support %d\n",cflag);
fprintf(stderr,"dsc_serv : dscn_chkin()     : c++ language support %d\n",Cflag);
fprintf(stderr,"dsc_serv : dscn_chkin()     : lisp language support %d\n"
        ,lflag);
fprintf(stderr,"dsc_serv : dscn_chkin()     : mathematica language support %d\n"
        ,mflag);
fprintf(stderr,"dsc_serv : dscn_chkin()     : maple language support %d\n"
        ,pflag);
fprintf(stderr,"dsc_serv : dscn_chkin()     : cpu rating %d\n",cpurating);
fprintf(stderr,"dsc_serv : dscn_chkin()     : mem available %d\n",memrating);
fprintf(stderr,"dsc_serv : dscn_chkin()     : # of processors %d\n"
        ,numprocessors);
fprintf(stderr,"dsc_serv : dscn_chkin()     : file system # %d\n",fsys);
fflush( stderr );
#endif

     /*                                  */
     /* if configured node is local node */
     /*                                  */
     if (strcmp( my_node, node_db[ncnt].nname) == 0) {
        nice1 = x;                        /* save the nice value DSC server  */
        nice2 = y;                        /* save the nice value subproblems */
        cpurate = cpurating;              /* save cpu rating                 */
        memrate = memrating;              /* save mem available              */
        numproc = numprocessors;          /* save # of processors            */
        fsystem = fsys;                   /* save the file system number     */
        cpu = 0;                          /* initialize cpu usage            */
        mem = 0;                          /* initialize memory usage         */
        continue;
        }  /* end if local node is listed */


     /*                                         */
     /* translate node name into a node address */
     /*                                         */
     hp = gethostbyname( node_db[ncnt].nname, DSC_NODEZ);
     if (hp == 0) {
        fprintf( stderr, "dsc_serv : dscn_chkin()     : %s is unknown\n"
        ,node_db[ncnt].nname);
        continue;
        }  /* end if node name is not known */

     /*                        */
     /* populate node database */
     /*                        */
     node_db[ncnt].index = ncnt;
     node_db[ncnt].portg = my_portg;  /* already in network order */
     node_db[ncnt].naddr = (unsigned long) ( *(long *)hp->h_addr);
     node_db[ncnt].act_prob = (-1);
     node_db[ncnt].blk_prob = (-1);
     node_db[ncnt].err_cnt = 0;
     node_db[ncnt].cff = cflag;
     node_db[ncnt].Cff = Cflag;   /* (AAP) */
     node_db[ncnt].lff = lflag;
     node_db[ncnt].mff = mflag;
     node_db[ncnt].pff = pflag;
     node_db[ncnt].cpurating = cpurating;
     node_db[ncnt].memrating = memrating;
     node_db[ncnt].numproc =  numprocessors;
     node_db[ncnt].fsystem = fsys;
     node_db[ncnt].cpu = 0;
     node_db[ncnt].mem = 0;

#ifdef NCHKINDEBUG
fprintf(stderr,"dsc_serv : dscn_chkin()     : sending checkin message\n");
fflush( stderr );
#endif

     rstat = dsc_dg_send( &msg, (int) sizeof(DSCKIN), sokg,
                          node_db[ncnt].portg, node_db[ncnt].naddr);
     ncnt++;
     dbtop++;
     }     /* end for all node slots in node database */

XDONE:
   dbtop--; /* Minus 1 for local node */
   if (fp != (NULL)) fclose( fp );

#ifdef DBLKDEBUG
fprintf(stderr,"dsc_serv : dscn_chkin()     : returning (%d) \n",rstat);
fflush(stderr);
#endif

  return( rstat);
} /* end dscn_chkin() */


/*****************************************************************************/
/*                                                                           */
/* dscn_find( flg, arg, filetype )                                           */
/*                                                                           */
/* Input: flg - 0 = find by node address                                     */
/*              1 = find node with less activity than arg                    */
/*              2 = find by offset (arg is ptr to offset)                    */
/*              3 = find node at arg (only if act_prob of node arg is != -1) */
/*              4 = find node by dynamic scheduling                          */ 
/*                                                                           */
/*        arg - node address,offset or index                                 */
/*                                                                           */
/*        filetype - 0 = C or C++ program                                    */
/*                   1 = lisp program                                        */
/*                   2 = mathematica program                                 */
/*                   3 = maple program                                       */
/*                                                                           */
/* Output: arg - ptr to next potential offset (if arg == 2)                  */
/*                                                                           */
/* Return:       (NODE_REC *)             = pointer to node record           */
/*               (NODE_REC *) (EOF)       = not found                        */
/*               (NODE_REC *) (DSCQUEUED) = if problem  must be queued       */
/*                                                                           */
/* Description:                                                              */
/*    search node database, return pointer to record or status               */
/*                                                                           */
/*****************************************************************************/

NODE_REC *dscn_find( flg, arg, filetype )
int flg;
unsigned long arg;
int filetype; 
{
   NODE_REC *nr;            /* node database pointer                         */
   int n;                   /* loop counter                                  */
   int min;                 /* # of active problems when finding node with   */
                            /* least activity in static scheduling           */
   int minn;                /* offset to node with least activity in static  */
                            /* scheduling                                    */
   int tcpu;                /* calculated cpu usage                          */
   int tmem;                /* calculated memory usage                       */
   LPTR head;               /* node selection list pointer                   */
   LPTR p;                  /* node selection list pointer                   */
   LPTR q;                  /* node selection list pointer                   */
   int i;                   /* loop counter                                  */
   int thres1;              /* minimum avaliable memory needed threshold     */
   int thres2;              /* swap factor threshold                         */ 

   /*                                  */
   /* set memory threshold defenitions */
   /*                                  */
   thres1 = MINMEM;    
   thres2 = SWAPFACTOR;    

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : int flg, int agr, int filetype\n"
       );
if ( flg == 0 )
fprintf(stderr,"dsc_serv : dscn_find()      : finding by node address %lu.%lu.%lu.%lu\n",
   (arg & 0xff000000) >> 24,
   (arg & 0x00ff0000) >> 16, 
   (arg & 0x0000ff00) >> 8,
   (arg & 0x000000ff)
);
if ( flg == 1 ) 
fprintf(stderr,"dsc_serv : dscn_find()      : finding with less activity than %lu\n",arg);
if ( flg == 2 )
fprintf(stderr,"dsc_serv : dscn_find()      : finding by offset %lu\n"
        ,arg);
if ( flg == 3 )
fprintf(stderr,"dsc_serv : dscn_find()      : finding node at index %lu\n",arg);
if ( flg == 4 )
fprintf(stderr,"dsc_serv : dscn_find()      : finding node by dynamic scheduling\n");
if ( filetype == 0 )
fprintf(stderr,"dsc_serv : dscn_find()      : filetype is C or C++\n");
if ( filetype == 1 )
fprintf(stderr,"dsc_serv : dscn_find()      : filetype is lisp \n");
if ( filetype == 2 )
fprintf(stderr,"dsc_serv : dscn_find()      : filetype is mathematica\n");
if ( filetype == 3 )
fprintf(stderr,"dsc_serv : dscn_find()      : filetype is maple \n");
fprintf(stderr,"dsc_serv : dscn_find()      : minimum avaliable memory needed %d\n",thres1);
fprintf(stderr,"dsc_serv : dscn_find()      : swap factor %d\n",thres2);
fflush( stderr );
#endif

   switch (flg) {

   /*                                            */
   /* find node by address                       */
   /* addresses are always in network byte order */
   /*                                            */
   case 0:

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : find node by address \n");
fflush( stderr );
#endif

      for (n=(ncnt-1); n>=0; n--) {
        nr = &node_db[n];
        if (nr->index == (-1)) continue;
        if (nr->naddr == arg) break;
      }    /* end for all nodes in database */
      if ((n<0) || (n==ncnt)) nr = (NODE_REC *) (EOF);

#ifdef NFINDDEBUG
if ( nr != (NODE_REC *) (EOF) )
fprintf(stderr,"dsc_serv : dscn_find()      : found record for %s\n",nr->nname);
else
fprintf(stderr,"dsc_serv : dscn_find()      : node record set to (NODE_REC *) (EOF)\n");
fflush( stderr );
#endif

      break;
    /*                                    */
    /* find node with least activity      */
    /* - do not account for node capacity */
    /* - avoid nodes with comm errors     */
    /*                                    */
    case 1:

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : find node with least activity \n"
       );
fflush( stderr );
#endif

      min = arg;
      minn = EOF;
      for (n=(ncnt-1); n>=0; n--) {
        nr = (node_db+n);
        if (nr->act_prob == (-1)) continue;  /* skip unavailable nodes */
        if (nr->err_cnt > 0) continue;       /* skip troublesome nodes */

        if ( (filetype == 0) && (nr->cff == 0) ) continue;
        else 
        if ( (filetype == 0) && (nr->Cff == 0) ) continue;  /* (AAP) */
        else
        if ( (filetype == 1) && (nr->lff == 0) ) continue;
        else 
        if ( (filetype == 2) && (nr->mff == 0) ) continue;
        else 
        if ( (filetype == 3) && (nr->pff == 0) ) continue;

        if (min > nr->act_prob) {
	  min = nr->act_prob;
	  minn = n;
	}
      }    /* end for all nodes in database */
      n = minn;
      nr = (node_db+n);
      if ((n<0) || (n==ncnt)) nr = (NODE_REC *) (EOF);

#ifdef NFINDDEBUG
if ( nr != (NODE_REC *) (EOF) )
fprintf(stderr,"dsc_serv : dscn_find()      : found record for %s\n",nr->nname);
else
fprintf(stderr,"dsc_serv : dscn_find()      : node record set to (NODE_REC *) (EOF)\n");
fflush( stderr );
#endif

      break;

    /*                                                    */
    /* find node by offset (from base of database)        */
    /* [not very commonly used, try to use 0 if possible] */
    /*                                                    */
    case 2:

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : find node by offset (from base of database) \n"
       );
fflush( stderr );
#endif

      n = *((long *)arg);
      if ((n<0) || (n==ncnt)) {
	nr = (NODE_REC *) (EOF);
	break;
        }
      while ((node_db[n].index != n) && (n < ncnt))
         n++;
      if (n < ncnt ) {
         nr = &node_db[n];
         n++;
         }
      else 
         n = -1;
      *((long *)arg) = n;

#ifdef NFINDDEBUG
if ( nr != (NODE_REC *) (EOF) )
fprintf(stderr,"dsc_serv : dscn_find()      : found record for %s\n",nr->nname);
else
fprintf(stderr,"dsc_serv : dscn_find()      : node record set to (NODE_REC *) (EOF)\n");
fflush( stderr );
#endif

      break;

    /*                                                          */
    /* find node at arg (only if act_prob of node arg is != -1) */
    /*                                                          */
    case 3:

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      :find node at arg  \n");
fflush( stderr );
#endif

      if ( arg > dbtop ) { nr = (NODE_REC *) (EOF);  goto XDONE; }
      nr = &node_db[arg]; 
      if (nr->act_prob == (-1)) nr = (NODE_REC *) (EOF);
      if ( strcmp( my_node, node_db[arg].nname) == 0  ) 
           nr = (NODE_REC *) (EOF);

#ifdef NFINDDEBUG
if ( nr != (NODE_REC *) (EOF) )
fprintf(stderr,"dsc_serv : dscn_find()      : found record for %s\n",nr->nname);
else
fprintf(stderr,"dsc_serv : dscn_find()      : node record set to (NODE_REC *) (EOF)\n");
fflush( stderr );
#endif

    break;

    /*                                 */
    /* find node by dynamic scheduling */
    /*   arg - cpu rating              */
    /*   filetype - mem rating         */
    /*                                 */
    case 4:

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : find node by dynamic scheduling\n"
       );
fflush( stderr );
#endif

      /* create head of the possible selection queue */
      head = (LPTR) malloc (sizeof (LST));
      head->c = 0;
      head->next = NULL;

      /*                                                */
      /* insert local node into the queue if there is a */
      /* minimum avaliable memory                       */
      /*                                                */
      tmem = (int)((255-mem)*memrate/255);
      tcpu = (int)((255-cpu)*numproc*cpurate/255);
      if (tmem >= thres1)
        insert (head, -1, tmem, tcpu);
 
      /*                                                     */
      /* insert remaining nodes into the queue if there is a */
      /* minimum avaliable memory                            */
      /*                                                     */
      for (i=0; i<= dbtop; i++)
       {
        tmem = (int)((255-node_db[i].mem)*node_db[i].memrating/255);
        tcpu = (int)((255-node_db[i].cpu)*node_db[i].numproc*
                      node_db[i].cpurating/255);
        if ( (tmem >= thres1) &&
             (node_db[i].err_cnt == 0) &&
             (node_db[i].act_prob != (-1) )
           )
        insert (head, i, tmem, tcpu);
       }

       /*                                                     */
       /* check if any nodes fulfilled the memory requirement */
       /* if not return a "(sub)problem must be queued flag   */
       /*                                                     */
       if (head->next == NULL) 
        {

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : no node fulfills memory threshold 1 \n");
fprintf(stderr,"dsc_serv : dscn_find()      : returning (%d) \n",
        ((NODE_REC *) (DSCQUEUED)));
fflush( stderr );
#endif

         p = head->next;

         while (p)
         {
          q = p;
          p = p->next;
          free (q);
         }
   
         free (head);

         return((NODE_REC *) (DSCQUEUED));
        }

#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : some node(s) fulfilled memory threshold\n");
        pr_list (head);
fprintf(stderr,"dsc_serv : dscn_find()      : searching for best cpu w/r to memory\n");
fflush( stderr );
#endif

       /*                                   */
       /* search for best cpu w/r to memory */
       /*                                   */
       p = head->next;
       q = head;
       while (p)
       {
        if (p->m - filetype < thres2) break;
        if (p->c > q->c) q = p;
        p = p->next;
       }

      if (q->c == 0)
       {
#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : no node fulfills swap factor threshold\n"); 
fprintf(stderr,"dsc_serv : dscn_find()      : returning (%d) \n",
        ((NODE_REC *) (DSCQUEUED)));
fflush(stderr);
#endif
         p = head->next;

         while (p)
         {
          q = p;
          p = p->next;
          free (q);
         }
  
         free (head);

         return((NODE_REC *) (DSCQUEUED));
        }

      /* check for local or remote node */
      if ( q->index == -1 ) nr = (NODE_REC *) (EOF);
      else nr = &node_db[q->index];

#ifdef NFINDDEBUG
if ( nr != (NODE_REC *) (EOF) )
fprintf(stderr,"dsc_serv : dscn_find()      : best choice is %s \n",
        node_db[q->index].nname);
else
fprintf(stderr,"dsc_serv : dscn_find()      : best choice is %s \n",
        my_node);
#endif

     
       /*                         */
       /* if chosen node is local */
       /*                         */
       if ( nr == (NODE_REC *) (EOF) ) 
         {
          /* adding assumed load for compensation */ 
          cpu = (cpu + arg/numproc);
          mem = mem + filetype*255/memrate;
         }
       /*                          */
       /* If chosen node is remote */
       /*                          */
       else
         {
          /* adding assumed load for compensation */
          nr->cpu = nr->cpu + arg/nr->numproc;
          nr->mem = nr->mem + filetype*255/nr->memrating; 
         }

      /*                 */
      /* deallocate list */
      /*                 */
      p = head->next;

      while (p)
      {
       q = p;
       p = p->next;
       free (q);
      }

      free (head);

      break;
    }              /* end switch */

XDONE:
#ifdef NFINDDEBUG
fprintf(stderr,"dsc_serv : dscn_find()      : returning (%d) \n",nr);
fflush(stderr);
#endif
  return( nr );
}

/*****************************************************************************/
/*                                                                           */
/* insert(s, i, mem, cpu)                                                    */
/*                                                                           */
/* Input: s - pointer to a node list                                         */
/*        i - node index number                                              */
/*        mem - current memory usage information                             */
/*        cpu - current cpu usage information                                */
/*                                                                           */
/* Output: none                                                              */
/*                                                                           */
/* Return: none                                                              */
/*                                                                           */
/* Description:                                                              */
/*    insert a node at end of LPTR s node list                               */
/*                                                                           */
/*****************************************************************************/

insert (s, i, mem, cpu)
LPTR    s;
int     i;
int mem,cpu;
{
   LPTR    p;               /* pointer to a node list                        */
   LPTR    q;               /* pointer to a node list                        */ 

#ifdef NINSERT 
fprintf(stderr,"dsc_serv : insert           : LPTR s, int i, int mem, int cpu\n");
fflush(stderr);
#endif

  q = s;
  p = s->next;

  while (p)
  {
    if (mem > p->m) break;
    q = p;
    p = p->next;
  }

  p = (LPTR) malloc (sizeof (LST));
  p->index = i;
  p->m = mem;
  p->c = cpu;
  p->next = q->next;
  q->next = p;

#ifdef NINSERT
fprintf(stderr,"dsc_serv : insert           : returning(void) \n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* pr_list (s)                                                               */
/*                                                                           */
/* Input: s - pointer to a node list                                         */
/*                                                                           */
/* Output: current node choice list                                          */
/*                                                                           */
/* Return: none                                                              */
/*                                                                           */
/* Description:                                                              */
/*    print current node choice list                                         */
/*                                                                           */
/*****************************************************************************/

pr_list (s)

LPTR    s;
{

  s = s->next;
#ifdef NPRLIST 
fprintf(stderr,"dsc_serv : pr_list()        : LPTR s \n");
fprintf(stderr,"dsc_serv : pr_list()        : possible nodes \n");
fflush(stderr);
#endif

  while (s)
  {
fprintf(stderr,"dsc_serv : pr_list()        : name %s\n",
       node_db[s->index].nname);
fprintf(stderr,"dsc_serv : pr_list()        : memory %d\t cpu %d\n",
        s->m, s->c);
fflush(stderr);
    s = s->next;
  }

#ifdef NPRLIST 
fprintf(stderr,"dsc_serv : pr_list()        : returning(void)\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscn_sock( msg )                                              */
/*                                                                           */
/* inputs:    msg - ptr to message containing socket address                 */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   sock - socket descriptor                                        */
/*           -1 - if socket cannot be built                                  */
/* purpose:  build stream socket with client process                         */
/*                                                                           */
/*****************************************************************************/

int dscn_sock( msg )
DSCPRBLK  *msg;
{
   int sok;
   int rstat;
   int count = 0;
   int tmperr;

#ifdef NSOCK
fprintf(stderr,"dsc_serv : dscn_sock()      : msg\n");
fprintf(stderr,"dsc_serv : dscn_sock()      : from PID %d\n",msg->from_PID);
fflush(stderr);
#endif

   msg->addr.sin_family = AF_INET;

#ifdef NSOCK
fprintf(stderr,"dsc_serv : dscn_sock()      : getting stream INET descriptor\n");
fflush(stderr);
#endif

   /*                                   */
   /* get stream INET socket descriptor */
   /*                                   */
   tmperr = ECONNREFUSED;
   while ( ( tmperr == ECONNREFUSED ) && ( count < 10 ) )
   {
   errno = 0;
   tmperr = 0;
   count++;
   sok = socket( AF_INET, SOCK_STREAM, 0);

   if (sok < 0) {

      perror( "opening stream socket to client");
fprintf(stderr,"dsc_serv : dscn_sock()      : opening stream socket to client\n");
fprintf(stderr,"dsc_serv : dscn_sock()      : errno %d\n",errno);
fflush(stderr);    

      exit( errno );
      }

   rstat = 1;
   setsockopt( sok, SOL_SOCKET, SO_REUSEADDR, &rstat, sizeof( rstat));

   /*                                 */
   /* connect with requesting process */
   /* may be local or remote          */
   /*                                 */

#ifdef NSOCK
fprintf(stderr,"dsc_serv : dscn_sock()      : getting ready to conect\n");
#endif

   if (connect( sok, &msg->addr, sizeof( msg->addr )) <0) {

     perror( "dscn_sok - connecting to client socket");
fprintf(stderr,"dsc_serv : dscn_sock()      : error connecting to client socket\n");
fprintf(stderr,"dsc_serv : dscn_sock()      : errno %d\n",errno);
fflush(stderr);

     tmperr = errno;
     close( sok );
     sok = EOF;
     sleep(1);
     }
#ifdef NSOCK
fprintf(stderr,"dsc_serv : dscn_sock()      : sok %d\n",sok);
fprintf(stderr,"dsc_serv : dscn_sock()      : count %d\n",count);
fflush(stderr);
#endif

 } /* End while errno == ECONNREFUSED */

XDONE:

#ifdef NSOCK
fprintf(stderr,"dsc_serv : dscn_sock()      : returning sok %d\n",sok);
fflush(stderr);
#endif

 return( sok ); 

} 

/*****************************************************************************/
/*                                                                           */
/* subroutine: dscn_allchkin( flag )                                         */
/*                                                                           */
/* inputs:   flag 0 = running                                                */
/*                1 = startup                                                */
/*                2 = halting                                                */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   0 = DSC_SUCCESS                                                 */
/*           n = error mask                                                  */
/*                                                                           */
/* purpose:  checkin with each active node                                   */
/*                                                                           */
/*****************************************************************************/

int dscn_allchkin( flag ) 
long flag; 
{ 
 int rstat;                /* return status of dscn_allchkin()               */
 DSCKIN msg;               /* node checkin message                           */
 int p;                    /* offset to dscn_find()                          */
 NODE_REC *np;             /* node record                                    */
 NODE_REC *dscn_find();    /* find node function                             */

#ifdef NALLCHKIN
fprintf(stderr,"dsc_serv : dscn_allchkin()  : flag %d\n",flag);
if (flag == 0 ) fprintf(stderr,"dsc_serv : dscn_allchkin()  : running\n");
if (flag == 1 ) fprintf(stderr,"dsc_serv : dscn_allchkin()  : startup\n");
if (flag == 2 ) fprintf(stderr,"dsc_serv : dscn_allchkin()  : halting\n");
fflush(stderr);
#endif

   /*                       */
   /* start executable code */
   /*                       */
   msg.msg_type = msg.sanity = htonl(MSG_CHECKIN);
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl(sizeof(DSCKIN));
   msg.fnode = my_addr;
   strcpy( msg.fnname, my_node);
   msg.act_prob = htonl(act_prob);
   msg.blk_prob = htonl(blk_prob);
   msg.flag = htonl(flag);
   msg.reply = htonl(FALSE);


#ifdef NALLCHKIN
fprintf(stderr,"dsc_serv : dscn_allchkin()  : sending checkin msg to active nodes\n");
#endif

   /*                                        */
   /* for each active node, send checkin msg */
   /*                                        */

   p = 0;
   while( (np=dscn_find( 2, &p, 0)) != (NODE_REC *)EOF) {
     if (np->act_prob < 0) continue;          /* skip inactive nodes */
      rstat = dsc_dg_send( &msg, (int) sizeof(DSCKIN), sockdg, 
         np->portg, np->naddr);
   } /* end while more nodes */

 XDONE:

#ifdef NALLCHKIN
fprintf(stderr,"dsc_serv : dscn_allchkin()  : returning %s\n",rstat);
fflush(stderr);
#endif

   return( rstat );

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: chkin_loop()                                                  */
/*                                                                           */
/* Input: none                                                               */
/*                                                                           */
/* Output: none                                                              */
/*                                                                           */
/* Return:                                                                   */
/*                                                                           */
/* Description:                                                              */
/*         each time called, find next problem executing off-node            */
/*         send checkin msg to that node                                     */
/*         increase error count by one                                       */
/*         if error count reaches threshold, assume node is dead             */
/*         for dead node, restart problem somewhere else                     */
/*                                                                           */
/*****************************************************************************/

int chkin_loop( ) 
{ 
   int rstat = DSC_SUCCESS; /* assume no problems with subroutine            */
   int p;                   /* loop variable                                 */
   int loop_cnt;            /* stop an infinite loop where all problems are  */
                            /* on node                                       */
   NODE_REC *np;            /* node record pointer                           */
   NODE_REC *dscn_find();   /* functin to find record of checking in node    */
   DSCDB *pp;               /* problem database entry                        */ 
   DSCDB *pa;               /* problem database entry of parent problem      */
   DSCDB *dscp_find();      /* function to find a problem in the database    */
   DSCKIN msg;              /* checkin request message                       */

   /*                       */
   /* start executable code */
   /*                       */

#ifdef CHKINLDEBUG
fprintf(stderr,"dsc_serv : chkin_loop()     : void \n");
fprintf(stderr,"dsc_serv : chkin_loop()     : loop position %d \n",loop_pos);
fflush( stderr );
#endif

   msg.msg_type = msg.sanity = htonl(MSG_CHECKIN);
   strcpy(msg.dsc_password , password );
   msg.msg_size = htonl(sizeof(DSCKIN));
   msg.fnode = my_addr;
   strcpy( msg.fnname, my_node);
   msg.act_prob = htonl(act_prob);
   msg.blk_prob = htonl(blk_prob);
   msg.flag = htonl(0);
   msg.reply = htonl(TRUE);

   if (loop_pos == 0) loop_flg = FALSE;   

   /*                                    */
   /* choose a node to send a message to */
   /*                                    */

#ifdef CHKINLDEBUG
fprintf(stderr,"dsc_serv : chkin_loop()     : choosing a node with a problem to send a message \n");
fflush( stderr );
#endif

   loop_cnt = 0;
   do {
     pp = dscp_find( 2, &loop_pos);
     if (pp == (DSCDB *)EOF) goto XDONE;     /* empty database               */
     if (loop_pos == (EOF)) {
       loop_pos = 0;                         /* loop thru list again         */
       /*                                                             */
       /* the next two lines stop an infinite loop where all problems */
       /* are on-node                                                 */
       /*                                                             */
       loop_cnt++;
       if (loop_cnt > 2) goto XDONE;
       }
   /*                                      */
   /* dont consider problems running local */
   /* or complete.                         */
   /*                                      */
   } while ((pp->enode_addr == my_addr) || (pp->compl_flg));



   np = dscn_find( 0, pp->enode_addr,0);
   if (np == (NODE_REC *) EOF)
   {

#ifdef CHKINLDEBUG
fprintf(stderr,"dsc_serv : chkin_loop()     : could not find problem node node database \n");
fflush( stderr );
#endif

     goto XDONE;
   }

#ifdef CHKINLDEBUG
fprintf(stderr,"dsc_serv : chkin_loop()     : sending message to %s \n",
        np->nname);
fflush( stderr );
#endif

   loop_flg = TRUE;

   if (np->err_cnt < 5) {
      np->err_cnt++;
      rstat = dsc_dg_send( &msg, (int) sizeof(DSCKIN), sockdg, 
         np->portg, np->naddr);
      }
   /*                                                           */ 
   /* if the target node will just not respond, declare it dead */
   /* and re-route the problem somewhere else                   */
   /*                                                           */
   else {
      DSCSTRT msg2;

fprintf( stderr, "dsc_serv : chkin_loop()     : node %s declared dead!\n"
        ,np->nname);
fprintf( stderr, "dsc_serv : chkin_loop()     : re-routing problem \n");
fflush( stderr );

      msg2.msg_type = msg2.sanity = htonl( MSG_START );
      strcpy(msg2.dsc_password , password );
      msg2.msg_size = htonl(sizeof(DSCSTRT));
      msg2.subflg = htonl(pp->depth);
      strcpy( msg2.prob_name, pp->prob_name);
      msg2.parent_PID = htonl(pp->parent_PID);
      msg2.fnode = pp->pnode_addr;
      msg2.ctyp = htonl(pp->ctyp);
      msg2.Ctyp = htonl(pp->Ctyp);                /* (AAP) */
      msg2.ltyp = htonl(pp->ltyp);
      msg2.mtyp = htonl(pp->mtyp);
      msg2.ptyp = htonl(pp->ptyp);
      msg2.otyp = htonl(pp->otyp);
      msg2.ityp = htonl(pp->ityp);
      msg2.dtyp = htonl(pp->dtyp);
      msg2.cpu_type = htonl(pp->cpu_type);
      msg2.cpurating = htonl(pp->cpurating);
      msg2.memrating = htonl(pp->memrating);
      strcpy( msg2.src_name, pp->src_name);
      strcpy( msg2.src_build, pp->src_build);
      strcpy( msg2.src_exec, pp->src_exec);
      if ( pssched == TRUE )
       {

#ifdef CHKINLDEBUG
fprintf(stderr,
"dsc_serv : chkin_loop()     : searching for node using dynamic scheduling \n");
fflush( stderr );
#endif

         /* find node using dynamic scheduling */
         np = dscn_find( 4 , pp->cpurating, pp->memrating);

#ifdef CHKINLDEBUG
if ( np == (NODE_REC *) (EOF) )
 fprintf(stderr,"dsc_serv : chkin_loop()     : local noded chosen %s\n",
         my_node);
else
if  (np == (NODE_REC *) (DSCQUEUED) )
 fprintf(stderr,
 "dsc_serv : chkin_loop()     : could not find a suitable node must queue\n");
else 
 fprintf(stderr,"dsc_serv : chkin_loop()     : %s chosen \n",np->nname);
fflush( stderr );
#endif

       }
      else
       {

#ifdef CHKINLDEBUG
fprintf(stderr,
"dsc_serv : chkin_loop()     : searching for node using static scheduling \n");
fflush( stderr );
#endif

         /* find node using static scheduling */
         np = dscn_find( 1, 9999, 0);    /* force export */

#ifdef CHKINLDEBUG
if ( np != (NODE_REC*)(EOF) )
 fprintf(stderr,"dsc_serv : chkin_loop()     : %s chosen \n",np->nname);
else
 fprintf(stderr,"dsc_serv : chkin_loop()     : local nodes chosen %s\n",
         my_node);
fflush( stderr );
#endif

       }
       /*                                                         */
       /* when using dynamic scheduling and there is no available */
       /* processor                                               */
       /*                                                         */
       if ( (pssched == TRUE ) && ( np == (NODE_REC *) (DSCQUEUED) ) )
        {
        pp->prob_PID = 0;
        waiting_on_sub++;

#ifdef CHKINLDEBUG
fprintf(stderr,
        "dsc_serv : chkin_loop()     : no room at the inn %s is waiting \n"
        ,pp->prob_name);
fflush( stderr );
#endif

         goto XDONE;
        }
      /*                        */
      /* if not running locally */
      /*                        */
      if (np != (NODE_REC *) EOF) {
       pp->enode_addr = np->naddr;
       rstat = dscp_export( pp, dscp_find(0,pp->parent_PID), &np, &msg2);
       if (rstat != DSC_SUCCESS)  np = (NODE_REC *) EOF;
      } /* End of not running locally */
      /*                    */
      /* if running locally */
      /*                    */
      if (np == (NODE_REC *) EOF) 
      {
       sprintf( logstr, 
             "dsc_serv : chkin_loop()     : %s: running (sub)problem locally\n"
             , pp->prob_name);
       dsc_log( pp, logstr);
       pp->enode_addr = my_addr;
       if (strlen( msg2.src_build ))
        {
         rstat = dscp_build( pp );
         if (rstat != DSC_SUCCESS) 
           {
            pa = dscp_find(0, ntohl(msg2.parent_PID));
            sprintf( logstr, 
                  "dsc_serv : chkin_loop()     : %s: build failed: %s\n",
                   pp->prob_name, msg2.src_build);
            dsc_log( pp, logstr);
            fprintf( stderr, logstr);
            dscp_compl( pp, pa, pp->pnode_addr, pp->prob_name,
                       DSCERR_P_BUILD|rstat);
            goto XDONE; 
           } /* if build failed */
        } /* if build command exists */
      } /* End if running locally */
     } /* End if node declared dead */

XDONE:

#ifdef CHKINLDEBUG
fprintf(stderr,"dsc_serv : chkin_loop()     : returning (void) \n");
fflush( stderr );
#endif

   return;
}
