/*****************************************************************************/
/*                                                                           */
/*                                dsc_pass.c                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* These Rountines implement the DSC_PASSWORD security                       */
/* system. Before a message is sent to a server the file "dsc_setup.csh"     */
/* located in the $HOME directory is searched for the DSC_PASSWORD. Once a   */
/* the receiving server gets the message, it compares the password to the    */
/* DSC_PASSWORD it finds in its "dsc_setup.csh" file in the $HOME directory  */
/* for validation of the message.                                            */
/*         *The dsc_setup.csh file must not be world read/write/execute      */
/*         *The DSC_PATH directory must not be world read/write/execute      */
/*         *If an illegal password is recieved the message is ignored and    */
/*          a note of this occurance appears in the server logfile.          */
/*         *If at anytime the "dsc_setup.csh" file does not contain the      */
/*          DSC_PASSWORD when a server or control program needs it the       */
/*          server or control program is forced to halt.                     */
/*         *Remeber that the only user that can have access to the           */
/*          "dsc_setup.csh" file is the user using the DSC package.          */
/*                                                                           */
/*****************************************************************************/

#include "dscrp.h"        /* include server external variables               */

/*****************************************************************************/
/*                                                                           */
/* functions :                                                               */
/*      dsc_getpass()                   search for dsc pass in setup file    */
/*      dsc_getpass_sub( file )         search for dsc pass in setup file    */
/*      dsc_chkpass( pass, sentpass )   compare file and sent password       */
/*      dsc_getpath()                   search for dsc path in setup file    */
/*      dsc_getpath_sub( file )         search for dsc path in setup file    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpass_sub( file )                                       */
/*                                                                           */
/* inputs:   file - pointer to memory containing the "dsc_setup.csh" file    */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PASSWORD = if found                                        */
/*            "NOT FOUND"  = if not found                                    */ 
/*                                                                           */
/* purpose: loop thru memory in search of the DSC_PASSWORD                   */
/*                                                                           */
/*****************************************************************************/

char* dsc_getpass_sub( file )
char* file;
{
   char *word;       /* word in the "dsc_setup.csh" file                     */
   char *wordbeg;    /* pointer to first character                           */
   int found;        /* flag                                                 */

#ifdef GETPSDEBUG
fprintf(stderr,"dsc_serv : dsc_getpass_sub  : char* file\n");
#endif

   found = FALSE;
   word = (char *)malloc(DSC_BUFFZ);
   wordbeg = word;

   /*                                                           */            
   /* loop through file searching for the DSC_PASSWORD variable */
   /*                                                           */

  while ( *file != EOF ) {
  
   /*                     */
   /* skip leading blanks */
   /*                     */
   while ( ( *file == ' ' ) || (*file == '\n') ) file ++;

   /*                                          */  
   /* get word , a word is delimited by spaces */
   /*                                          */ 
   while  ( ( *file != ' ') && ( *file != EOF ) && (*file != '\n') )
    { *word = *file;
      word ++; file++;
    }
   *word = '\0'; 

   /*                                                         */
   /* the pass word is the word after the string DSC_PASSWORD */
   /*                                                         */
   if ( found == TRUE )
     { 
#ifdef GETPSDEBUG
fprintf(stderr,"dsc_serv : dsc_getpass_sub  : found password %s\n",wordbeg);
#endif
      return(wordbeg);
     }

   if (strcmp(wordbeg,DSCTEXTPASS) == 0 )  found = TRUE;

   word = wordbeg;

  } 

#ifdef GETPSDEBUG
fprintf(stderr,"dsc_serv : dsc_getpass_sub  : could not find password\n");
#endif

 return("NOT_FOUND");
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpass()                                                 */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   DSC_PASSWORD on success                                         */
/*           halt otherwise                                                  */
/*                                                                           */
/* purpose:  search the "dsc_setup.csh" file for a password                  */
/*                                                                           */
/*****************************************************************************/

char *dsc_getpass()
{
  char setupfile[DSC_SETUPFZ]; /* memory copy of the dsc_setup.csh file      */
  char *homedir;               /* user's HOME directory path                 */
  char *setupname[DSC_FILEZ];  /* full path and name of the set up file      */
  char *passwrd;               /* DSC password text variable                 */
  char let;
  int count;
  int rstat;
  FILE *fp;                    /* file pointer to dsc_setup.csh              */
  int errcount;

#ifdef GETPDEBUG
fprintf(stderr,"dsc_serv : dsc_getpass      : void \n");
#endif

  /*                        */
  /* Create setup file name */
  /*                        */
  homedir = getenv("HOME");   
  sprintf(setupname,"%s/",homedir);
  strcat(setupname,DSC_SETUP_NAME);

  fp = fopen (setupname,"r");

  if (fp == NULL) {
   fprintf(stderr,"dsc_serv : dsc_getpass      : can not read %s\n",setupname);
   fprintf(stderr,"dsc_serv : dsc_getpass      : forcing a server halt!\n");
   fprintf(stderr,"dsc_serv : dsc_getpass      : %s halting\n",my_node);
   fflush(stderr);
   exit(DSC_FAILURE);
  }

  /*                                              */
  /* Read set up file and store it in 'setupfile' */
  /*                                              */ 
  count = 0;
  errcount = 0;

#ifdef GETPDEBUG
fprintf(stderr,"dsc_serv : dsc_getpass      : reading %s\n",setupname);
#endif

   /*                                                                  */
   /* due to problems relating to reading the setup file to soon after */
   /* another server has accessed it the first getc call might return  */
   /* an EOF. Try some repeated fopens.                                */
   /*                                                                  */
   let = getc(fp);

#ifdef GETPDEBUG
if ( let == EOF ) 
fprintf(stderr,"dsc_serv : dsc_getpass      : 1st read is an EOF\n");
#endif
 
   while ( ( let == EOF ) && ( errcount < 1000 ) ) {

    fclose(fp);
    fp = fopen (setupname,"r");
    if (fp == NULL) {
    fprintf(stderr,"dsc_serv : dsc_getpass      : in errcount and EOF loop\n");
    fprintf(stderr,"dsc_serv : dsc_getpass      : can not read %s\n",setupname);
    fprintf(stderr,"dsc_serv : dsc_getpass      : forcing a server halt!\n");
    fprintf(stderr,"dsc_serv : dsc_getpass      : %s halting\n",my_node);
    fflush(stderr);
    exit(DSC_FAILURE);
    }
    let = getc(fp);
    errcount++;
#ifdef GETPDEBUG
fprintf(stderr,"dsc_serv : dsc_getpass      : try #%d \n",errcount);    
#endif
   } 

   /*                               */
   /* populate 'setupfile' variable */
   /*                               */
   setupfile[count] = let;
   while (let != EOF )
   {
    count++;
    let = getc(fp);
    setupfile[count] = let;
    }
    setupfile[count+1] = EOF;

    fclose(fp);

   /*                                              */
   /* get the password from the dsc_setup.csh file */
   /*                                              */

   passwrd = dsc_getpass_sub(setupfile);

   if ( strcmp(passwrd,"NOT_FOUND") == 0 ) 
   {
    fprintf(stderr,"dsc_serv : dsc_getpass      : could not find password\n");
    fprintf(stderr,"dsc_serv : dsc_getpass      : in %s\n",setupname);
    fprintf(stderr,"dsc_serv : dsc_getpass      : forcing a server halt!\n");
    fprintf(stderr,"dsc_serv : dsc_getpass      : %s halting\n",my_node);
    fflush(stderr);
    exit(DSC_FAILURE);
   }

#ifdef GETPDEBUG
fprintf(stderr,"dsc_serv : dsc_getpass      : returning %s as password\n"
               ,passwrd );
#endif

 return ( passwrd  );
}

/*****************************************************************************/
/*                                                                           */
/* subroutine:  dsc_chkpass_sub( pass, sentpass )                            */
/*                                                                           */
/* inputs:    pass - password read in the "dsc_setup.csh" file for           */
/*                   associated to this server.                              */
/*        sentpass - password sent to the server via the datagram message    */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_SUCCESS = passwords are the same                           */
/*            DSC_FAILURE = passwords are not the same                       */
/*                                                                           */
/* purpose: string compare passwords                                         */
/*          log invalid password received in ther server log file            */
/*                                                                           */
/*****************************************************************************/

int dsc_chkpass( pass, sentpass)
char *pass;
char *sentpass;
{

 if (strcmp(pass,sentpass) != 0 ) {
  fprintf(stderr,"dsc_serv : dsc_chkpass      : incorrect password recieved\n");
  fprintf(stderr,"dsc_serv : dsc_chkpass      : actual %s\n",pass);
  fprintf(stderr,"dsc_serv : dsc_chkpass      : recieved %s\n",sentpass);
  fprintf(stderr,"dsc_serv : dsc_chkpass      : can not service request\n");
  fflush(stderr);
  return(DSC_FAILURE);
}

#ifdef CHKPDEBUG
  fprintf(stderr,"dsc_serv : dsc_chkpass      : passwords match\n");
  fprintf(stderr,"dsc_serv : dsc_chkpass      : returning DSC_SUCCESS\n");
#endif

 return(DSC_SUCCESS);

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpath_sub( file )                                       */
/*                                                                           */
/* inputs:   file - pointer to memory containing the "dsc_setup.csh" file    */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    DSC_PATH = if found                                            */
/*            "NOT FOUND"  = if not found                                    */
/*                                                                           */
/* purpose: loop thru memory in search of the DSC_PATH                       */
/*                                                                           */
/*****************************************************************************/

char* dsc_getpath_sub( file )
char* file;
{
   char *word;       /* word in the "dsc_setup.csh" file                     */
   char *wordbeg;    /* pointer to first character                           */
   int found;        /* flag                                                 */

#ifdef GETPTSDEBUG
fprintf(stderr,"dsc_serv : dsc_getpath_sub  : char* file\n");
#endif

   found = FALSE;
   word = (char *)malloc(DSC_BUFFZ);
   wordbeg = word;

   /*                                                        */
   /* loop through file searching for the DSC_PATH  variable */
   /*                                                        */

  while ( *file != EOF ) {

   /*                     */
   /* skip leading blanks */
   /*                     */
   while ( ( *file == ' ' ) || (*file == '\n') ) file ++;

   /*                                          */
   /* get word , a word is delimited by spaces */
   /*                                          */
   while  ( ( *file != ' ') && ( *file != EOF ) && (*file != '\n') )
    { *word = *file;
      word ++; file++;
    }
   *word = '\0';

   /*                                                */
   /* the path is the word after the string DSC_PATH */
   /*                                                */
   if ( found == TRUE )
     {
#ifdef GETPTSDEBUG
fprintf(stderr,"dsc_serv : dsc_getpath_sub  : found DSC_PATH %s\n",wordbeg);
#endif
      return(wordbeg);
     }

   if (strcmp(wordbeg,"DSC_PATH") == 0 )  found = TRUE;

   word = wordbeg;

  }

#ifdef GETPTSDEBUG
fprintf(stderr,"dsc_serv : dsc_getpath_sub  : could not find DSC_PATH\n");
#endif

 return("NOT_FOUND");
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_getpath())                                                */
/*                                                                           */
/* inputs:   none                                                            */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   DSC_PATH on success                                             */
/*           halt otherwise                                                  */
/*                                                                           */
/* purpose:  search the "dsc_setup.csh" file for DSC_PATH                    */
/*                                                                           */
/*****************************************************************************/

char *dsc_getpath()
{
  char setupfile[DSC_SETUPFZ]; /* memory copy of the dsc_setup.csh file      */
  char *homedir;               /* user's HOME directory path                 */
  char *setupname[DSC_FILEZ];  /* full path and name of the set up file      */
  char *dpath;                 /* DSC path env variable                      */
  char let;
  int count;
  int rstat;
  FILE *fp;                    /* file pointer to dsc_setup.csh              */
  int errcount;

#ifdef GETPTDEBUG
fprintf(stderr,"dsc_serv : dsc_getpath()    : void \n");
#endif

  /*                        */
  /* Create setup file name */
  /*                        */
  homedir = getenv("HOME");
  sprintf(setupname,"%s/",homedir);
  strcat(setupname,DSC_SETUP_NAME);

  fp = fopen (setupname,"r");

  if (fp == NULL) {
   fprintf(stderr,"dsc_serv : dsc_getpath()    : can not read %s\n",setupname);
   fprintf(stderr,"dsc_serv : dsc_getpath()    : forcing a server halt!\n");
   fprintf(stderr,"dsc_serv : dsc_getpath()    : %s halting\n",my_node);
   fflush(stderr);
   exit(DSC_FAILURE);
  }

  /*                                              */
  /* Read set up file and store it in 'setupfile' */
  /*                                              */
  count = 0;
  errcount = 0;

#ifdef GETPTDEBUG
fprintf(stderr,"dsc_serv : dsc_getpath()    : reading %s\n",setupname);
#endif

   /*                                                                  */
   /* due to problems relating to reading the setup file to soon after */
   /* another server has accessed it the first getc call might return  */
   /* an EOF. Try some repeated fopens.                                */
   /*                                                                  */
   let = getc(fp);

#ifdef GETPTDEBUG
if ( let == EOF )
fprintf(stderr,"dsc_serv : dsc_getpath()    : 1st read is an EOF\n");
#endif

   while ( ( let == EOF ) && ( errcount < 1000 ) ) {

    fclose(fp);
    fp = fopen (setupname,"r");
    if (fp == NULL) {
fprintf(stderr,"dsc_serv : dsc_getpath()    : in errcount and EOF loop\n");
fprintf(stderr,"dsc_serv : dsc_getpath()    : can not read %s\n",setupname);
fprintf(stderr,"dsc_serv : dsc_getpath()    : forcing a server halt!\n");
fprintf(stderr,"dsc_serv : dsc_getpath()    : %s halting\n",my_node);
    fflush(stderr);
    exit(DSC_FAILURE);
    }
    let = getc(fp);
    errcount++;
#ifdef GETPTDEBUG
fprintf(stderr,"dsc_serv : dsc_getpath()    : try #%d \n",errcount);
#endif
   }

   /*                               */
   /* populate 'setupfile' variable */
   /*                               */
   setupfile[count] = let;
   while (let != EOF )
   {
    count++;
    let = getc(fp);
    setupfile[count] = let;
    }
    setupfile[count+1] = EOF;

    fclose(fp);

   /*                                              */
   /* get the DSC_PATH from the dsc_setup.csh file */
   /*                                              */

   dpath = dsc_getpath_sub(setupfile);

   if ( strcmp(dpath,"NOT_FOUND") == 0 )
   {
    fprintf(stderr,"dsc_serv : dsc_getpath()    : could not find DSC_PATH\n");
    fprintf(stderr,"dsc_serv : dsc_getpath()    : in %s\n",setupname);
    fprintf(stderr,"dsc_serv : dsc_getpath()    : forcing a server halt!\n");
    fprintf(stderr,"dsc_serv : dsc_getpath()    : %s halting\n",my_node);
    fflush(stderr);
    exit(DSC_FAILURE);
   }

#ifdef GETPTDEBUG
fprintf(stderr,"dsc_serv : dsc_getpath()    : returning %s as DSC_PATH\n"
               ,dpath );
fflush(stderr);
#endif

 return ( dpath  );

}
