/*****************************************************************************/
/*                                                                           */
/*                                 dsc_ps.c                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* functions:                                                                */
/*      main         - loop and send/save statistics                         */
/*      dsc_psinit() - initialize communication to DSC-Server                */
/*      dsc_sendps() - wrap up datagram and send it to DSC-Server            */
/*      err_exit()   - write error message to log file and exit              */
/*      get_stat()   - get node status from ps command                       */
/*      get_time()   - get local system time and compute week of year        */
/*                                                                           */
/*****************************************************************************/

#include "dscrp.h"        /* include server external variables               */
#include "unistd.h"

/*****************************************************************************/
/*                                                                           */
/*                                 definitions                               */
/*                                                                           */
/*****************************************************************************/

typedef struct
{
  u_char	chour[24],
		mhour[24],
		cwday[7],
		mwday[7],
		cweek[52],
		mweek[52];

} STAT_REC, *STAT_PTR;

#define LLEN		256
#define STAT_SIZE	sizeof (STAT_REC)
#define NAP_TIME	400	

#define C1		1
#define C2		2
#define C3		4
#define C4		8

/*****************************************************************************/
/*                                                                           */
/*                           externals & prototypes                          */
/*                                                                           */
/*****************************************************************************/

void dsc_psinit ();
void dsc_sendps ();
void err_exit ();
void get_stat ();
void get_time ();

/*****************************************************************************/
/*                                                                           */
/*                           module level variables                          */
/*                                                                           */
/*****************************************************************************/

static int	stat_file;
static FILE	*log_file;

static int	cur_cpu=0;
static int	cur_mem=0;

static u_char	table[STAT_SIZE];

static int	weight[3] = { 0, 0, 0 };
static int	wsum;

static fd_set fdset;                   /* in types.h                         */
static struct timeval to;              /* in time.h                          */
static struct sockaddr_in saddr;       /* in netinet/in.h                    */
static struct sockaddr_in rsaddr;      /* in netinet/in.h                    */
static struct linger lo;               /* in socket.h                        */
static int on;                         /* acess option values for setsockopt */
static int length;                     /* size of valid address              */
static DSCPSSTAT ps_msg;
static DSCPSRSTAT rps_msg; 
static NODE_REC node_db[DSC_NODEX];
static int ncnt;
static int dbtop;

int             portnm;                 /* int datagram port number          */
int             sockdg;                 /* datagram descriptor               */
int             my_pid;                 /* this process's PID                */
char            my_node[32];            /* local node name                   */
unsigned long   my_addr;                /* local host address                */
char            pwd[100];               /* expanded path information         */

int             n_proc;                 /* number of processors              */

/*****************************************************************************/
/*                                                                           */
/* subroutine: main (argc, argv)                                             */
/*                                                                           */
/* inputs:    argc    - number of command line arguments                     */
/*            argv[1] - log file name                                        */
/*            argv[2] - datagram port number                                 */
/*            argv[3] - number of processors                                 */
/*                                                                           */
/* outputs:  file updates, current processor status                          */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  daemon program providing load information to the DSC-Server     */
/*                                                                           */
/* updates:  January 1993    - Extended version for hosts with several       */
/*                             processors                                    */
/*                                                                           */
/* authors:  Markus Hitz, Angel Diaz                                         */
/*                                                                           */
/*****************************************************************************/

int main (argc, argv)

int	argc;
char  **argv;
{
STAT_PTR	st;
int		i, nonempty;
int		hour_stat=0, alive=1;
int		cur_h, cur_d, cur_w;
int		prev_h, prev_d, prev_w;
int		tmp_cpu, tmp_mem;
u_char		res_cpu, res_mem;

  if (argc > 3)
  {
    log_file = fopen (argv[1], "a+");
    dup2 (fileno (log_file), fileno (stdout));
    dup2 (fileno (log_file), fileno (stderr));
    portnm = atoi (argv[2]);
    n_proc = atoi (argv[3]);
  }
  else
  err_exit ("dsc_ps   : main()           : insufficient number of parameters");

#ifdef PSMAIN
fprintf(stderr,"dsc_ps   : main()           : int argc %d\n",argc);
for(i=0;i<argc;i++)
fprintf(stderr,"dsc_ps   : main()           : argv[%d] %s\n",i,argv[i]);
i=0;
fflush(stderr);
#endif

  dsc_psinit();

  /*          */
  /* set mask */
  /*          */

  st = (STAT_PTR) table;

  /*                         */
  /* get current date & time */
  /*                         */

  get_time (&cur_h, &cur_d, &cur_w);

  prev_h = cur_h;
  prev_d = cur_d;
  prev_w = cur_w;

  /*                           */
  /* create name for stat_file */
  /*                           */

  i = strcspn (my_node, ".");
  my_node[i] = '\0';
  strcat (my_node, "_dsc_serv.stat");

  /*                       */
  /* try to open stat file */
  /*                       */

  if ((stat_file = open (my_node, O_RDWR, S_IREAD | S_IWRITE)) == -1)
  {
    if (errno != ENOENT)
      err_exit ("dsc_ps   : main()           : open (exist?) failed");

    if ( (stat_file = open (my_node, O_RDWR | O_CREAT, S_IREAD | S_IWRITE)) 
           == -1
       )
      err_exit ("dsc_ps   : main()           : open (creat) failed");

    for (i=0; i<STAT_SIZE; i++) table[i] = 0;	/* initialize table */

    write (stat_file, table, STAT_SIZE);
  }
  else  
  /*                               */
  /* read stat_file and process it */
  /*                               */
  {
    if ((read (stat_file, table, STAT_SIZE)) != STAT_SIZE)
     err_exit ("dsc_ps   : main()           : read failed");

    /*                                               */
    /* set weights for existent long-term statistics */
    /*                                               */

    if (st->chour[cur_h] || st->mhour[cur_h]) weight [2] = C3;
    if (st->cwday[cur_d] || st->mwday[cur_d]) weight [1] = C2;
    if (st->cweek[cur_w] || st->mweek[cur_w]) weight [0] = C1;
  }

  wsum = C4 + weight[2] + weight[1] + weight[0];

#ifdef PSMAIN
fprintf(stderr,"dsc_ps   : main()           : starting alive loop\n");
fflush(stderr);
#endif

  while (alive)
  {
    /*                                   */
    /* if hour passed then do rollover   */
    /* and save new long-term statistics */
    /*                                   */

    get_time (&cur_h, &cur_d, &cur_w);
    hour_stat++;

    if (cur_h != prev_h)
    {
      tmp_cpu = st->chour[prev_h] + cur_cpu;
      st->chour[prev_h] = tmp_cpu / hour_stat;

      tmp_mem = st->mhour[prev_h] + cur_mem;
      st->mhour[prev_h] = tmp_mem / hour_stat;

      hour_stat = 1;
      cur_cpu = 0; cur_mem = 0;
      prev_h = cur_h;

      weight [2] = (st->chour[cur_h] || st->mhour[cur_h] ? C3 : 0 );

      if (cur_d != prev_d)
      {
        tmp_cpu = 0; tmp_mem = 0;
        nonempty = 0;

        for (i=0; i<24; i++)
        {
          if (st->chour[i] || st->mhour[i]) nonempty++;
          tmp_cpu += st->chour[i];
          tmp_mem += st->mhour[i];
        }

        if (st->cwday[prev_d] || st->mwday[prev_d]) nonempty++;
        tmp_cpu += st->cwday[prev_d];
        tmp_mem += st->mwday[prev_d];

        st->cwday[prev_d] = tmp_cpu / nonempty;
        st->mwday[prev_d] = tmp_mem / nonempty;

        prev_d = cur_d;

        weight [1] = (st->cwday[cur_d] || st->mwday[cur_d] ? C2 : 0 );

        if (cur_w != prev_w)
        {
          tmp_cpu = 0; tmp_mem = 0;
          nonempty = 0;

          for (i=0; i<7; i++)
          {
            if (st->cwday[i] || st->mwday[i]) nonempty++;
            tmp_cpu += st->cwday[i];
            tmp_mem += st->mwday[i];
          }

          if (st->cweek[prev_w] || st->mweek[prev_w]) nonempty++;
          tmp_cpu += st->cweek[prev_w];
          tmp_mem += st->mweek[prev_w];

          st->cweek[prev_w] = tmp_cpu / nonempty;
          st->mweek[prev_w] = tmp_mem / nonempty;

          prev_w = cur_w;

          weight [0] = (st->cweek[cur_w] || st->mweek[cur_w] ? C1 : 0 );
        }
      }

      /*                                  */
      /* save modified table to stat_file */
      /*                                  */

      if (lseek (stat_file, 0, SEEK_SET) == -1)
        err_exit ("lseek");

      write (stat_file, table, STAT_SIZE);

      wsum = C4 + weight[2] + weight[1] + weight[0];
    }

    /*                  */
    /* get current load */
    /*                  */

    get_stat (&res_cpu, &res_mem);

    cur_cpu += res_cpu;
    cur_mem += res_mem;

    tmp_cpu = C4 * cur_cpu / hour_stat;
    tmp_mem = C4 * cur_mem / hour_stat;

    /*                                                     */
    /* send current load corrected by long-term statistics */
    /* to DSC-server                                       */
    /*                                                     */

    tmp_cpu += st->chour[cur_h] * C3;
    tmp_cpu += st->cwday[cur_d] * C2;
    tmp_cpu += st->cweek[cur_w] * C1;

    tmp_mem += st->mhour[cur_h] * C3;
    tmp_mem += st->mwday[cur_d] * C2;
    tmp_mem += st->mweek[cur_w] * C1;

    res_cpu = tmp_cpu / wsum;
    res_mem = tmp_mem / wsum;

    dsc_sendps (res_cpu, res_mem);

    /*            */
    /* take a nap */
    /*            */

    sleep (NAP_TIME);
  }

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_psinit()                                                  */
/*                                                                           */
/* inputs:    none                                                           */
/*                                                                           */
/* ext vars:  my_pid, my_node, saddr, hp, sockdg, ps_msg                     */ 
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   initialize communication to DSC-Server                         */
/*                                                                           */
/*****************************************************************************/

void dsc_psinit()
{
int rstat;
char *cptr;
FILE *fp=(NULL);
struct hostent *hp, *gethostbyname();
struct tm *tmp;
time_t tt;

#ifdef PSINIT
fprintf(stderr,"dsc_ps   : dsc_psinit()     : void\n");
#endif

  my_pid = getpid();
  if ( getcwd(pwd,sizeof(pwd)) == NULL )
   {
    perror( "error getting the current woring directory");
    fflush( stderr );
    err_exit("dsc_psinit");
   }

  gethostname(my_node, DSC_NODEZ);
  hp = gethostbyname(my_node);
  saddr.sin_family = AF_INET;
  bcopy( hp->h_addr, &saddr.sin_addr, hp->h_length);
  saddr.sin_port = htons(portnm);
  sockdg = socket(AF_INET,SOCK_DGRAM,0);
  on=1;
  setsockopt( sockdg, SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on));

  /*                          */
  /* Local DSC server message */
  /*                          */
  ps_msg.msg_type = ps_msg.sanity = MSG_PSSTAT;
  ps_msg.msg_size = sizeof(DSCPSSTAT);
  ps_msg.fnode = (unsigned long) (*(long *)hp->h_addr);
  strcpy(ps_msg.fnname,my_node);

  /*                           */
  /* Remote DSC server message */
  /*                           */
  rps_msg.msg_type = rps_msg.sanity = MSG_PSRSTAT;
  rps_msg.msg_size = sizeof(DSCPSRSTAT);
  rps_msg.fnode = (unsigned long) (*(long *)hp->h_addr);
  strcpy(rps_msg.fnname,my_node);

  fp = fopen( DSCRP_NETFILE, "r");
   if (fp == NULL) {
       rstat = errno;
       perror( "error opening DSC net configuration file");
       fprintf( stderr, "\t%s\n", DSCRP_NETFILE);
       fflush( stderr );
       err_exit("dsc_psinit");
       }

  for (ncnt = 0; ncnt<DSC_NODEX; ) {
     cptr = node_db[ncnt].nname;
     fgets( cptr, DSC_NODEZ, fp);
     if (feof(fp)) break;
     if (*cptr == '*') continue;            /* comment line */
     cptr = strchr( cptr, '|');
     if (cptr != 0) {
       *cptr++ = 0;
     }

#ifdef PSINIT
fprintf(stderr,"dsc_ps   : dsc_psinit()     : adding %s to node dbase\n"
                  ,node_db[ncnt].nname);
fflush(stderr);
#endif

     hp = gethostbyname( node_db[ncnt].nname, DSC_NODEZ);
     if (hp == 0) {
        fprintf( stderr, "dsc_ps   : dsc_psinit()     : %s is unknown\n",
                 node_db[ncnt].nname);
        continue;
        }  /* end if node name is not known */

     /*                        */
     /* populate node database */
     /*                        */
     node_db[ncnt].index = ncnt;
     node_db[ncnt].portg = portnm;  /* already in network order */
     node_db[ncnt].naddr = (unsigned long) ( *(long *)hp->h_addr);
     node_db[ncnt].act_prob = (-1);
     node_db[ncnt].blk_prob = (-1);
     node_db[ncnt].err_cnt = 0;
     ncnt++;
     dbtop++;
     }     /* end for all node slots in node database */
 
  /*                                          */
  /* Determine curent time for the Time stamp */
  /*                                          */
  time(&tt);
  tmp = localtime(&tt);

  fprintf (stderr, "Distributed Symbolic Computation environment\n");
  fprintf( stderr, "\t(c) 1990, 1991, 1992, 1993, 1994\n");
  fprintf( stderr, 
"\t(c) Kurt W. Schmitz, Angel L. Diaz, Erich Kaltofen, Markus Hitz\n");
  fprintf( stderr, "time stamp = %s",asctime(tmp));
  fprintf (stderr, "dsc_ps pid = %d\n", my_pid);
  fprintf (stderr, "dsc_ps port number = %d\n",portnm);
  fflush (stderr);

#ifdef PSINIT
fprintf(stderr,"dsc_ps   : dsc_psinit()     : returning void\n"); 
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: dsc_sendps()                                                  */
/*                                                                           */
/* inputs:    cpu - current cpu load                                         */
/*            mem - current memory usage                                     */
/*                                                                           */
/* ext vars:  saddr, sockdg, ps_msg                                          */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   wrap up datagram and send it to DSC-Server                     */
/*                                                                           */
/*****************************************************************************/

void dsc_sendps (cpu ,mem)
int	cpu, mem;
{
int	rlen;
char    mmy_node[DSC_NODEZ];
struct hostent *hp, *gethostbyname();

#ifdef PSSENDPS
fprintf(stderr,"dsc_ps   : dsc_sendps()     : int cpu %d\n",cpu);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : int mem %d\n",mem);
fflush(stderr);
#endif

  ps_msg.cpu = htonl(cpu);
  ps_msg.mem = htonl(mem);
  strcpy (ps_msg.dsc_password,DSCPSPASS);
  rps_msg.cpu = htonl(cpu);
  rps_msg.mem = htonl(mem);
  strcpy (rps_msg.dsc_password,DSCPSPASS);

#ifdef PSSENDPS
fprintf(stderr,"\n");
fprintf(stderr,"dsc_ps   : dsc_sendps()     : Message Contents \n");
fprintf(stderr,"dsc_ps   : dsc_sendps()     : msg_type %d \n",ps_msg.msg_type);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : sanity %d \n",ps_msg.sanity);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : msg_size %d \n",ps_msg.msg_size);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : fnname %s \n",ps_msg.fnname);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : dsc_password %s \n"
              ,ps_msg.dsc_password);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : cpu %d \n",ps_msg.cpu);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : mem %d \n",ps_msg.mem);
fflush(stderr);
#endif 
 

  /*                        */
  /* Send to remote servers */
  /*                        */
  gethostname(mmy_node, DSC_NODEZ);

  for (ncnt = 0; ncnt< dbtop; ){
  hp = gethostbyname(node_db[ncnt].nname);
  rsaddr.sin_family = AF_INET;
  bcopy( hp->h_addr, &rsaddr.sin_addr, hp->h_length);
  rsaddr.sin_port = htons(portnm);
  if (strcmp(mmy_node,node_db[ncnt].nname)!=0)
  {
  do
  {
    errno = 0;
    rlen = sendto (sockdg,
                   (char *) &rps_msg,
                   sizeof(DSCPSSTAT),
                   0,
                   &rsaddr,
                   sizeof(rsaddr));

    if (errno == ENOBUFS) sleep(2);
  } while ((errno == EINTR) || (errno == ENOBUFS));

#ifdef PSSENDPS
fprintf(stderr,"dsc_ps   : dsc_sendps()     : sent cpu %d mem %d to %s\n"
              ,ps_msg.cpu,ps_msg.mem,node_db[ncnt].nname);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : rlen %d sizeof is %d \n",
               rlen,sizeof(DSCPSSTAT));
fflush(stderr);
#endif

  }
   ncnt++;
  }


  /*                      */
  /* Send to local server */
  /*                      */
  do
  {
    errno = 0;
    rlen = sendto (sockdg,
		   (char *) &ps_msg,
		   sizeof(DSCPSSTAT),
		   0,
		   &saddr,
		   sizeof(saddr));

    if (errno == ENOBUFS) sleep(2);
  } while ((errno == EINTR) || (errno == ENOBUFS));

#ifdef PSSENDPS
fprintf(stderr,"dsc_ps   : dsc_sendps()     : sent message to local server %s\n"
              ,my_node);
fprintf(stderr,"dsc_ps   : dsc_sendps()     : rlen %d sizeof is %d \n"
              ,rlen,sizeof(DSCPSSTAT));
fprintf(stderr,"dsc_ps   : dsc_sendps()     : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: err_exit()                                                    */
/*                                                                           */
/* inputs:    s - error message                                              */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   write error message to log file, set error condition and exit  */
/*                                                                           */
/*****************************************************************************/

void err_exit (s)
char	*s;
{

#ifdef PSEXIT
fprintf(stderr,"dsc_ps   : err_exit         : s %s\n",s);
fprintf(stderr,"dsc_ps   : err_exit         : errno %d\n",errno);
fflush(stderr);
#endif
 
  fprintf (log_file, "DSC_PS: %s\n", s);
  fprintf (log_file, "Process terminated\n");
  fclose (log_file);

#ifdef PSEXIT
fprintf(stderr,"dsc_ps   : err_exit         : exiting!\n");
fflush(stderr);
#endif

  exit (1);

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: get_stat()                                                    */
/*                                                                           */
/* inputs:    none                                                           */
/*                                                                           */
/* ext vars:  n_proc                                                         */
/*                                                                           */
/* outputs:   current CPU and MEM load (0..255)                              */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   open pipe to "ps -agux" ("ps agux" for IBM AIX)                */
/*            for each line in stream do                                     */
/*              if not own "ps" extract CPU and MEM                          */
/*            close pipe                                                     */
/*            convert float to u_char                                        */
/*                                                                           */
/*****************************************************************************/

void get_stat (cpu, mem)
u_char	*cpu, *mem;
{

float	pcpu, pmem;
float	totcpu, totmem;
char	line[LLEN];
FILE	*fp;

#ifdef PSSTAT
fprintf(stderr,"dsc_ps   : get_stat()       : int cpu %d\n",*cpu);
fprintf(stderr,"dsc_ps   : get_stat()       : int mem %d\n",*mem);
fflush(stderr);
#endif

  totcpu = 0.0;
  totmem = 0.0;

  if ((fp = popen ("ps -agux", "r")) == NULL)
    err_exit ("popen failed");

  fgets (line, LLEN, fp);

  while (fgets (line, LLEN, fp))
  {
    sscanf (line+14, "%f %f", &pcpu, &pmem);

    if (strstr (line+56, "ps -agux") == NULL)
    {
      totcpu += pcpu;
      totmem += pmem;
    }
  }

  if (ferror (fp))
    err_exit ("fgets from pipe failed");

  pclose (fp);

  totcpu /= n_proc;

  *cpu = (totcpu >= 100.0 ? 255 : 2.55 * totcpu);
  *mem = (totmem >= 100.0 ? 255 : 2.55 * totmem);

#ifdef PSSTAT
fprintf(stderr,"dsc_ps   : get_stat()       : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: get_time()                                                    */
/*                                                                           */
/* inputs:    none                                                           */
/*                                                                           */
/* outputs:   current hour, day and week of year                             */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   get local system time and compute week of year                 */
/*                                                                           */
/*****************************************************************************/

void get_time (h, d, w)
int	*h, *d, *w;
{
time_t		nosec;
struct tm	*tmstr;
int		dy, dd;

#ifdef PSTIME
fprintf(stderr,"dsc_ps   : get_time()       : h %d\n",*h); 
fprintf(stderr,"dsc_ps   : get_time()       : d %d\n",*d);
fprintf(stderr,"dsc_ps   : get_time()       : w %d\n",*w);
fflush(stderr);
#endif

  time (&nosec);

  tmstr = localtime (&nosec);

  *h = tmstr->tm_hour;
  *d = tmstr->tm_wday;
  dy = tmstr->tm_yday;

  dd = *d - dy%7;
  if (dd < 0) dd += 7;

  *w = (dy + dd)/7;
  if (*w > 51) *w = 0;

#ifdef PSTIME
fprintf(stderr,"dsc_ps   : get_time()       : returning void\n");
fflush(stderr);
#endif

}
