##############################################################################
#                                                                            #
#                                   distdet.mpl                              #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
#                               debugging hooks                              #
#                                                                            #
##############################################################################

DET_D_DEBUG := 'DET_D_DEBUG';
DET_D_DEBUG := 1;

with(linalg);

##############################################################################
#                                                                            #
# subroutine: distdet                                                        #
#                                                                            #
# inputs:  subprobsf - source file (without path information) of the         #
#                      subproblems                                           #
#          solfile   - the file name that will contain the determinant       #
#                      of the matrix A                                       #
#          A         - square matrix                                         #
#          iprod     - starting point for prime moduli                       #
#                                                                            # 
#                                                                            #
# outputs: determinant                                                       #
#                                                                            #
# return:  determinant of matrix A                                           #
#                                                                            #
# purpose: maple procedure to compute the determinant of the integer matrix  #
#          A by homomorphic imaging and Chinese remaindering                 #
#                                                                            #
##############################################################################

distdet := proc(subprobsf,solfile,A,iprod)
 local CPU,          # estimated CPU usage
       MEM,          # estimated memory usage in megabytes
       PWD,          # real path of the current working directory
       SOURCE_FILE,  # sourcefile with real path information
       INDEX,        # DSC subproblem index integer 
       fid,          # file id
       build,        # maple build command
       exec,         # maple execute command
       y,            # loop counter
       NUMPROBS,     # the number of primes needed given the initial value
       moduli,       # the array of moduli
       residue,      # the array of determinant residues
       tmod,         # moduli for each subroblem (save can only save names)
       ubound,       # the upper bound for the product of the moduli
       prod,         # the product of the moduli  
       inc,          # the next starting point for the moduli search
       u,            # determinant
       TPROD;        # temporary product
       TU;           # temporary det
       difference;   # difference of result of the chrem and product

 if DET_D_DEBUG = 1 then
   print(`distdet : distdet          : subprobsf `.subprobsf);
   print(`distdet : distdet          : solfile `.solfile);
   print(`distdet : distdet          : iprod `.iprod);
 fi;

 #                                              #
 # Read in DSC Maple User Functions and Library #
 #                                              #
 read(`dsc_maple.mpl`);
 read(`dsc_maple_lib.mpl`);

 #                        #
 # Set problem parameters #
 #                        #
 CPU:= 240;
 MEM:= 10;

 if DET_D_DEBUG = 1 then
   print(`distdet : distdet          : starting dsc_maple deamon`);
 fi;

 #                        #
 # Start dsc_maple deamon #
 #                        #
 system(`dsc_maple commandfile messagepid dscmaplepid &`);
 system(`sleep 2`);

 #                           #
 # Get real path information #
 #                           #
 PWD := dscpr_pwd(`commandfile`,`messagepid`,`dscmaplepid`,`answerfile`);

 SOURCE_FILE:= cat(PWD,`/`,subprobsf);

  if DET_D_DEBUG = 1 then
   print(`distdet : distdet          : PWD `.PWD);
  fi;

 #                                              #
 # Log problem in local server problem database #
 #                                              #
 INDEX := dscdbg_start(`commandfile`,`messagepid`,`dscmaplepid`,PWD,
                       `answerfile`);

  if DET_D_DEBUG = 1 then
   print(`distdet : distdet          : dscdbg_start index `.INDEX);
  fi;

 ubound := evalf(2*hadamard(A)); 

  if DET_D_DEBUG = 1 then
   print(`distdet : distdet          : computing moduli starting at `.iprod);
  fi;
 
 #                    #
 # Compute the Moduli #
 #                    #
 prod := 1; NUMPROBS:= 0;
 inc :=iprod;

 while prod < ubound do
  moduli[NUMPROBS] := nextprime(inc);
  inc := moduli[NUMPROBS]+1;
  prod := prod * moduli[NUMPROBS];
  NUMPROBS := NUMPROBS+1; 
  if DET_D_DEBUG = 1 then
   print(`distdet : distdet          : moduli[`.nm.`] `);
   print(moduli[nm]);
 fi;

 od; 

  if DET_D_DEBUG = 1 then
   print(`distdet : distdet          : preparing files`);
  fi;

 #                                  #
 # Prepare coresponding input files #
 #                                  #
 for y from 0 to (NUMPROBS - 1) do 
 fid:=cat(SOURCE_FILE,y,`.`,`i`);
 tmod := moduli[y];
 save A , tmod, fid;
 od;
 
 #                               #
 # Make copies of the subproblem #
 #                               #
 copyfiles(NUMPROBS,SOURCE_FILE,`mpl`);

 
 for y from 0 to (NUMPROBS - 1) do
     # Prepare arguments
     fid:= cat(SOURCE_FILE,y);
     build:= cat(`dsc_maple.build`);
     exec:= cat(`dsc_maple.int `,subprobsf,y,`.mpl `,subprobsf,y,`.d`,
                ` `,subprobsf,y,`.i`);

    if DET_D_DEBUG = 1 then
     print(`distdet : distdet          : fid`.fid);
    fi;

     #                   #
     # Spawn subproblems #
     #                   #
     rval:= dscpr_sub(`commandfile`,`messagepid`,`dscmaplepid`,
                       PWD,fid,`pi`,build,exec,CPU,MEM,
                      `answerfile`);

    if DET_D_DEBUG = 1 then
     print(`distdet : distdet          : index `.index);
    fi;

 od;

 if DET_D_DEBUG = 1 then
  print(`distdet : distdet          : waiting on all sub tasks`);
 fi;

 #                         #
 # Wait on all subproblems #
 #                         #
 rval := dscpr_wait(`commandfile`,`messagepid`,`dscmaplepid`,
                    PWD,-1,`answerfile`);

 if DET_D_DEBUG = 1 then
  print(`distdet : distdet          : wait returned `.rval);
  print(`distdet : distdet          : computing det from images`);
 fi;

 dscpr_maple_kill(`commandfile`,`messagepid`,`dscmaplepid`,PWD); 

 #                            #
 # get residues from .d files #
 #                            #
 for y from 0 to (NUMPROBS - 1) do
  fid:= cat(SOURCE_FILE,y,`.d`);
  read(fid);
  residue[y] := ddet;
 od;

 u := chrem(convert(residue,list),convert(moduli,list));

 TU := u;
 TPROD := prod;

 difference := TU-TPROD;

 if DET_D_DEBUG = 1 then
  print(`distdet : distdet          : product         `.TPROD);
  print(`distdet : distdet          : chrem           `.TU);
  print(`distdet : distdet          : chrem - product `.difference);
  print(`distdet : distdet          : returning `.TU);
 fi;

 save TU , TPROD , difference, solfile;

 RETURN(TU);
end;

read(`IM_1_50_50`);
answer := distdet(`det`,`distdet.d`,A,3000);
#read(`IM_1_10_10`);
#answer := distdet(`det`,`distdet.d`,A,10000000);
#A := matrix(2,2);
#A[1,1] := 100; A[1,2] := 200;
#A[2,1] := 300; A[2,2] := 400;
#answer := distdet(`det`,`distdet.d`,A,100);

quit;
