##############################################################################
#                                                                            #
#                              rsa_decode.mpl                                #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
# description:                                                               #
#       Subproblem for RSA demonstration package with DSC distributed        # 
#       decoding                                                             #
#       Copyright: Erich Kaltofen, 8/9/1989(Mathematica) and 8/10/1992(Maple)#
#                  Rostyslav Cisyk, 1994                                     #
#       Permission to use provided the copyright notice is not removed       #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
# functions:                                                                 #
#      unscramble     - decode one integer                                   #
#      power_mod      - compute a^e mod m                                    #
#      decode_int     - decode an integer                                    #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
#                                   globals                                  #
#                                                                            #
##############################################################################

# DSC_IFILE - input file name
# DSC_OFILE - output file name

##############################################################################
#                                                                            #
# subroutine: unscramble                                                     #
#                                                                            #
# inputs:  codedint       - encoded integer                                  #
#          magicexp       - magic exponent (computed from 2 private keys)    #
#          publickey      - product of two private keys                      #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: maple procedure to decode one integer                             #
#                                                                            #
##############################################################################

unscramble := proc(codedint, magicexp, publickey)
   power_mod(codedint, magicexp,  publickey)
end;

##############################################################################
#                                                                            #
# subroutine: power_mod                                                      #
#                                                                            #
# inputs:  a              - an integer                                       #
#          e              - an integer                                       #
#          m              - an integer                                       #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  integer a^e mod m                                                 #
#                                                                            #
# purpose: maple procedure to compute a^e mod m                              #
#                                                                            #
##############################################################################

power_mod := proc(a, e, m)
   local half; # a^(e/2) mod m

   if e = 0 then 
     RETURN(1) 
   fi;

   if e = 1 then 
     RETURN(irem(a, m)) 
   fi;

   half := power_mod(a, iquo(e, 2), m);
   if type(e,even) then 
     RETURN(Expand(half*half) mod m)
   else 
     RETURN(Expand(half*half*a) mod m)
   fi
end;

##############################################################################
#                                                                            #
# subroutine: decode_int                                                     #
#                                                                            #
# inputs:  one_input      - input file for subproblem                        #
#          one_output     - output file for subproblem                       #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: data file                                                         #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: maple procedure to read encoded integer, magic exponent and       #
#          public key from one_input, decode the integer and place the       #
#          result in one_output file                                         #
#                                                                            #
##############################################################################

decode_int := proc(one_input,one_output)
   local commandfile,  # dsc_maple DSC command file
         messagepid,   # dsc_maple_message process id file
         dscmaplepid,  # dsc_maple process id file
         answerfile,   # subtask solution file
         msgstr,       # temporary string
         PWD,          # real path
         decode_int;   # decoded integer

   #                                  #
   # Read in DSC Maple User Functions #
   #                                  #
   read(`dsc_maple.mpl`);

   #                        #
   # Start dsc_maple daemon #
   #                        #
   commandfile := cat(one_input,`.cmd`);
   messagepid  := cat(one_input,`.mpid`);
   dscmaplepid := cat(one_input,`.ppid`);
   answerfile  := cat(one_input,`.ans`);
   msgstr := cat(`dsc_maple `,commandfile,` `,messagepid,` `,dscmaplepid,` &`);
   system(msgstr);

   #                          #
   # Wait a while for startup #
   #                          #
   system(`sleep 2`);

   #                           #
   # Get real path information #
   #                           #
   PWD := dscpr_pwd(commandfile,messagepid,dscmaplepid,answerfile);

   #                       #
   # Kill dsc_maple_daemon #
   #                       #
   dscpr_maple_kill(commandfile,messagepid,dscmaplepid,PWD);

   #                            #
   # Wait a while for shut down #
   #                            #
   system(`sleep 2`);

   #              #
   # Remove files #
   #              #
   msgstr := `rm -f `;
   msgstr := cat(msgstr,commandfile);
   system(msgstr);
   msgstr := `rm -f `;
   msgstr := cat(msgstr,messagepid);
   system(msgstr);
   msgstr := `rm -f `;
   msgstr := cat(msgstr,dscmaplepid);
   system(msgstr);
   msgstr := `rm -f `;
   msgstr := cat(msgstr,answerfile);
   system(msgstr);

   #                                                   #
   # Read coded integer, magic exponent and public key #
   #                                                   #
   msgstr := PWD;
   msgstr := cat(msgstr,`/`,one_input);
   read(msgstr);

   #                                                             #
   # Decode the integer and save the result into the output file #
   #                                                             #
   decode_int := unscramble(codedint, magicexp, publickey);
   save decode_int , one_output;

end;

##############################################################################
#                                                                            #
#                              main                                          #
#                                                                            #
##############################################################################

decode_int(DSC_IFILE,DSC_OFILE);

quit;
