##############################################################################
#                                                                            #
#                            rsa_funcs.mpl                                   #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
# description:                                                               #
#       RSA demonstration package with DSC distributed decoding              #
#       Copyright: Erich Kaltofen, 8/9/1989(Mathematica) and 8/10/1992(Maple)#
#                  Rostyslav Cisyk, 1994                                     #
#       Permission to use provided the copyright notice is not removed       #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
# functions:                                                                 #
#      convert/toASCII     - get an ASCII code of a character                #
#      convert/fromASCII   - get a character corresponding to an ASCII code  #
#      encode              - encode text with RSA algorithm                  #
#      blocking_factor     - determine maximum size of an encoded block      #
#      pack                - convert an array of strings to an array of      #
#                            integers                                        #
#      compress            - remove multiple blanks at the end of a line     #
#      unpack              - convert an array of integers back into an array #
#                            of strings                                      #
#      scramble            - encode one integer                              #
#      prepare_text        - output an array of maple strings as a TEXT      #
#                            structure into a file                           #
#      decode_dsc          - decode an array of integers using RSA algorithm #
#                            and DSC                                         #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
#                                   globals                                  #
#                                                                            #
##############################################################################

# CTRLSTRING               - global variable containing printable ASCII 
#                            characters

CTRLSTRING := `.........	
..................... !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_``abcdefghijklmnopqrstuvwxyz{|}~.`;

##############################################################################
#                                                                            #
# subroutine: convert/toASCII                                                #
#                                                                            #
# inputs:  char       - character to be converted                            #
#                                                                            #
# globals: CTRLSTRING - global variable containing printable ASCII characters#
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  ASCII code of a character                                         #
#                                                                            #
# purpose: maple procedure to convert a character to its ASCII code          #
#                                                                            #
##############################################################################

`convert/toASCII` := proc(char)
   local i;    # loop iterator
   
   #                                                #
   # Try to convert the character to its ASCII code #
   #                                                #
   if char = `.` then 
    RETURN(46) 
   fi;
   for i to 128 do
       if substring(CTRLSTRING, i..i) = char then RETURN(i-1) fi
   od;
   
   #                              #
   # The character is unprintable #
   #                              #
   print(`Could not convert character to ASCII`);
   RETURN(0);
end;

##############################################################################
#                                                                            #
# subroutine: convert/fromASCII                                              #
#                                                                            #
# inputs:  ascii_val  - character to be converted                            #
#                                                                            #
# globals: CTRLSTRING - global variable containing printable ASCII characters#
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  a character                                                       # 
#                                                                            #
# purpose: maple procedure to get a character corresponding to the ASCII code#
#                                                                            #
##############################################################################

`convert/fromASCII` := proc(ascii_val)
   substring(CTRLSTRING, (ascii_val+1)..(ascii_val+1))
end;

##############################################################################
#                                                                            #
# subroutine: encode                                                         #
#                                                                            #
# inputs:  cleartext  - an array of strings represting the text              #
#          publickey  - product of two private keys                          # 
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  an array of encoded integers                                      #
#                                                                            #
# purpose: maple procedure to encode cleartext using RSA algorithm           #
#                                                                            #
##############################################################################

encode := proc(cleartext, publickey)
   local b,          # blocking factor
         i,          # loop iterator
         ints,       # array of integers representing cleartext
         codedints,  # array of encoded integers
         string_num, # number of strings in cleartext
         strn;       # loop iterator

   b := blocking_factor(publickey);
   string_num := nops(cleartext);

   #                                            #
   # Convert cleartext to the array of integers #
   #                                            #
   ints := pack(cleartext, b);

   #                                  #
   # Encode each integer in the array #
   #                                  #
   for strn to string_num do
      codedints[strn] := [];
      for i to nops(ints[strn]) do
          codedints[strn] := [op(codedints[strn]), scramble(ints[strn][i], \
                              publickey)]
      od;
   od;

   RETURN(codedints)
end;

##############################################################################
#                                                                            #
# subroutine: blocking_factor                                                #
#                                                                            #
# inputs:  key  - public key                                                 #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: maple procedure to determine maximum size of a block which can be #
#          uniquely encoded with key                                         #
#                                                                            #
##############################################################################

blocking_factor := proc(key) 
   iquo(trunc( evalf(log(key)/log(2)) ), 8) 
end;

##############################################################################
#                                                                            #
# subroutine: pack                                                           #
#                                                                            #
# inputs:  text  - array of strings                                          #
#          b     - blocking factor                                           #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  an array of integers representing text                            #
#                                                                            #
# purpose: maple procedure to convert text into an array of integers         #
#                                                                            #
##############################################################################

pack := proc(text, b)
   local chars,      # string of text with appended blanks
         len,        # correct number of blocks for the string
         i,          # loop iterator
         j,          # loop iterator
         k,          # loop iterator
         strn,       # loop iterator 
         ints,       # array of integers representing text 
         code,       # an element of the array of integers ints 
         nb,         # number of blocks
         blanks,     # string of blanks to be added to a string of text
         string_num, # number of strings in text
         ascii_val;  # an ASCII code of a character

   #                                     #
   # Determine number of strings in text #
   #                                     #
   string_num := nops(text);
   
   #                                   #
   # Convert each string to integer(s) #
   #                                   #
   for strn to string_num do
      nb := length(text[strn])/b;
      if type(nb, integer) then len := nb;
                           else len := trunc(nb + 1);
      fi;
      
      #                                                                   #
      # Append some blanks so that blocking doesn't run out of characters #
      #                                                                   #
      blanks := ``;
      for i to len*b - length(text[strn]) do
         blanks := `` . blanks . ` `;
      od;
      chars := text[strn];
      chars := ``.chars.blanks;

      j := 0; ints[strn] := []; 
      for i to len do
         code := 0;

         #                                                              #
         # Compute code - an integer representation of the string chars #
         #                                                              #
         for k to b do
            j := j+1;
            code := 2**8 * code;
            if j <= length(chars) then
               ascii_val := convert(substring(chars, j..j), toASCII);
               if ascii_val = 0 then
                  print(`string: `.strn.`     j = `.j);
               fi;
               code := code + ascii_val;
            fi;
         od;

         ints[strn] := [op(ints[strn]), code];
      od;
   od;

   RETURN(ints)
end;

##############################################################################
#                                                                            #
# subroutine: compress                                                       #
#                                                                            #
# inputs:  string  - a string to be compressed                               #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  string with trailing blanks removed                               #
#                                                                            #
# purpose: maple procedure to remove multiple blanks added by pack procedure #
#          at the end of the string                                          #
#                                                                            #
##############################################################################

compress := proc(string)
   local str_length, # length of the string without trailing blanks
         i;          # loop iterator

   str_length := length(string);

   for i from length(string) by -1 to 1 do
      if substring(string, i..i) = ` ` then
         str_length := str_length - 1;
      else 
         break;
      fi;
   od;

   RETURN(substring(string, 1..str_length));
end;

##############################################################################
#                                                                            #
# subroutine: unpack                                                         #
#                                                                            #
# inputs:  intlist  - an array of integers representing strings of the text  #
#          b        - blocking factor                                        #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  an array of strings                                               #
#                                                                            #
# purpose: maple procedure to convert an array of integers back into an array#
#          of strings                                                        #
#                                                                            #
##############################################################################

unpack:= proc(intlist, b)
   local len,     # number of integer blocks for a string
         i,       # loop iterator
         j,       # an element of the array 
         k,       # loop iterator 
         block,   # a block of a string
         text,    # an array of strings corresponding to intlist 
         textstr, # a string  
         strn;    # loop iterator    
   
   strn := 1;
   text := [];
   
   #                                                       #
   # Convert each iteger[or group of integers] to a string #
   #                                                       #
   while type(intlist[strn][1], integer) do
      len := nops(intlist[strn]);
      textstr := ``;
      for i to len do
          block := ``; j := intlist[strn][i];
          for k to b do
              block := cat(convert(irem(j, 2**8), fromASCII), block);
              j := iquo(j, 2**8)
          od;
          textstr := cat(textstr, block);
      od;
      
      #                                         #
      # Remove blanks added by pack() procedure #
      #                                         #
      text := [op(text), compress(textstr)];
      strn := strn + 1;

   od;

   RETURN(text)
end;

##############################################################################
#                                                                            #
# subroutine: scramble                                                       #
#                                                                            #
# inputs:  int        - an integer to scramble                               #
#          publickey  - product of two private keys                          # 
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: maple procedure to encode one integer                             #
#                                                                            #
##############################################################################

scramble := proc(int, publickey)
   irem(irem(int*int,  publickey) * int, publickey)
end;

##############################################################################
#                                                                            #
# subroutine: prepare_text                                                   #
#                                                                            #
# inputs:  string_array - an array of strings                                #
#          filename     - name of the output file                            #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: maple procedure which outputs TEXT structure with strings from    #
#          the input array into a file (Reading this file with "read" will   #
#          cause the text to be printed)                                     #
#                                                                            #
##############################################################################

prepare_text := proc(string_array, filename)
   local tmp_str,    # a string from string_array 
         string_num, # number of strings in string_array
         i;          # loop iterator

   #                                        #
   # Do not echo prompt and line separators #
   #                                        #
   interface(echo=0);

   string_num := nops(string_array);

   #                                  #
   # Write "TEXT(" to the output file #
   #                                  #
   writeto(filename);
   lprint(`TEXT(`);
   writeto(terminal);

   #                                                               #
   # Write each line of string_array as part of the TEXT structure #
   #                                                               #
   for i to string_num do
      tmp_str  := string_array[i];
      appendto(filename);
      lprint(``.````.tmp_str.````);
      if i <> string_num then
         lprint(`,`);
      fi;
      writeto(terminal);
   od;

   #                                    #
   # Terminate TEXT structure with ");" #
   #                                    #
   appendto(filename);
   lprint(`);`);
   writeto(terminal);

end;

##############################################################################
#                                                                            #
#  subroutine: decode_dsc                                                    #
#                                                                            #
#  inputs:  subprobsf   - source file (without path information) of the      #
#                         subproblems                                        #
#           solfile     - the file that will contain decoded text            # 
#           codedints   - array of encoded integers                          #
#           privatekeys - two private keys                                   #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  an array of decoded strings                                       #
#                                                                            #
# purpose: maple procedure which decodes codedints using Distributed         #
#          Symbolic Computation environment                                  #
#                                                                            #
##############################################################################

decode_dsc := proc(subprobsf, solfile, codedints, privatekeys)
   local LOW_CPU,     # estimated CPU usage
         MEM,         # estimated memory usage in megabytes
         PWD,         # real path
         SOURCE_FILE, # source file with real path information
         INDEX,       # subproblem index file
         fid,         # temporary file id
         build,       # maple build command
         exec,        # maple execute command
         y,           # loop iterator
         NUMSTRINGS,  # number of strings in the text
         codedint,    # a coded integer
         P,           # first private key
         Q,           # second private key
         magicexp,    # magic exponent
         publickey,   # product of two private keys
         b,           # blocking factor
         ints,        # decoded integers
         dec_text,    # decoded text
         strn,        # loop counter
         numints,     # number of integers per string
         rval;        # return value from dscpr_sub

   #                                  #
   # Read in DSC Maple User Functions #
   #                                  #
   read(`dsc_maple.mpl`);

   #                        #
   # Set problem parameters #
   #                        #
   subprobsf:= subprobsf;
   LOW_CPU := 240;
   MEM := 5;

   #                        #
   # Start dsc_maple daemon #
   #                        #
   system(`dsc_maple commandfile messagepid dscmaplepid &`);
   system(`sleep 2`);

   #                           #
   # Get real path information #
   #                           #
   PWD := dscpr_pwd(`commandfile`,`messagepid`,`dscmaplepid`,`answerfile`);

   SOURCE_FILE:= cat(PWD,`/`,subprobsf);

   #                                              #
   # Log problem in local server problem database #
   #                                              #
   INDEX := dscdbg_start(`commandfile`,`messagepid`,`dscmaplepid`,PWD,
                         `answerfile`);

   #                                         #
   # Calculate magic exponent and public key #
   #                                         #
   P := privatekeys[1];
   Q := privatekeys[2];
   magicexp := iquo(2*(P-1)*(Q-1) + 1, 3); 
   publickey := P*Q;

   #                                                          #
   # Prepare corresponding input files (one integer per file) #
   #                                                          #
   strn := 1;
   while type(codedints[strn][1], integer) do
      numints := nops(codedints[strn]);
      for y from 1 to numints do
         fid := cat(SOURCE_FILE,strn,y,`.`,`i`);
         codedint := codedints[strn][y];   
         save  codedint, magicexp, publickey, fid;
      od;
      strn := strn + 1;
   od;

   NUMSTRINGS := strn - 1;

   #                                #
   # Make copies of the subproblems #
   #                                #
   copyfiles(NUMSTRINGS, codedints, SOURCE_FILE, `mpl`);

   #                         #
   # Decode encoded integers #
   #                         #
   for strn from 1 to NUMSTRINGS do
      for y from 1 to nops(codedints[strn]) do
         #                   #
         # Prepare arguments #
         #                   #
         fid := cat(SOURCE_FILE,strn, y);
         build := cat(`dsc_maple.build`);
         exec := cat(`dsc_maple.int `, subprobsf, strn, y, `.mpl `, subprobsf, 
                      strn, y, `.d`, ` `, subprobsf , strn, y, `.i`);

         #                  #
         # Spawn subproblem #
         #                  #
         rval:= dscpr_sub(`commandfile`,`messagepid`,`dscmaplepid`,
                          PWD,fid,`pi`,build,exec,LOW_CPU,MEM,
                          `answerfile`);
         system(`sleep 1`);
      od;  
   od;

   #                         #
   # Wait on all subproblems #
   #                         #
   rval := dscpr_wait(`commandfile`,`messagepid`,`dscmaplepid`,
                      PWD,-1,`answerfile<`);

   #                            #
   # Shut down dsc_maple daemon #
   #                            #
   dscpr_maple_kill(`commandfile`,`messagepid`,`dscmaplepid`,PWD); 

   #                                                                  #
   # Collect decoded integers, unpack them and write text into a file #
   #                                                                  #
   for strn from 1 to NUMSTRINGS do
      ints[strn] := [];
      for y from 1 to nops(codedints[strn]) do
         fid:= cat(SOURCE_FILE,strn,y,`.d`);
         read(fid);
         ints[strn] := [op(ints[strn]), decode_int];
      od;
   od;

   b := blocking_factor(publickey);

   dec_text := unpack(ints, b);

   prepare_text(dec_text, solfile);

   RETURN(dec_text);
end;

##############################################################################
#                                                                            #
#  subroutine: copyfiles                                                     #
#                                                                            #
#  inputs:  numstrings  - number of encoded strings                          #
#           codedints   - an array of encoded integers                       #
#           SOURCE_FILE - the path and name of the source file without the   #
#                         extension                                          #
#           extension   - the language extension of the source file          #
#                                                                            #
# globals: none                                                              #
#                                                                            #
# outputs: none                                                              #
#                                                                            #
# return:  none                                                              #
#                                                                            #
# purpose: maple procedure to create copies of the source files              #
#                                                                            #
##############################################################################

copyfiles := proc(numstrings,codedints,SOURCE_FILE,extension)
   local readfile,     # name of original file
         writefile,    # name of copy
         y,            # loop iterator
         intfs_echo,   # old value of echo
         intfs_pprint, # old value of prettyprint
         copystr,      # string to be executed
         strn;         # loop iterator

   #                                               #
   # Get the current value of echo and prettyprint #
   #                                               #
   intfs_echo:=interface(echo);
   intfs_pprint:=interface(prettyprint);

   #                                                                         #
   # Do not echo under any circumstances                                     #
   # use the lprint method to render maple expressions in the user interface #
   #                                                                         #
   interface(echo=0,prettyprint=false);
 
   #               #
   # Create copies # 
   #               #
   for strn to numstrings do
      for y to nops(codedints[strn])  do
         readfile:= cat(SOURCE_FILE,`.`,extension);
         writefile:= cat(SOURCE_FILE,strn,y,`.`,extension);
         copystr:= cat(`cp `,readfile,` `,writefile);
         system(copystr);
      od;
   od;
  
   #                                        #
   # Restore values of echo and prettyprint #
   #                                        #
   interface(echo=intfs_echo,prettyprint=intfs_pprint);

end;
