/*****************************************************************************/
/*                                                                           */
/*                            one_level_driver.C                             */
/*                                                                           */
/* Change History:                                                           */
/* ~~~~~~~~~~~~~~~                                                           */
/* (AAP) - Removed unused variables and tightened up declarations.           */
/* (AAP) - Added stream I/O.                                                 */
/* (AAP) - Added call to dscpr_getabs().                                     */
/* (AAP) - Added use of bif abstraction.                                     */
/*                                                                           */
/*****************************************************************************/

#include "bif.C"               // BIF methods

#define NUMPROBS   3           // number of one_level#.C subproblems
#define PERCENTAGE 75          // BIF successful completion percentage
#define RETRY      5           // number of times for BIF to resubmit problems
#define SF         "one_level" // subproblem prefix
#define DEBUG      1           // debugging hook

/*****************************************************************************/
/*                                                                           */
/* subroutine: main(argc,argv)                                               */
/*                                                                           */
/* inputs:   calling sequence one_level_driver one_level_driver.d            */
/*                                                                           */
/* outputs:  none                                                            */
/*                                                                           */
/* return:   0 = DSC_SUCCESS                                                 */
/*           n = error mask                                                  */
/*                                                                           */
/* purpose: This program can be used as a template for using HOMOGENEOUS     */
/*          BIF's to run one level distribution problems within the DSC      */
/*          environment.                                                     */
/*                                                                           */
/*          This module is the driver program which creates a bag of         */
/*          incomplete futures using a homogeneous BIF contstructor which    */
/*          generates a list of identical subtasks inside the BIF.           */
/*                                                                           */
/*          To demonstrate their use, the homogeneous versions of the        */
/*          RemoveTasks() and AddTasks() routines are called to delete       */
/*          and then replace all tasks in the newly created BIF.  Once this  */
/*          has been done, all tasks are submitted to DSC via the RunTasks() */
/*          routine.                                                         */
/*                                                                           */
/*          After all tasks have been run, GetDataFile() reads their data    */
/*          files and the BIF is emptied with a call to Empty().  An empty   */
/*          BIF still exists and can be refilled with new tasks.  In order   */
/*          to destroy a BIF, a call to its destructor must be made.  (This  */
/*          will occur when main() exits.)                                   */
/*                                                                           */
/*****************************************************************************/

main (int argc, char **argv)
{
 int rval=0;             // DSC C user interface return values
 int sum=0;              // sum of data files (one_level#.d)
 int number;             // number read from (one_level#.d) file
 char fileprf[132];      // one_level# file name w/o path
 char fid[132];          // subproblem file id
 char SOURCE_FILE[132];  // full path to source file w/o extention
 char exec[132];         // subproblem exec command
 char build[132];        // subproblem build command
 char scrfile[132];      // local problem scratch file
 char pwd[100];          // current working directory
 FILE *scr;              // pointer to local problem scratch file
 FILE *fp;               // one_level#.d file pointer

 void writefiles (int, char *, int, FILE *);  // creates source and input files
 void quit (FILE *);                          // exits

  //
  // Get absolute path to source file.
  //
  strcpy (SOURCE_FILE, dscpr_getabs (SF));
//                     ~~~~~~~~~~~~~~~~~
  strcpy (scrfile, "one_level_driver.scr");

#ifdef DEBUG
cerr << "driver   : main()           : opening " << scrfile << "\n";
cerr << flush;
#endif

  //
  // Open scratch file.
  //
  if ((scr = fopen (scrfile, "w+")) == NULL) {
     cerr << "driver   : main()           : can not open " << scrfile << "\n";
     cerr << "driver   : main()           : errno " << errno << "\n";
     cerr << flush;
     quit (scr);
    } 

#ifdef DEBUG
fprintf (scr, "driver   : main()           : starting\n");
fprintf (scr,
         "driver   : main()           : creating %d input and source files\n",
         NUMPROBS);
fflush (scr);
#endif

  //
  // Create input files and subproblems
  //
  writefiles (NUMPROBS, SOURCE_FILE, 2, scr);

#ifdef DEBUG
fprintf (scr, "driver   : main()           : Placing tasks in BIF\n");
fflush (scr);
#endif

  //
  // Create a BIF with all the necessary data - in this example, we are
  //   defining a HOMOGENEOUS BIF with a success percentage of PERCENTAGE.
  //
  BIF *one_level = new BIF (NUMPROBS, PERCENTAGE, SF, scr, HIGH_CPU, 8);
  //  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  //
  // Call RemoveTask to delete all tasks from bif.  This is done here
  //   simply to demonstrate the HOMOGENEOUS RemoveTask() method, which
  //   will remove the task most recently added to the BIF's task list.
  //
  for (int i = 0 ; i < NUMPROBS ; i++) {
#ifdef DEBUG
fprintf (scr, "driver   : main()           : Removing a task\n");
fflush (scr);
#endif
      one_level -> RemoveTask ();
  //  ~~~~~~~~~~~~~~~~~~~~~~~~~~
  }

  //
  // Call AddTask to add all tasks back into bif.  This is done here
  //   simply to demonstrate the HOMOGENEOUS AddTask() method, which will
  //   add a task to the BIF which is identical to the tasks defined by
  //   the parameters specified in the homogeneous BIF constructor above.
  //
  for (i = 0 ; i < NUMPROBS ; i++) {
#ifdef DEBUG
fprintf (scr, "driver   : main()           : Adding a task\n");
fflush (scr);
#endif
       one_level -> AddTask ();
  //   ~~~~~~~~~~~~~~~~~~~~~~~
  }

#ifdef DEBUG
fprintf (scr, "driver   : main()           : Running tasks\n");
fflush (scr);
#endif

  //
  // Call RunTasks to run tasks in BIF.  This routine will wait on the
  //   successful completion of the BIF's percentage of tasks, resubmitting 
  //   incomplete tasks RETRY times, if necessary, to achieve success.
  //
  rval = one_level -> RunTasks (RETRY);
  //     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef DEBUG
fprintf (scr, "driver   : main()           : Tasks completed\n");
fprintf (scr, "driver   : main()           : RunTasks rval %d\n", rval);
fprintf (scr, "driver   : main()           : reading data files\n");
fflush (scr);
#endif

  //
  // Read date file from successfully completed subtasks.
  //
  while ((strcmp ((strcpy (fid, one_level -> GetDataFile ())), "nil")) != 0) {
  //                            ~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef DEBUG
fprintf (scr, "driver   : main()           : reading %s\n",fid);
fflush (scr);
#endif
         fp = fopen (fid, "r");
         if (fp == NULL) {
           cerr << "driver   : main()           : can not open " << fid << endl;
           cerr << "driver   : main()           : errno " << errno << endl;
           cerr << flush;
           quit (scr);
         }
         fscanf (fp, "%d", &number);

#ifdef DEBUG
fprintf (scr, "driver   : main()           : read in %d\n", number);
fflush (scr);
#endif
         fclose (fp);
         sum = sum + number ;
  }

  //
  // Write sum to data file.
  //
  strcpy (fid, argv[1]);

  //
  // Empty BIF, so that it can be reused, if desired.
  //
#ifdef DEBUG
fprintf (scr, "driver   : main()           : emptying BIF\n");
fflush (scr);
#endif
   one_level -> Empty();
// ~~~~~~~~~~~~~~~~~~~~

#ifdef DEBUG
fprintf (scr, "driver   : main()           : opening %s\n", fid);
fflush (scr);
#endif

  fp = fopen (fid, "w");
  if (fp == NULL) {
    cerr << "driver   : main()           : cannot open " << fid << endl;
    cerr << "driver   : main()           : errno " << errno << endl;
    cerr << flush;
    quit (scr);
  }

#ifdef DEBUG
fprintf (scr, "driver   : main()           : writing sum %d\n",sum);
fflush (scr);
#endif

    fprintf (fp, "%d", sum);

    fflush (fp);
    fclose (fp);

    quit (scr);
}


/*****************************************************************************/
/* subroutine: quit (FILE *)                                                 */
/*                                                                           */
/* purpose:    exit                                                          */
/*****************************************************************************/
void quit (FILE *scr)
{
#ifdef DEBUG
fprintf(scr, "driver   : main()           : exiting!\n");
#endif
     cerr << flush;
     fflush (scr);
     fclose (scr);

     exit (0);
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: writefiles(numfiles, SOURCE_FILE , number, scr)               */
/*                                                                           */
/* inputs:    numfiles    - number of file pairs to create                   */
/*            SOURCE_FILE - c source file w/o extension                      */
/*            number      - number to populate input file                    */ 
/*            scr         - pointer to scratch file                          */
/*                                                                           */
/* outputs:  input and source files                                          */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  create numfiles source and input file pairs                     */
/*                                                                           */
/*****************************************************************************/

void writefiles (int numfiles, char *SOURCE_FILE, int number, FILE *scr)
{
  int y;
  char fid[64];
  char temp[80];
  char readfile[132];
  char writefile[132];
  FILE *fp;
  FILE *readptr; 
  FILE *writeptr;

  strcpy(readfile,SOURCE_FILE);
  strcat(readfile,".C"); 

  /*                        */
  /* Set the nth input file */
  /*                        */

  for (y=0; y < numfiles; y++)
  {
    sprintf(fid,"%s%d",SOURCE_FILE,y);
    strcat(fid,".i");

#ifdef DEBUG
fprintf(scr,"driver   : writefile        : opening %s\n",fid);
fflush(scr);
#endif

   fp = fopen(fid,"w");
   if (fp == NULL)
     {
      fprintf(scr,"driver   : writefile        : can not open %s\n",fid);
      fprintf(scr,"driver   : writefile        : errno %d\n",errno);
      fflush(scr);
      fclose(scr);
      fclose(fp);
      exit(errno);
     }
  
#ifdef DEBUG
fprintf(scr,"driver   : writefile        : putting in %d %d\n",
           number,number);
fflush(scr);
#endif

    fprintf(fp,"%d %d",number,number);
    fclose(fp);
  }

  /*                         */
  /* Set the nth source file */
  /*                         */

  for (y=0; y < numfiles; y++)
  {
  strcpy(writefile,SOURCE_FILE);
  strcat(writefile,"%d.C");
  sprintf(writefile,writefile,y);

#ifdef DEBUG
fprintf(scr,"driver   : writefile        : reading from %s\n",readfile);
fprintf(scr,"driver   : writefile        : writing to %s\n",writefile);
fflush(scr);
#endif

  readptr = fopen(readfile,"r");
   if (readptr == NULL)
     {
      fprintf(scr,"driver   : writefile        : can not open %s\n",readfile);
      fprintf(scr,"driver   : writefile        :errno %d\n",errno);
      fflush(scr);
      fclose(scr);
      fclose(fp);
      exit(errno);
     }

  writeptr = fopen(writefile, "w");
   if (writeptr == NULL)
     {
      fprintf(scr,"driver   : writefile        : can not open %s\n",writefile);
      fprintf(scr,"driver   : writefile        : errno %d\n",errno);
      fflush(scr);
      fclose(scr);
      fclose(fp);
      exit(errno);
     }

  /*                  */
  /* copy source code */
  /*                  */
  while( !feof(readptr)) {
          temp[0] = '\0';
          fgets( temp, 80,readptr);
          if (strlen(temp) > 0) fputs( temp,writeptr);
        }
  fclose(readptr);
  fclose(writeptr);
  
  }

}
