##############################################################################
#                                                                            #
#                           one_level_driver.mpl                             #
#                                                                            #
##############################################################################

##############################################################################
#                                                                            #
#                               debugging hooks                              #
#                                                                            #
##############################################################################

DSC_M_ONE_DRIVER := 'DSC_M_ONE_DRIVER';
DSC_M_ONE_DRIVER := 1;

##############################################################################
#                                                                            #
# subroutine: one_level_driver                                               #
#                                                                            #
# inputs:  subprobsf - source file (without path information)                #
#                      of the subproblems                                    #
#          one_answer - the file name that will contains the sum of          #
#                       the subproblem datafile                              #
#                                                                            #
# outputs: sub tasks, files ...                                              #
#                                                                            #
# return:  sum of subproblem datafiles                                       #
#                                                                            #
# purpose: spwan NUMPROBS Maple code found in one_level.mpl file using       #
#          the DSC Maple user interface functions, wait on all subproblems   #
#          to complete and store the sum of the subproblem datafiles in      #
#          the one_answer file                                               #
#                                                                            #
##############################################################################

one_level_driver:= proc(subprobsf,one_answer)
 local SF,           # source file w/o path information
       NUMPROBS,     # number of subproblems
       y,            # loop iterator
       rval,         # return value of DSC Maple Interface functions
       total,        # sum of data files
       fid,          # file id string
       SOURCE_FILE,  # source file with path information 
       exec,         # execute command
       build,        # build command
       PWD,          # path of current working directory
       INDEX,        # subroblem index
       LOW_CPU,      # estimated cpu usage
       LOW_MEM;      # estimated mem usage

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : subprobsf %s\n`,subprobsf);
  printf(`dirver   : main             : one_answer %s\n`,one_answer);
 fi;

 #                                  #
 # Read in DSC Maple User Functions #
 #                                  #
 read(`dsc_maple.mpl`);

 #                                          # 
 # Read in DSC Maple Library User Functions #
 #                                          #
 read(`dsc_maple_lib.mpl`);

 #                        #
 # Set problem parameters #
 #                        #
 SF:= subprobsf;
 NUMPROBS:= 3;
 total := 0;
 LOW_CPU:= 80;
 LOW_MEM:= 5;

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : starting dsc_maple deamon\n`);
 fi;

 #                        #
 # Start dsc_maple deamon #
 #                        #
 system(`dsc_maple commandfile messagepid dscmaplepid &`);
 system(`sleep 2`);

 #                           #
 # Get real path information #
 #                           # 
 PWD := dscpr_pwd(`commandfile`,`messagepid`,`dscmaplepid`,`answerfile`);

 SOURCE_FILE:= cat(PWD,`/`,SF);

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : PWD %s\n`,PWD);
 fi;

 #                                              # 
 # Log problem in local server problem database #
 #                                              #
 INDEX := dscdbg_start(`commandfile`,`messagepid`,`dscmaplepid`,PWD,
                       `answerfile`);

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : dscdbg_start index %d\n`,INDEX);
  printf(`dirver   : main             : making copies of sub tasks\n`);
 fi;

 #                               #
 # Make copies of the subproblem #
 #                               # 
 copyfiles(NUMPROBS,SOURCE_FILE,`mpl`);

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : preparing input files\n`);
 fi;

 #                                  #
 # Prepare coresponding input files #
 #                                  # 
 for y from 0 to (NUMPROBS - 1) do
     fid:= cat(SOURCE_FILE,y,`.i`);
     I1 := y;
     I2 := y;
     save I1, I2, fid;
 od; 

 for y from 0 to (NUMPROBS - 1) do
     #                   #
     # Prepare arguments #
     #                   #
     fid:= cat(SOURCE_FILE,y);
     build:= cat(`dsc_maple.build`);
     exec:= cat(`dsc_maple.int `,SF,y,`.mpl `,SF,y,`.d`,` `,SF,y,`.i`);

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : spawning %s\n`,fid);
 fi;

     #                   #
     # Spawn subproblems #
     #                   #
     rval:= dscpr_sub(`commandfile`,`messagepid`,`dscmaplepid`,
                       PWD,fid,`pi`,build,exec,LOW_CPU,LOW_MEM,
                      `answerfile`);

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : index %d\n`,rval);
 fi;

 od;

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : waiting on all sub tasks\n`);
 fi;

     #                         #
     # Wait on all subproblems #
     #                         #
     rval := dscpr_wait(`commandfile`,`messagepid`,`dscmaplepid`,
                         PWD,-1,`answerfile`);

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : wait returned %d\n`,rval);
  printf(`dirver   : main             : reading and summing data files\n`);
 fi;
 
 #                        #
 # Read and sum datafiles #
 #                        #
 for y from 0 to (NUMPROBS-1) do
  fid:= cat(SOURCE_FILE,y,`.d`);
    if is_file_exist(fid) = 1 then
     read(fid);
     total := total + sum;
     sum := 'sum';
    fi;
 od;

 save total , one_answer;

 dscpr_maple_kill(`commandfile`,`messagepid`,`dscmaplepid`,PWD);

 if DSC_M_ONE_DRIVER = 1 then
  printf(`dirver   : main             : returning %d\n`,total);
 fi;

RETURN(total);

end;

answer := one_level_driver(`one_level`,`one_level_driver.d`);
quit;
