/*****************************************************************************/
/*                                                                           */
/*                               dsc_seestat.c                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* functions:                                                                */
/*      main         - display statistics from stat file                     */
/*      err_exit()   - write error message to log file and exit              */
/*      get_stat()   - get node status from ps command                       */
/*      get_time()   - get local system time and compute week of year        */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sys/timeb.h>

/*****************************************************************************/
/*                                                                           */
/*                                 definitions                               */
/*                                                                           */
/*****************************************************************************/

typedef struct
{
  u_char	chour[24],
		mhour[24],
		cwday[7],
		mwday[7],
		cweek[52],
		mweek[52];

} STAT_REC, *STAT_PTR;

#define LLEN		256
#define STAT_SIZE	sizeof (STAT_REC)
#define NAP_TIME	400

#define C1		1
#define C2		2
#define C3		4
#define C4		8

/*****************************************************************************/
/*                                                                           */
/*                           externals & prototypes                          */
/*                                                                           */
/*****************************************************************************/

extern int      errno;
void err_exit ();
void get_stat ();
void get_time ();

/*****************************************************************************/
/*                                                                           */
/*                           module level variables                          */
/*                                                                           */
/*****************************************************************************/

static int	stat_file;

static int	cur_cpu=0;
static int	cur_mem=0;

static u_char	table[STAT_SIZE];

static int	weight[3] = { 0, 0, 0 };
static int	wsum;
int             n_proc;                 /* number of processors              */

/*****************************************************************************/
/*                                                                           */
/* subroutine: main (argc, argv)                                             */
/*                                                                           */
/* inputs:    argc    - number of command line arguments                     */
/*            argv[1] - log file name                                        */
/*            argv[2] - number of processors                                 */
/*                                                                           */
/* outputs:   stats found in argv[1]                                         */
/*                                                                           */
/* return:   none                                                            */
/*                                                                           */
/* purpose:  display stats found in argv[1]                                  */
/*                                                                           */
/* authors:  Markus Hitz, Angel Diaz                                         */
/*                                                                           */
/*****************************************************************************/

main (argc, argv)

int     argc;
char  **argv;
{
STAT_PTR	st;
int		i, hour_stat=0, alive=1;
int		cur_h, cur_d, cur_w;
int		prev_h, prev_d, prev_w;
int		tmp_cpu, tmp_mem;
u_char		res_cpu, res_mem;

  /* set mask */

  st = (STAT_PTR) table;

  /* get current date & time */

  get_time (&cur_h, &cur_d, &cur_w);

  prev_h = cur_h;
  prev_d = cur_d;
  prev_w = cur_w;

  if ( argc > 2 )
   n_proc = atoi (argv[2]);
  else 
  {
   fprintf(stderr,"dsc_seest: main()       : insufficient arguments\n");
   fflush(stderr);
   err_exit ("insufficient args"); 
  }

  /* open stat_file */

  if ((stat_file = open (argv[1], O_RDWR)) == -1)
    err_exit ("open");

  read (stat_file, table, STAT_SIZE);

  printf ("\nH-CPU:");
  for (i=0; i<24; i++) printf (" %u", st->chour[i]);

  printf ("\nH-MEM:");
  for (i=0; i<24; i++) printf (" %u", st->mhour[i]);

  printf ("\nD-CPU:");
  for (i=0; i<7; i++) printf (" %u", st->cwday[i]);

  printf ("\nD-MEM:");
  for (i=0; i<7; i++) printf (" %u", st->mwday[i]);

  printf ("\nW-CPU:");
  for (i=0; i<52; i++) printf (" %u", st->cweek[i]);

  printf ("\nW-MEM:");
  for (i=0; i<52; i++) printf (" %u", st->mweek[i]);
  printf ("\n");

  close (stat_file);

  /* get current load */

  get_stat (&res_cpu, &res_mem);

  printf ("Time: h %d  d %d  w %d\n", cur_h, cur_d, cur_w);
  printf ("\nLoad: CPU %u  MEM %u\n", res_cpu, res_mem);
}

/*****************************************************************************/
/*                                                                           */
/* subroutine: err_exit()                                                    */
/*                                                                           */
/* inputs:    s - error message                                              */
/*                                                                           */
/* outputs:   none                                                           */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   write error message to log file, set error condition and exit  */
/*                                                                           */
/*****************************************************************************/

void err_exit (s)
char    *s;
{

#ifdef PSEXIT
fprintf(stderr,"dsc_seest: err_exit         : s %s\n",s);
fprintf(stderr,"dsc_seest: err_exit         : errno %d\n",errno);
fflush(stderr);
#endif

  fprintf (stderr,"DSC_PS: %s\n", s);
  fprintf (stderr, "Process terminated\n");

#ifdef PSEXIT
fprintf(stderr,"dsc_seest: err_exit         : exiting!\n");
fflush(stderr);
#endif

  exit (1);

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: get_stat()                                                    */
/*                                                                           */
/* inputs:    none                                                           */
/*                                                                           */
/* ext vars:  n_proc                                                         */
/*                                                                           */
/* outputs:   current CPU and MEM load (0..255)                              */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   open pipe to "ps -agux" ("ps agux" for IBM AIX)                */
/*            for each line in stream do                                     */
/*              if not own "ps" extract CPU and MEM                          */
/*            close pipe                                                     */
/*            convert float to u_char                                        */
/*                                                                           */
/*****************************************************************************/

void get_stat (cpu, mem)
u_char  *cpu, *mem;
{

float   pcpu, pmem;
float   totcpu, totmem;
char    line[LLEN];
FILE    *fp;

#ifdef PSSTAT
fprintf(stderr,"dsc_seest: get_stat()       : int cpu %d\n",*cpu);
fprintf(stderr,"dsc_seest: get_stat()       : int mem %d\n",*mem);
fflush(stderr);
#endif
  totcpu = 0.0;
  totmem = 0.0;

  if ((fp = popen ("ps -agux", "r")) == NULL)
    err_exit ("popen failed");

  fgets (line, LLEN, fp);

  while (fgets (line, LLEN, fp))
  {
    sscanf (line+14, "%f %f", &pcpu, &pmem);

    if (strstr (line+56, "ps -agux") == NULL)
    {
      totcpu += pcpu;
      totmem += pmem;
    }
  }

  if (ferror (fp))
    err_exit ("fgets from pipe failed");

  pclose (fp);

  totcpu /= n_proc;

  *cpu = (totcpu >= 100.0 ? 255 : 2.55 * totcpu);
  *mem = (totmem >= 100.0 ? 255 : 2.55 * totmem);

#ifdef PSSTAT
fprintf(stderr,"dsc_seest: get_stat()       : returning void\n");
fflush(stderr);
#endif

}

/*****************************************************************************/
/*                                                                           */
/* subroutine: get_time()                                                    */
/*                                                                           */
/* inputs:    none                                                           */
/*                                                                           */
/* outputs:   current hour, day and week of year                             */
/*                                                                           */
/* return:    void                                                           */
/*                                                                           */
/* purpose:   get local system time and compute week of year                 */
/*                                                                           */
/*****************************************************************************/

void get_time (h, d, w)
int     *h, *d, *w;
{
time_t          nosec;
struct tm       *tmstr;
int             dy, dd;

#ifdef PSTIME
fprintf(stderr,"dsc_seest: get_time()       : h %d\n",*h);
fprintf(stderr,"dsc_seest: get_time()       : d %d\n",*d);
fprintf(stderr,"dsc_seest: get_time()       : w %d\n",*w);
fflush(stderr);
#endif

  time (&nosec);

  tmstr = localtime (&nosec);

  *h = tmstr->tm_hour;
  *d = tmstr->tm_wday;
  dy = tmstr->tm_yday;

  dd = *d - dy%7;
  if (dd < 0) dd += 7;

  *w = (dy + dd)/7;
  if (*w > 51) *w = 0;

#ifdef PSTIME
fprintf(stderr,"dsc_seest: get_time()       : returning void\n");
fflush(stderr);
#endif

}
